//------------------------------------------------------------
// $Id: PErecvStandAlone.cc,v 1.5.2.1 2000/06/14 15:00:10 yamauchi Exp $ 
//------------------------------------------------------------
//	PE.cc Processing Element 
//	1996/3/22, 1997/1/21
//	(C) 1997--1998 Copyright YAMAUCHI Hitoshi  
//------------------------------------------------------------
#ifndef FOR_STANDALONE
# error "This file is for stand alone mode."
#endif

#include <stdio.h>
#include "PE.h"
#include "Option.h"
#include "FileInfo.h"
#include "volume.h"
#include "mpi2_funcs.h"

//------------------------------------------------------------
// for stand alone
//------------------------------------------------------------
void PEClass::
receiveSysinfoNinitSpaceInfo(const SysInfoClass&    sys,
			     const ViewDataClass&   vw,
			     const SpaceParam&	    sp)
{
    if(this->mObjList.empty() == Q_NOT_EMPTY) // ʤˤ
    {
	// ޤĶ狼ʤΤ object ˥åȤƤ롣
	cout << "Error! : there are objects in list PE(" << this->mID << 
	    ") at PEClass::receiveSysinfo" << endl;
	this->exitprog();
    }

    this->mSysInfo 	= sys;
    this->mViewData	= vw;
    this->mSpcParam	= sp;

    // initialize after receiving sysinfo, viewdata, spaceparam
    this->initInternalUnits();

    if(Option::getInstance()->getVerboseMode() == YES)
    {
	cout << genHd(VERBOSEHEAD, this->getID()) << "receive sysinfo data."
	<< endl;
    }
}

// receive materials 
void PEClass::
receiveMaterials(const materialClass &mat)
{
    this->mMatList.append(mat);
    // this->outRecvMaterialInfo();ȤꤿƤͭפʾǤϤʤΤǽϤʤ
}

//------------------------------
// volume Ϣ
//------------------------------
void PEClass::
receiveVolume(VolumeClass* pVolume)
{
    if(pVolume->isNullVolume() == NO)
    {
	VolumeClass* pRecvVolume = new VolumeClass(*pVolume);
	this->mVolList.append(pRecvVolume);
    }else
    {
	// all null volume is received
	this->initRecvVolumeSet();
	// ޤǤϳ volume ݻƤ뤳Ȥ狼ä
    }
}

//------------------------------
// objects (sizeof = ) μ
//------------------------------
void PEClass::
receiveObjects(objBaseClass *pObj)
{
    if(pObj->who() != NULL_OBJ_E)
    {
	this->mObjList.append(pObj);
    }else
    {
	// get NULL Object ... λ
	this->initObjectMaterialSet();
	this->outRecvObjectInfo();
    }
}

//------------------------------
// ʪ¸ɽμ
//------------------------------
void PEClass::
receiveObjInSpcMap(int bytesize,  OBJINSPC_IMPLEMENT_TYPE * pDatArray)
{
    int size[3];
    size[0] = size[1] = size[2] = this->mSpcParam.getSubDivNum();
    const int totalSize = size[0] * size[1] * size[2];
    if((totalSize <= 0) ||	//  SpcParam ϼѤߤΤϤ
       (totalSize != bytesize))
    {
	cout << "Internal Error! : PEClass::receiveObjInSpcMap" << endl;
	commMpi2_exit(1);
    }

    // Map ΰȽ
    this->mDDAUnit.mObjInSpcMap.initBySizeWithBaseZero(size); 
    
    // ƤΥԡ
    for(int i = 0; i < this->mDDAUnit.mObjInSpcMap.getTotalSize(); i++)
    {
	objectExistInASpaceClass oeias(pDatArray[i]);
	this->mDDAUnit.mObjInSpcMap.setAsOneDimArray(i, oeias);
    }
}

//------------------------------------------------------------
// write geometry file
// @param isShotRad : YES ξˤ shot ΤߤϤ
//	  	      NO  ξˤ unshot ΤߤϤ
// ǽ̤񤯾ˤ PEClass::finalizeForRadiosity Ƥ֤礹
//------------------------------------------------------------
void PEClass::
writeGeometryFile(FILE *pFp, YorN isShotRad)
{
    char buf[1024];
    int outputType = ((isShotRad == YES) ? 
		      WITH_SHOT_RADOUT_e : WITH_UNSHOT_RADOUT_e);
    int oldmatid = -1;
    Pix pidx;
    for(pidx = this->mObjList.first(); pidx != NULL; this->mObjList.next(pidx))
    {
	if(oldmatid != this->mObjList(pidx)->getMaterialID())
	{
	    oldmatid = this->mObjList(pidx)->getMaterialID();
	    fprintf(pFp, "m %s\n", 
		    (this->mMaterialSet.get(oldmatid))->getName());
	}
	this->mObjList(pidx)->getMpi2Form(outputType, buf);
	fprintf(pFp, "%s\n",  buf);
    }
}

//------------------------------------------------------------
// output radiosity intermediate result
// @param isShotRad == YES λˤ shot radiosity Ϥ
//		       NO  λˤ unshot radiosity Ϥ
//------------------------------------------------------------
void PEClass::
outRadiosityMpi2Data(const char *pOutBaseFname, YorN isShotRad)
{
    // filename buf
    const int FNAMEBUFSIZE = 128;
    char buf[FNAMEBUFSIZE];
    sprintf(buf, "%s.r%03d.mpi2.geo", pOutBaseFname, this->getMyRank());
    FILE *pFp = NULL;

    char realWriteFName[FNAMEBUFSIZE];
    if((pFp = getUniqueWriteFP(buf, realWriteFName, FNAMEBUFSIZE)) == NULL)
    {
	cout << "Error! : can not open [" << buf << "]" << endl;
	commMpi2_exit(1);
    }

    cout << NORMALHEAD << " radiosity intermediate result out. Writing ["
	 << buf << "] ..."; cout.flush();

    // write Hemi-cube num 
    fprintf(pFp, "# create HC num is %d\n", this->mGStat.getHCGenNum());
    this->writeGeometryFile(pFp, isShotRad);
    fclose(pFp);

    // gzip output file
    char combuf[FNAMEBUFSIZE];
    sprintf(combuf, "gzip %s", buf);
    system(combuf);

    cout << " save done." << endl;
}
