/**-----------------------------------------------------------
 * $Id: ProfileMon.h,v 1.2 1999/12/07 03:50:02 yamauchi Exp $
 *------------------------------------------------------------
 *	Profile Monitor 
 *	1999/10/14 
 *	Copyright (C) 1999 YAMAUCHI Hitoshi :  
 *
 * ǽư
 *	send  recv Ʊ event ³ˤϽϤʤ
 */
#ifndef _ProfileMon_h
#define _ProfileMon_h 1

#include <iostream.h>
#include <stdio.h>
#include "myString.h"
#include "ElapsedTimer.h"
#include "config.h"
#include "FileInfo.h"

/*
 * USE_PROFILE_MON  define ƤʤˤϤδؿϻѤʤ
 */
#ifdef USE_PROFILE_MON
static const int Use_Profile_Monitor = 1; // Use
#else 
static const int Use_Profile_Monitor = 0; // not use
#endif

/*
 * USE_IBM_PMARRAY  define Ƥʤˤ ibm_pm_array 
 * δؿѤʤ 
 */
#ifdef USE_IBM_PMARRAY 
static const int Use_IBM_Program_MarkerArray = 1; // Use
#else 
static const int Use_IBM_Program_MarkerArray = 0; // not use
#endif

/*
 * USE_MPI2_PMARRAY  define Ƥʤˤ mpi2_pm_array ط
 * δؿѤʤ 
 */
#ifdef USE_MPI2_PMARRAY 
static const int Use_Mpi2_Program_MarkerArray = 1; // Use
#else 
static const int Use_Mpi2_Program_MarkerArray = 0; // not use
#endif
    
/**
 * ǽ
 */
class ProfileMon {
private:
    ProfileMon(const ProfileMon& rhs);
    const ProfileMon &operator=(const ProfileMon& rhs);
    
private:
    // singleton method
    static ProfileMon *PProfMon;
public:
    static ProfileMon *getInstance() 
    {
	if(PProfMon == NULL)
	{
	    PProfMon = new ProfileMon();
	}
	return(PProfMon);
    }

public:
    /*
     * stage 
     */
    enum ProfileMon_stage {
	ProfSt_radiosity,
	ProfSt_prepareShading,
	ProfSt_raytracing,
	ProfSt_Else
    };

    /*
     * event  
     */
    enum ProfileMon_event {
	ProfEv_genHemiCube,	// radiosity
	ProfEv_genRays,		// ray-tracing
	ProfEv_StopRayGen,	// global load balancing
	ProfEv_ResumeRayGen,	// 
	ProfEv_ProcRadRays,	// processing radiosity		*
	ProfEv_ProcRayRays,	// processing ray-tracing
	ProfEv_SendingWait,	// send waiting only
	ProfEv_Idle,		// Idle

	ProfEv_Send,		
	ProfEv_Recv,		
	ProfEv_ELSE
    };

private:
    int 	mStage;
    int 	mLastEvent;
    int 	mThisRank;
    // tracefile γƹԤΥإå (stdout λ¾ξʬΥ˻)
    myString	mOutHeadder;	
    FILE *	mpOutFileFp;    
    double 	mStartTime;

    myString *  mpStageNameTab;
    myString *  mpEventNameTab;

private:
    double getElapsedTime() const
    {
	double etime = 
	    ElapsedTimerClass::getNowRawTime() - this->mStartTime;
	return(etime);
    }
    
public:
    /**
     * constructor
     */
    ProfileMon()
    {
	this->mStage		= ProfSt_Else;
	this->mLastEvent	= ProfEv_ELSE;
	this->mThisRank		= -1;
	this->mpOutFileFp	= NULL;
	this->mStartTime	= 0.0;

	this->mpStageNameTab	= new myString[ProfSt_Else];
	this->mpStageNameTab[ProfSt_radiosity] = "radiosity";
	this->mpStageNameTab[ProfSt_prepareShading] = "prepareshade";
	this->mpStageNameTab[ProfSt_raytracing] = "raytrace";
	
	this->mpEventNameTab	= new myString[ProfEv_ELSE];
	this->mpEventNameTab[ProfEv_genHemiCube] = "gh";
	this->mpEventNameTab[ProfEv_genRays]	 = "gr";
	this->mpEventNameTab[ProfEv_StopRayGen]	 = "sr";
	this->mpEventNameTab[ProfEv_ResumeRayGen]= "rr";
	this->mpEventNameTab[ProfEv_ProcRadRays] = "rad";
	this->mpEventNameTab[ProfEv_ProcRayRays] = "ray";
	this->mpEventNameTab[ProfEv_SendingWait] = "sw";
	this->mpEventNameTab[ProfEv_Idle]	 = "id";
	this->mpEventNameTab[ProfEv_Send]	 = "s";		
	this->mpEventNameTab[ProfEv_Recv]	 = "r";
    }

    /**
     * initialize 
     *	@param myrank 		 trace data ϤץΥ
     *	@param outfilebasename 	trace file  basename
     * 	       "trace file 򼨤Ƭ + outfilebasename + rank ֹ" 
     * 	        trace file ̾ˤʤ롥`-' 뤤 
     *	       `(stdout)' ξˤϽϤ stdout ˥إå()ʸղä
     * 	       ƽϤ롥
     */
    void init(int myrank, const myString &outfilebasename)
    {
	if(!Use_Profile_Monitor)
	{
	    return;		// not use profile monitor
	}

	this->mStage		= ProfSt_Else;
	this->mLastEvent	= ProfEv_ELSE;
	this->mThisRank		= myrank;
	this->mpOutFileFp	= NULL;
	this->mStartTime	= ElapsedTimerClass::getNowRawTime();

	if(FileInfo::getInstance()->isProgramMarkerOn() == YES)	
	{
	    if(Use_IBM_Program_MarkerArray)
	    {
	    }

	    if(Use_Mpi2_Program_MarkerArray)
	    {
	    }
	}
	    
	if(FileInfo::getInstance()->isTraceFileOut() == YES)
	{
	    if((outfilebasename == "-") ||
	       (outfilebasename == "(stdout)"))
	    {
		// to standard out 
		this->mpOutFileFp = stdout;
		this->mOutHeadder = TRACE_STDOUT_HEAD; // Trace Υإå
	    }else
	    {
		char tracefilename[256];
#ifdef HAVE_SNPRINTF
		snprintf(tracefilename, 256, "trace.%s.%03d", 
			 (const char*)outfilebasename, this->mThisRank);
#else
		sprintf(tracefilename, "trace.%s.%03d", 
			(const char*)outfilebasename, this->mThisRank);
#endif
	    
		if((this->mpOutFileFp = fopen(tracefilename, "r")) != NULL)
		{
		    cout << "Error! : There is already trace file [" << 
			tracefilename << "], exit." << endl;
		    fclose(this->mpOutFileFp);
		    exit(1);
		}

		if((this->mpOutFileFp = fopen(tracefilename, "w")) == NULL)
		{
		    cout << "Error! : can not open trace file [" << 
			tracefilename << "], exit." << endl;
		    fclose(this->mpOutFileFp);
		    exit(1);
		}
		// եؽϤˤϾʬΥΤǥإå
		this->mOutHeadder = "";
	    }
	    fprintf(this->mpOutFileFp, 
		    "%s%s Trace file : rank %d : starttime %g\n",
		    (const char*)this->mOutHeadder, VERBOSEHEAD, 
		    this->mThisRank, this->mStartTime);
	}
    }

    /*------------------------------------------------------------
     * ʲϳ event νϤԤݤ˸Ƥ init ˸ƤФƤʤ
     * Ϥʤʤ
     *-----------------------------------------------------------*/
    /**
     * stage  start  call 뤳ȤǤϿ롥
     */
    void startStage(int stage)
    {
	if(!Use_Profile_Monitor)
	{
	    return;		// not use profile monitor
	}
	
	this->mStage = stage;
	
	if(FileInfo::getInstance()->isProgramMarkerOn() == YES)	
	{
	    if(Use_IBM_Program_MarkerArray)
	    {
	    }
	    
	    if(Use_Mpi2_Program_MarkerArray)
	    {
	    }
	}
	
	if(FileInfo::getInstance()->isTraceFileOut() == YES)
	{
	    // startstage 
	    double etime = 
		ElapsedTimerClass::getNowRawTime() - this->mStartTime;
	    fprintf(this->mpOutFileFp, 
		    "%sstartstage %s R %d T %g\n",
		    (const char*)this->mOutHeadder, 
		    (const char*)this->mpStageNameTab[this->mStage],
		    this->mThisRank, etime);
	}
	this->mLastEvent = ProfEv_ELSE;
    }

    /**
     * stage  end   call 뤳ȤǵϿԤ
     */
    void endStage(int stage)
    {
	if(!Use_Profile_Monitor)
	{
	    return;		// not use profile monitor
	}

	if(this->mStage != stage)
	{
	    cout << "Error! : ProfileMon : Stage " << stage << 
		" is not start, but done." << endl;
	    exit(1);
	}
	
	if(FileInfo::getInstance()->isProgramMarkerOn() == YES)	
	{
	    if(Use_IBM_Program_MarkerArray)
	    {
	    }

	    if(Use_Mpi2_Program_MarkerArray)
	    {
	    }
	}

	if(FileInfo::getInstance()->isTraceFileOut() == YES)
	{
	    // endstage R (rank) T (elapsed time)
	    fprintf(this->mpOutFileFp, 
		    "%sendstage %s R %d T %g\n",
		    (const char*)this->mOutHeadder, 
		    (const char*)this->mpStageNameTab[this->mStage],
		    this->mThisRank, this->getElapsedTime());
	}
	this->mLastEvent = ProfEv_ELSE;
    }

    /** 
     * event ̤εϿ
     * 	Ĥ̤ʵϿɬפʤ (genHemicube, procRays, send,
     * 	recv)  event εϿ
     */
    void occurEv(int evType)
    {
	if(!Use_Profile_Monitor)
	{
	    return;		// not use profile monitor
	}

	if(FileInfo::getInstance()->isProgramMarkerOn() == YES)
	{
	    if(Use_IBM_Program_MarkerArray)
	    {
		/*
		    pm_change_stat(PC_proc_ray_e); 
		*/
	    }

	    if(Use_Mpi2_Program_MarkerArray)
	    {
	    }
	}
	
	if(FileInfo::getInstance()->isTraceFileOut() == YES)
	{
	    switch(evType)
	    {
	    case ProfEv_genHemiCube:
	    case ProfEv_genRays:	    
	    case ProfEv_StopRayGen:
	    case ProfEv_ResumeRayGen:
	    case ProfEv_SendingWait:
	    case ProfEv_Idle:
	    case ProfEv_ProcRadRays:
	    case ProfEv_ProcRayRays:
		if(this->mLastEvent != evType)
		{
		    // (event name) R (rank) T (elapsed time)
		    fprintf(this->mpOutFileFp, "%s%s R %d T %g\n", 
			    (const char*)this->mOutHeadder, 
			    (const char*)this->mpEventNameTab[evType],
			    this->mThisRank,
			    this->getElapsedTime());
		}
		this->mLastEvent = evType;
		break;

	    case ProfEv_Send:
	    case ProfEv_Recv:	    
		cout << "Internal Error! : do not use occurEv for "
		    "this event. " << this->mpEventNameTab[evType] << endl;
		exit(1);
		break;

	    default:
		cout << "Internal Error! : unrecognized event" << endl;
		exit(1);
		break;
	    }
	}
    }

    /** 
     * HemiCube ξϿ
     */
    void outHemiCubeInfo(int rest_hemicubes, double restEnergy)
    {
	if(!Use_Profile_Monitor)
	{
	    return;		// not use profile monitor
	}

	if(FileInfo::getInstance()->isProgramMarkerOn() == YES)
	{
	    if(Use_IBM_Program_MarkerArray)
	    {
	    }

	    if(Use_Mpi2_Program_MarkerArray)
	    {
	    }
	}

	if(FileInfo::getInstance()->isTraceFileOut() == YES)
	{
	    // hcinfo R (rank) T (elapsed time) : (hemicubes in queue)
	    // (rest of unshotted energy)
	    fprintf(this->mpOutFileFp, "%s%s R %d T %g : %d %g\n", 
		    (const char*)this->mOutHeadder, "hcinfo",
		    this->mThisRank,   		    this->getElapsedTime(),
		    rest_hemicubes, 		    restEnergy);
	}
    }

    /** 
     * ray εϿ
     */
    /*
	ϿˤƤߤʤȤ狼ʤ
	Hemicube ξϿΤ褦ȯ ray ο count 
	뤫?    
    void genRays(int procRays)
    {
	if(!Use_Profile_Monitor)
	{
	    return;		// not use profile monitor
	}

	if(Use_IBM_Program_MarkerArray)
	{
	    
	}

	if(Use_Mpi2_Program_MarkerArray)
	{
	    // gr R (rank) T (elapsed time) #r (number of generated rays)
	    fprintf(this->mpOutFileFp, "%s%s R %d T %g #r %d\n", 
		    (const char*)this->mOutHeadder, 
		    (const char*)this->mpEventNameTab[ProfEv_genRays],
		    this->mThisRank,   this->getElapsedTime(), procRays);
	}
    }
    */

    /** 
     * send εϿ
     *	@param destRank 
     *  @param size     object 
     *  @param tag      򼨤 tag
     */
    void send(int destRank, int size, int tag)
    {
	if(!Use_Profile_Monitor)
	{
	    return;		// not use profile monitor
	}

	if(FileInfo::getInstance()->isProgramMarkerOn() == YES)
	{
	    if(Use_IBM_Program_MarkerArray)
	    {
	    }

	    if(Use_Mpi2_Program_MarkerArray)
	    {
	    }
	}

	if(FileInfo::getInstance()->isTraceFileOut() == YES)
	{
	    // s R (rank) T (elapsed time) : (destination rank) (size) (tag)
	    fprintf(this->mpOutFileFp, "%s%s R %d T %g : %d %d %d\n", 
		    (const char*)this->mOutHeadder, 
		    (const char*)this->mpEventNameTab[ProfEv_Send],
		    this->mThisRank,   this->getElapsedTime(), 
		    destRank, size, tag);
	}
	this->mLastEvent = ProfEv_Send;
    }

    /** 
     * recv εϿ
     *	@param srcRank 
     *  @param size    object 
     *  @param tag     򼨤 tag
     */
    void recv(int srcRank, int size, int tag)
    {
	if(!Use_Profile_Monitor)
	{
	    return;		// not use profile monitor
	}

	if(FileInfo::getInstance()->isProgramMarkerOn() == YES)
	{
	    if(Use_IBM_Program_MarkerArray)
	    {
		
	    }

	    if(Use_Mpi2_Program_MarkerArray)
	    {
		
	    }
	}

	if(FileInfo::getInstance()->isTraceFileOut() == YES)
	{
	    // r R (rank) T (elapsed time) : (source rank) (size) (tag)
	    fprintf(this->mpOutFileFp, "%s%s R %d T %g : %d %d %d\n", 
		    (const char*)this->mOutHeadder, 
		    (const char*)this->mpEventNameTab[ProfEv_Recv],
		    this->mThisRank,   this->getElapsedTime(), 
		    srcRank, size, tag);
	}
	this->mLastEvent = ProfEv_Recv;
    }
};

// singleton pattern
#ifdef THIS_IS_GLOBAL_DEFINE_FILE
ProfileMon *ProfileMon::PProfMon = NULL;	
#endif				// #ifdef THIS_IS_GLOBAL_DEFINE_FILE

#endif				// _ProfileMon_h







