//------------------------------------------------------------
// $Id: ray.cc,v 1.1.1.1 1999/11/25 06:26:12 yamauchi Exp $
//------------------------------------------------------------
//	ray
//	1996/7/21 
//	Copyright (C) 1993--1998 YAMAUCHI Hitoshi  
//	
//------------------------------------------------------------
#include <assert.h>
#include <iostream.h>
#include "ray.h"

#ifdef TRACE_RAY
static int RayID = 0;

void Ray::
setNextID(void)
{ 
    this->mID = RayID;	//----- debug
    RayID++;
}
#endif

//----------------------------------------
// constructor
//----------------------------------------
Ray::
Ray(void)
{
    //     this->mStartPos;
    //     this->mDir;
#ifndef NDEBUG
    this->mColor.set(   -1.0, -2.0, -3.0); 
    this->mRefCoeff.set(-1.0, -4.0, -9.0); 
#endif
    this->mpRadObj	= NULL;
    // this->mHitPoint;
    this->mSpcPos.set(-1, -2, -3);
    this->mScreenX	= -1;
    this->mScreenY	= -1;
    this->mReflectedNum	= 0;
    
    //     this->mSpcDInfo;
    //     this->mVTinfo;	
#ifdef TRACE_RAY
    this->setNextID();
    // this->mOrgPEID = IllegalPEID;
    //    cout << "#R n0 " << this->mID << endl;
#endif
    // #ifdef COUNT_SKIP_SPACE
    //     this->mNumOfSkipSpaces 	= 0;
    //     this->mNumOfTraverseSpaces 	= 0;    
    // #endif
    //tokunaga changed:
    this->mSide         = OUT;
}

Ray::
Ray(const Ray& rhs)
{
    assert((rhs.mScreenX >= -2) && (rhs.mScreenY >= -2));

    this->mStartPos	= rhs.mStartPos;
    this->mDir		= rhs.mDir;
    this->mColor	= rhs.mColor;
    this->mRefCoeff	= rhs.mRefCoeff;
    this->mpRadObj	= rhs.mpRadObj;
    this->mHitPoint	= rhs.mHitPoint;
    this->mSpcPos	= rhs.mSpcPos;
    this->mScreenX	= rhs.mScreenX;
    this->mScreenY	= rhs.mScreenY;
    this->mReflectedNum	= rhs.mReflectedNum;

    this->mSpcDInfo	= rhs.mSpcDInfo;
    this->mVTinfo	= rhs.mVTinfo;    
#ifdef TRACE_RAY
    this->mID		= rhs.mID; // operator= inherit ID
    // this->mOrgPEID	= rhs.mOrgPEID;
    // cout << "#R n1 " << this->mID << endl;
#endif
    // #ifdef COUNT_SKIP_SPACE
    //     this->mNumOfSkipSpaces 	= rhs.mNumOfSkipSpaces;
    //     this->mNumOfTraverseSpaces 	= rhs.mNumOfTraverseSpaces;
    // #endif
    //tokunaga changed:
    this->mSide         = rhs.mSide;
}

//----------
// constructor for ray tracing ( object ʤɤϤʤ)
//----------
Ray::
Ray(int screenX, int screenY, 
    const Vector3& viewPoint, const Vector3& destVecFromVp, 
    const IntVec3& spcpos,    const double distRaysFactor
    //, const IntVec3& PEID
    )
{
    assert((screenX >= 0) && (screenY >= 0));

    this->mStartPos 	= viewPoint; 
    this->mDir	    	= destVecFromVp;	
    assert(fabs(this->mDir.getNorm()) > 1e-5);    // this->mDir.VNrmlize();
    this->mColor.set(   0.0, 0.0, 0.0);	// ǽ鿧Ϥʤ

    // ray-tracing ǤϸΨ 1 ˽뤬 distributed
    // ray-tracing Τᡢ ray δͿ 1 꾮ʤ롣 
    // distributed ΤνŤߤ򤫤ƽ
    assert((0.0 < distRaysFactor) && (distRaysFactor <= 1.0));
    this->mRefCoeff.set(distRaysFactor, distRaysFactor, distRaysFactor);
    
    this->mpRadObj 	= NULL;	
    // this->mpHitObj 	= NULL;	
    this->mHitPoint.set(PlusInfty, PlusInfty, PlusInfty);
    this->mSpcPos  	= spcpos; 
    this->mScreenX	= screenX;
    this->mScreenY	= screenY;
    this->mReflectedNum	= 0;	// 줿Ф

    this->mSpcDInfo.invalidateInit();
    this->mVTinfo.invalidateInit();

#ifdef TRACE_RAY
    this->setNextID();
    //    this->mOrgPEID 	= PEID;
    // cout << "#R n2 " << this->mID << endl;
#endif
// #ifdef COUNT_SKIP_SPACE
//     this->mNumOfSkipSpaces 	= 0;
//     this->mNumOfTraverseSpaces 	= 0;
// #endif
    //tokunaga changed:
    this->mSide         = OUT;
}

//----------
// constructor for radiosity
//----------
Ray::
Ray(const Vector3& viewPoint, const Vector3& destVecFromVp, 
	 const Vector3& radiosityVal, 
	 objBaseClass * pGenerateObject,
	 const IntVec3& spcpos
	 // ,    const IntVec3& PEID
	 )
{
    this->mStartPos 	= viewPoint; 
    this->mDir	    	= destVecFromVp;	
    assert(fabs(this->mDir.getNorm()) > 1e-5); // this->mDir.VNrmlize();

    this->mColor	= radiosityVal;
    this->mRefCoeff.set(1.0, 1.0, 1.0);	// ȿͤǽ:Ψ1˽
    this->mpRadObj 	= pGenerateObject;	
    this->mHitPoint.set(PlusInfty, PlusInfty, PlusInfty);
    this->mSpcPos  	= spcpos; 
    this->mScreenX	= -2;	// screen ¸ߤʤ
    this->mScreenY	= -2;	// screen ¸ߤʤ
    this->mReflectedNum	= 0;	// 줿Ф

    this->mSpcDInfo.invalidateInit();
    this->mVTinfo.invalidateInit();
#ifdef TRACE_RAY
    this->setNextID();
    //this->mOrgPEID 	= PEID;
    // cout << "#R n3 " << this->mID << endl;
#endif
// #ifdef COUNT_SKIP_SPACE
//     this->mNumOfSkipSpaces 	= 0;
//     this->mNumOfTraverseSpaces	= 0;
// #endif
    //tokunaga changed:
    this->mSide         = OUT;
}

//------------------------------------------------------------
// 2 쥤νؿ 
//------------------------------------------------------------
// Raldiosity ˡ
//   @param pOldRay 		 ray ʪΤȿͤ ray 
//   @param dirR    		ȿ
//   @param newDeltaRadiosity 	ȿͤƸꤷ radiosity 
//   @param pHitObj 		򺹤ʪΤؤΥݥ(Ʊ object 
//   				 2 ŤθȽԤʤ褦)
//------------------------------
void Ray::
initForReflectionRadiosity(const Ray *const pOldRay, const Vector3& dirR, 
			   const Vector3& newDeltaRadiosity, 
			   objBaseClass *pHitObj)
{
    this->mStartPos	= pOldRay->mHitPoint; // ȿ ray λϸ
    this->mDir		= dirR;		      // ٥ȥȿ͡Ʃ
    this->mColor        = newDeltaRadiosity;  //  delta-radiosity 
    // this->mRefCoeff;	      	              // 그radiosityǤ̵ط
    assert(pHitObj != NULL); 		      // ʪΤ 2 쥤ʤΤ¸
    this->mpRadObj	= pHitObj;	      // ray ȯʪΤϵʪ
#ifndef NDEBUG
    this->mHitPoint.set(-1.0, -1.0, -1.0);    // ʪΤȤθϤޤʤ
#endif
    this->mSpcPos	= pOldRay->mSpcPos;   // ¸ߤindexƱ
    // this->mScreenX;  		      // Display Screenɸ̵ط
    // this->mScreenY;

    // ray ȿͲ PEClass::calcNsetIntensityRadiosity ˤƴä
    this->mReflectedNum	= pOldRay->mReflectedNum;	

    // 3DDDA ΤξƷ׻ʤɬפ
    this->mSpcDInfo.invalidateInit();
    this->mVTinfo.invalidateInit();

#ifdef TRACE_RAY
    this->setNextID();
    // this->mOrgPEID	= pOldRay->mOrgPEID;  // ȯ PE Ʊ
    // cout << "#R n4 " << this->mID << endl;
#endif
    // #ifdef COUNT_SKIP_SPACE
    //     this->mNumOfSkipSpaces 	= 0;
    //     this->mNumOfTraverseSpaces 	= 0;
    // #endif
    //tokunaga changed:
    this->mSide           = pOldRay->mSide;
}

//------------------------------
// Ray Tracing ˡ
//   @param pOldRay 		 ray ʪΤȿͤ ray 
//   @param dirR    		ȿ
//   @param reflectedRayCoef 	그(Ѿ軻)
//   @param pHitObj 		򺹤ʪΤؤΥݥ(Ʊ object 
//   				 2 ŤθȽԤʤ褦)
//------------------------------
void Ray::
initForReflectionRayTracing(const Ray *const pOldRay, const Vector3& dirR, 
			    const Vector3& reflectedRayCoef, 
			    objBaseClass *pHitObj)
{
    this->mStartPos	= pOldRay->getHitPos(); // ȿ ray λϸ
    this->mDir		= dirR;			// ٥ȥȿ͡Ʃ
    this->mColor.set(0.0, 0.0, 0.0);		// IC ǵٷ׻,åȤϸ
    this->mRefCoeff	= reflectedRayCoef;	// ray θ그Ϲ

    assert(pHitObj != NULL);
    this->mpRadObj	= pHitObj; 		// ray ȯʪΤϵʪ
#ifndef NDEBUG
    this->mHitPoint.set(-1.0, -1.0, -1.0); 	// ʪΤȤθϤޤʤ
#endif
    this->mSpcPos	= pOldRay->mSpcPos;   	// ¸ߤindexƱ
    this->mScreenX	= pOldRay->mScreenX; 	// Display ScreenɸƱ
    this->mScreenY	= pOldRay->mScreenY;

    // ray ȿͲäƤ꤬ʤä
    this->mReflectedNum	= (pOldRay->mReflectedNum + 1);	

    // 3DDDA ΤξƷ׻ʤɬפ
    this->mSpcDInfo.invalidateInit();
    this->mVTinfo.invalidateInit();

#ifdef TRACE_RAY
    this->setNextID();
    // this->mOrgPEID	= pOldRay->mOrgPEID;  // ȯ PE Ʊ
    // cout << "#R n4 " << this->mID << endl;
#endif
    // #ifdef COUNT_SKIP_SPACE
    //     this->mNumOfSkipSpaces 	= 0;
    //     this->mNumOfTraverseSpaces 	= 0;
    // #endif
    //tokunaga changed:
    this->mSide           = pOldRay->mSide;
}			

//------------------------------
// constructor for radiosity ( object ʤɤ)
//------------------------------
// Ray::
// Ray(const Vector3& viewPoint, const Vector3& destVecFromVp, 
// 	 const IntVec3&PEID);

//------------------------------
// destructor
//------------------------------
Ray::
~Ray(void)
{
    // cout << "#R d " << this->mID << endl;
#ifndef NDEBUG
    Vector3 destructedVector(-1.0, -2.0, -3.0);
    IntVec3 destructedIntVec(-1, -2, -3);
    this->mStartPos	= destructedVector;
    this->mDir		= destructedVector;
    this->mColor	= destructedVector;
    this->mRefCoeff	= destructedVector;
    this->mpRadObj	= NULL;
    this->mSpcPos	= destructedIntVec;
    this->mScreenX	= -3;
    this->mScreenY	= -3;
#endif
}

//------------------------------
// operator=
//------------------------------
// const Ray& Ray::
void Ray::
operator=(const Ray& rhs)
{
    if(this != &rhs)
    {
	this->mStartPos		= rhs.mStartPos;
	this->mDir		= rhs.mDir;
	this->mColor		= rhs.mColor;
	this->mRefCoeff		= rhs.mRefCoeff;
	this->mpRadObj		= rhs.mpRadObj;
	this->mHitPoint		= rhs.mHitPoint;
	this->mSpcPos		= rhs.mSpcPos;
	this->mScreenX		= rhs.mScreenX;
	this->mScreenY		= rhs.mScreenY;
	this->mReflectedNum	= rhs.mReflectedNum;

	this->mSpcDInfo		= rhs.mSpcDInfo;
	this->mVTinfo		= rhs.mVTinfo;
#ifdef TRACE_RAY
	this->mID		= rhs.mID;
	//	this->mOrgPEID		= rhs.mOrgPEID;
#endif
// #ifdef COUNT_SKIP_SPACE
// 	this->mNumOfSkipSpaces 	= rhs.mNumOfSkipSpaces;
// 	this->mNumOfTraverseSpaces	= rhs.mNumOfTraverseSpaces;
// #endif
	//tokunaga changed:
	this->mSide             = rhs.mSide;
    }
    // return(*this);
}

//------------------------------
// ¾
//------------------------------
void Ray::
print(const char *pHead) const
{
    cout << pHead << "----- Ray -----\n";
    cout << pHead << " startpos( " << this->mStartPos
	<< " ) dir( " << this->mDir << " )\n";        
    cout << pHead << " color   (" << this->mColor << ")\n";
    cout << pHead << " reflectance coefficient (" << this->mRefCoeff << ")\n";
    cout << pHead << " ray org object: "
	<< ((this->mpRadObj != NULL) ?  "There is, " : "NO, ");
    cout << pHead << " ray in spcpos(" << this->mSpcPos << ")\n";
    cout << pHead << " voxel idx    (" << this->mVTinfo.getVolIdx() << ")\n";
    cout << pHead << ((this->mSide == IN) ? "in" : "out") << " of object\n";
#ifdef TRACE_RAY
    cout << pHead << " ray ID = " << this->getID() << endl;
#endif
    // #ifdef COUNT_SKIP_SPACE
    //     cout << pHead << " skip/total " << this->mNumOfSkipSpaces << 
    // 	"/" << this->mNumOfTraverseSpaces << endl;
    // #endif
}

void Ray::
printDDA(const char *pHead) const
{
#ifndef READTEST		// READ TEST Ǥ ddainfo νϤϤʤ
    this->mSpcDInfo.print(pHead);
    this->mVTinfo.getVolDDA().print(pHead);
#else
    cout << "Error! : Ray::printDDA() is not implemented "
	"in READTEST. : pHead = " << pHead << endl;
#endif
}

//----- Ray ֤ȥС PE id 򥻥åȤƼؤ
#ifdef TRACE_RAY
// void Ray::
// setOrgPE(const IntVec3& anOrgPEID)
// {
//     this->mOrgPEID = anOrgPEID;
// }
// 
// YorN Ray::
// isSameToOrgPE(const IntVec3& aHerePEID) const
// {
//     if(this->mOrgPEID == aHerePEID)
//     {
// 	return(YES);
//     }else
//     {
// 	return(NO);
//     }
// }
#endif
