//------------------------------------------------------------
// $Id: ray.h,v 1.2 2000/06/14 13:29:37 yamauchi Exp $
//------------------------------------------------------------
//	ray
//	1996/7/21 
//	(C) Copyright Hitoshi Yamauchi
//	
//------------------------------------------------------------
#ifndef _ray_h
#define _ray_h 1

#include <assert.h>
#include "IntVec3.h"
#include "Vector3.h"
#include "config.h"
#include "constants.h"
#include "ddaInfo.h"
#include "mpi2.h"
#include "voltravinfo.h"

class objBaseClass;

// #define TRACE_RAY 1
//------------------------------------------------------------
// ray is in or out :tokunaga
//------------------------------------------------------------
enum side_e {
    OUT,
    IN
};

//------------------------------------------------------------
// ray
//------------------------------------------------------------
class Ray {
protected:
    //  ray tracing 쥤
    Ray(int screenX, int screenY,
	const Vector3& viewPoint, const Vector3& destVecFromVp, 
	const IntVec3& spcpos,    const double distRaysFactor
	//,    const IntVec3& PEID
	);
    //  radiosity 쥤
    Ray(const Vector3& viewPoint, const Vector3& destVecFromVp, 
	const Vector3& radioVal,  objBaseClass * pGenerateObject, 
	const IntVec3& spcpos	//,    const IntVec3& PEID
	);

    // ȿ delta-radiosity ray 
    void initForReflectionRadiosity(const Ray *const pOldRay, 
				    const Vector3& reflectedRayDir,
				    const Vector3& newDeltaRadiosity,
				    objBaseClass *pHitObj);
    
    // ȿ ray-tracing ray 
    void initForReflectionRayTracing(const Ray *const pOldRay, 
				     const Vector3& reflectedRayDir,
				     const Vector3& reflectedRayCoef,
				     objBaseClass *pHitObj);
protected:
    // ray ΤΤξ
    Vector3 	  mStartPos;	// ray ȯ
    Vector3 	  mDir;		// ray ٥ȥ
    Vector3 	  mColor;	// ray ε 
    Vector3 	  mRefCoeff;	// ray θ그 in radiosity = no mean
    objBaseClass *mpRadObj;	// ray ȯʪ (¾֤Ǥ NULL)
    Vector3 	  mHitPoint;	// ʪΤȤθ
    IntVec3 	  mSpcPos;	// ¸ߤ index
    int 	  mScreenX, mScreenY; // Display Screen ɸ
    int		  mReflectedNum;      // ray ȿͲ

    DDAInfoClass  mSpcDInfo;	//  traverse Τ DDA ݻ
    volTravInfoClass mVTinfo;	// volume traverse Τξݻ
#ifdef TRACE_RAY
    int 	  mID;			//----- debug
    // IntVec3       mOrgPEID;		// ȯ PE
#endif
//------------------------------
// ray ֤ɤ skip ξ ray ĤΤɤray 
// ͭξȤɬפǤϤʤΤ PE ˻뤳Ȥˤ롣 
//------------------------------
// #ifdef COUNT_SKIP_SPACE		// ray ֤ɤΰ̥åפ
//     int		  mNumOfSkipSpaces;
//     int		  mNumOfTraverseSpaces;
// #endif

public:
    Ray(void);		       // default constructor
    //  ray tracing 쥤
    static Ray* newForRayTrace(int screenX, int screenY,
			       const Vector3& viewPoint, 
			       const Vector3& destVecFromVp, 
			       const IntVec3& spcpos,    
			       const double   distRaysFactor
			       //,    const IntVec3& PEID
			       )
    {
	return(new Ray(screenX, screenY, viewPoint, destVecFromVp, 
		       spcpos,  distRaysFactor
		       //,    const IntVec3& PEID
		       ));
    }
    //  radiosity 쥤
    static Ray *newForRadiosity(const Vector3& viewPoint, 
				const Vector3& destVecFromVp, 
				const Vector3& radioVal,
				objBaseClass * pGenerateObject, 
				const IntVec3& spcpos
				//,    const IntVec3& PEID
				)
    {
	return(new Ray(viewPoint, destVecFromVp, 
		       radioVal,  pGenerateObject, spcpos
		       //,    const IntVec3& PEID
		       ));
    }
    // ȿͥ쥤 (ray-tracing)
    // pOldRay ȿʪ(=ȯʪ)ξ⤢
    static Ray *newForReflectionRayTracing(const Ray *const pOldRay, 
					   const Vector3& reflectedRayDir,
					   const Vector3& reflectedRayCoef,
					   objBaseClass *pHitObj)
    {
	Ray *pRay = new Ray();
	pRay->initForReflectionRayTracing(pOldRay, reflectedRayDir,
					  reflectedRayCoef, pHitObj);
	return(pRay);
    }
    // ȿͥ쥤 (radiosity)
    // pOldRay ȿʪ(=ȯʪ)ξ⤢
    static Ray *newForReflectionRadiosity(const Ray *const pOldRay, 
					  const Vector3& reflectedRayDir,
					  const Vector3& newDeltaRadiosity,
					  objBaseClass *pHitObj)
    {
	Ray *pRay = new Ray();
	pRay->initForReflectionRadiosity(pOldRay, reflectedRayDir,
					 newDeltaRadiosity, pHitObj);
	return(pRay);
    }
    
    Ray(const Ray& rhs);     // buff Фcopy construct
    ~Ray(void);
    // const Ray& operator=(const Ray& rhs);
    void operator=(const Ray& rhs);
    void setPosDir(const Vector3& aStartPos, const Vector3& aDir)
    {
	this->mStartPos = aStartPos;
	this->mDir 	= aDir;
    }
    // void setPos(const Vector3& aPos)    { this->mStartPos = aPos;	}
    Vector3 getPos(void) const		   
    {
	return(this->mStartPos); 	
    }
    void    setDir(const Vector3& aDir)    
    {
	this->mDir = aDir;		
    }
    Vector3 getDir(void) const	   	   
    {
	return(this->mDir);	
    }
    void    setColor(const Vector3& aCol)  
    {
	this->mColor = aCol;	
    }
    Vector3 getColor(void) const	   
    {
	return(this->mColor);	
    }
    void    setRefCoeff(const Vector3& aRef)
    {
	this->mRefCoeff = aRef;	
    }
    Vector3 getRefCoeff(void) const	   
    { 
	return(this->mRefCoeff);	
    }
    void    setRadObj(objBaseClass * apObj)
    {
	this->mpRadObj = apObj;	
    }
    objBaseClass *getRadObj(void) const    
    {
	return(this->mpRadObj);	
    }
    void    setHitPos(const Vector3& aHPos)
    {
	this->mHitPoint = aHPos;	
    }
    Vector3 getHitPos(void) const	   
    {
	return(this->mHitPoint);	
    }
    void    setSpcPos(const IntVec3& aSpc) 
    {
	this->mSpcPos  = aSpc;	}   
    
    const IntVec3 getSpcPos(void) const    
    {
	return(this->mSpcPos);    	}   
    
    int     getScreenPosX() const	
    {
	return(this->mScreenX);
    }
    int     getScreenPosY() const
    {
	return(this->mScreenY);
    }
    int	 getReflectedNum(void) const 	   
    {
	return(this->mReflectedNum); 
    }
    void incReflectedNum(void) 		   
    {
	this->mReflectedNum++;	
    }

    // DDA ط 
    //-----  traverse  DDA
    void   setSpcDDAInfo(const DDAInfoClass& dinfo) 
    {
	this->mSpcDInfo = dinfo; 
    }
    DDAInfoClass getSpcDDAInfo(void) const 
    {
	return(this->mSpcDInfo);	 
    }
    DDAInfoClass* getSpcDDAInfoPtr(void)   
    {
	return(&(this->mSpcDInfo)); 
    }
    void   setSpcDinfoIllegalDir(void)	
    { 
	this->mSpcDInfo.setGoingDir(DIR_ILLEGAL); 
    }

    //-----  voxel traverse  DDA
    void    invalidateVolData(void) 
    {
	this->mVTinfo.invalidateInit();	 
    }
    // volume DDA
    YorN   isVolDDAinited(void) const 	
    {
	// improve 󥹥ȥ饯ȤΤǸΨ
	return((this->mVTinfo.getVolDDA()).isDDAinited());
    }
    void   setVolDDAInfo(const DDAInfoClass& dinfo) 
    {
	this->mVTinfo.setVolDDA(dinfo);
    }
    DDAInfoClass getVolDDAInfo(void) const 
    {
	return(this->mVTinfo.getVolDDA());
    }
    //    volume ID
    void   setVolID(int vid) 		
    {
	this->mVTinfo.setVolID(vid);	
    }
    int    getVolID(void) const		
    {
	return(this->mVTinfo.getVolID()); 
    }
    //    volume index
    void    setVoxelIdx(const IntVec3& voxelIdx) 
    {
	this->mVTinfo.setVolIdx(voxelIdx);
    }
    IntVec3 getVoxelIdx(void) const 
    {
	return(this->mVTinfo.getVolIdx()); 
    }
    void    setAccumOpacity(double accumOpacity)
    {
	this->mVTinfo.setAccumOpacity(accumOpacity);
    }
    double getAccumOpacity(void)const
    {
	return(this->mVTinfo.getAccumOpacity());
    }
				       
    //----- ¾
    void print(const char *pLineHead) const;
    void printDDA(const char *pLineHead) const;
    int  isOK(void) const 
    {
	return((this->mSpcPos.get(Vector3::X) >= 0) ? 1 : 0);
    }
    

#ifdef TRACE_RAY
    int  getID(void) const 
    {
	return(this->mID); 
    }
protected:
    void setNextID(void);
    //void setOrgPE(const IntVec3& aOrgPEID); // ľˤ PE
    //YorN isSameToOrgPE(const IntVec3& aHerePEID) const; // 
#endif
// #ifdef COUNT_SKIP_SPACE
// public:
//     void incSkipSpaceNum(void)		{ this->mNumOfSkipSpaces++; }
//     int  getSkipSpaceNum(void) const	{ return(this->mNumOfSkipSpaces);}
//     // number of traverse spaces  updateSpc ä
// #endif    

    // tokunaga changed:
    // Ʃ:쥤ΰ֤Ƚ    
protected:
     side_e mSide;

public:
    // ʪ⤫
    void changeSide(void) 
    {	
	this->mSide = (this->mSide == OUT) ? IN : OUT;
    }
    YorN  isInside(void)  const {
	if(this->mSide == IN)
	{
	    return(YES); 
	}
	else
	{
	    return(NO);
	}
    }

public:
    // ֤
    static communication_tags_e getTAG(void) 
    {
	return(TAG_RAY_e);	
    }
};

#endif				// #ifndef _ray_h 
