//------------------------------------------------------------
// $Id: volumeRayCast.cc,v 1.2.2.1 2000/06/14 15:00:10 yamauchi Exp $
//------------------------------------------------------------
//	volumeClass : ray casting 
//	1998/2/10
//	Copyright (C) 1996--1998 YAMAUCHI Hitoshi  
//------------------------------------------------------------
#include <stdlib.h>
#include <math.h>
#include "SLList.h"

#include "searchNextSpcDDA.h"
#include "IntVec3.h"
#include "volume.h"
#include "volumeFuncs.h"

#ifndef ONLY_VOLREN
#include "mpi2_funcs.h"
#else
// external functions 
extern YorN isPosInBox(const Vector3& pos, 
		       const Vector3& minpos, 
		       const Vector3& maxpos);
extern isPosInBoxWithRes(const Vector3& pos, 
			 const Vector3& minpos, 
			 const Vector3& maxpos,
			 double res);
extern DIR_3D isHitBoundingBox(const Vector3& minBox,    
			       const Vector3& maxBox, 
			       const Vector3& rayOrigin,
			       const Vector3& rayDir,
			       Vector3 *pHitPoint);
#endif


//------------------------------------------------------------
// RP Block νΤvolume ȤߺѤ׻ʬ
//	ray ȸ򺹤뤳ζ֤¸ߤ volume ᡢΥζᤤ
//	¤٤
//	call from processingOneRayTrace
//
// 	mHitVIdxCand 򺹥ܥ index ȤʤäƤΤϡ
// 	֤¸ߤȤ롢뤤Ϥζ֤ voxel  ray 
// 	Ѥ׻ΤɬפȤ voxel Ȥθ򺹤Ǥ롣
//
//	-----+
//	--+---+
//        |  ||<-- voxel 
//      --+---+
//           |<--- ֶ
//	-----+
//
//  	ޤΤ褦 voxel ζ֤ζϰפʤȤ¿
//  	ξˤ٤֤ voxel ξäƤơǴ
//       voxel η׻򤷤ƤޤäƤ뤫⤷ʤ³
//      ׻뤿 ray  mVoxelIdx Ĥvolume ˸򺹤뤫
//      ɤϤζ֤ȽǤɬפ롣Ǥζ֤λ 
//      volume Ȥθǽ mHitVIdxCand Ȥˤʤ櫓Ǥ
//      롣
//------------------------------------------------------------
// volume ȸ򺹾ݻ
//------------------------------------------------------------
class VolumeClass;

class VolDistClass {
public:
    double       mDistance;	// εΥ if(dist < 0){no hit}
    VolumeClass *mpVol;		// volume
    IntVec3 	 mHitVIdxCand;	// 򺹥ܥ index 
    double 	 mMaxDistance;	// volume ̲ǤΥ()
				// ʪΤȸ򺹤ƤˤϤޤ
public:
    VolDistClass(void) : mDistance(-1.0), mpVol(NULL)
    {
	this->mHitVIdxCand.set(-5, -6, -7); // 
    }
};

//------------------------------------------------------------
// implementation
//------------------------------------------------------------
// pos ϡ stride Ƕڤ줿 org ϤޤΤɤ index 
// ˤ뤫
//                 
//	|----|----|----|----|----|----|----|
//     org          ^pos      -->|----|<--
//                                stride
//------------------------------
inline int getPosIdx(double org_pos, double stride, double position)
{
    // org_pos + stride * index $\leq$ position 
    //		->
    // index = \lfloor \frac{position - org_pos}{stride} \rfloor
    return((int)(floor((position - org_pos) / stride)));
}

//------------------------------------------------------------
// class methods
//------------------------------------------------------------
// pos Ϥ volume ΤɤǤ¸ߤ뤫
//  volume attr 
//------------------------------
IntVec3 VolumeClass::
getGlobalVolumeIndexFromPos(const Vector3& pos)
{
    // 0.0001  顼ĺ
    assert(isPosInBoxWithRes(pos, this->mSubGrid.getMinPos(),
			     this->mSubGrid.getMaxPos(), 0.0001) == YES); 
    int i;
    int ret[3];
    for(i = 0; i < 3; i++)
    {
	ret[i] = getPosIdx((this->mGlobalGrid.getMinPos()).get(i), 
			   this->mGlobalGrid.getVoxelEdgeLen(),
			   pos.get(i));
	assert(this->mGrid.getMinIdx(i) <= ret[i]);
	assert(ret[i] <= this->mGrid.getMaxIdx(i));

	// volume ξˤϺʬäˤϰ⤬ volume
	// index Ȥʤ롣
	if(ret[i] == this->mGrid.getMaxIdx(i))
	{
	    ret[i]--;
	    assert(this->mGrid.getMinIdx(i) <= ret[i]);
	}
    }

    return(IntVec3(ret[0], ret[1], ret[2]));
}


//------------------------------
//  volume  ray 򺹤뤫Υƥ
//	castRay ƤФ줿Ǵ volume  ray 򺹤뤫ϥƥ
//	ȺѤߤǤ롣
//------------------------------
static YorN checkRayIsInVolume(const Vector3& subVolMinPos, 
			       const Vector3& subVolMaxPos,
			       const Vector3& rayPos,
			       const Vector3& rayDir)
{
    // ray  volume Ⱥǽ˸򺹤
    Vector3 intersectPos;	
    
    // ray λ volume γ椫
    if(isPosInBox(rayPos, subVolMinPos, subVolMaxPos) == YES) 
    {
	// ray λ volume ¸ߤ => 򺹤
	return(YES);	    	
    }else
    {
	// ray λ volume γˤ: ray  volume ȸ򺹤뤫?
	if(isHitBoundingBox(subVolMinPos, subVolMaxPos, rayPos, rayDir,
			    &intersectPos) == DIR_ILLEGAL)
	{
	    // 򺹤ʤäʤˤ⤷ʤ֤롣
	    return(NO);
	}else
	{
	    // 򺹤
	    return(YES);	    
	}
    }
}

//------------------------------------------------------------
// cast ray 
//   @param *pRay 		: ray
//   @param limitDistance	: ɤޤפ뤫
//   @param *pRetColor		: ׻̤ο
//
//   maxDistance ¸ߤΤϡvolume ʪΤ¸ߤ
//   ˤʪΤ ray ֤Ĥäʬߤ뤿ᡣ
//	return : 
//	    ray  volume ȸ򺹤(׻)?  YES
//
// DDA unit (ץå) DDA info (ǡ)ΰ㤤
//------------------------------------------------------------
YorN VolumeClass::
castRay(Ray *pRay, const VolDistClass* const pVolDist, Vector3 *pRetColor)
{
    // ray Ȥ volume ϸ򺹤Ϥ
    assert(checkRayIsInVolume(this->mSubGrid.getMinPos(), 
			      this->mSubGrid.getMaxPos(),
			      pRay->getPos(),	   pRay->getDir()) == YES);

    //------------------------------------------------------------
    // DDAinfo Ƥˤ⤫餺volume ID ۤʤ
    // ˤ volume ŤʤäƤ뤫 DDAinfo ̤ѤˤƤʤ
    // Τ褦ʾ check Ƥ
    // DDAinfo ˽ƤΤʤФ ray ٤ζ֤Ǥʬ䤷
    // Ʊ volume  trace ƤϤ
    //
    // ⤷volDDA ƤСvolID >= 0   ID 
    // 뤤ϽƤʤ
    //------------------------------------------------------------
    assert(((pRay->isVolDDAinited() == YES) && 
	    ((pRay->getVolID() >= 0) && (this->getID() == pRay->getVolID())))
	   || (pRay->isVolDDAinited() == NO));
    
    if(pRay->getAccumOpacity() > 1.0)
    {
	// Ʃ٤ 1.0 ۤƤΤǷ׻ɬפϤʤ
	// Ǥڤäˤϸ DDA η׻ʤΤ̵
	// Ǥ
	// Ǥڤʤˤ accumOpacity  1 ۤƤޤ
	// 硤
	// 	double 	hereOpacity = c * (1.0 - accumOpacity);
	// Ƿ׻ƤΤ opacity ˤʤ = color ˤʤ뤿
	// Ჿ餫ν֤ɬ
	return(NO);
    }

    //------------------------------------------------------------
    //  volume  traverse 뤿 DD ˥åȤ
    // DD Unit ϳ PE, volume Ȥ˾֤ۤʤΤǽɬפ
    // 롣
    //  DDAinfo ϰٽƱ volume, ʬʤн
    // ɬפϤʤ
    //------------------------------------------------------------
    searchNextSpcDDA nssDDAunit;

    int spcMinIdx[3], spcMaxIdx[3];
    for(int i = 0; i < 3; i++)
    {
	// grid ͭϰϤ [spcMinIdx, spcMaxIdx] ʤΤDDAUnit 
	//  [min, max) ȤƷ׻Τ Max  1 䤷Ƥ
	spcMinIdx[i] = (this->mSubGrid.getMinIdx()).get(i);
	spcMaxIdx[i] = (this->mSubGrid.getMaxIdx()).get(i) + 1; 
    }
    
    nssDDAunit.initWithoutObjInSpcMap(this->mGlobalGrid.getMinPos(),
				      this->mGlobalGrid.getVoxelEdgeLen(),
				      spcMinIdx, spcMaxIdx,
				      NoSkip_rayprop,
				      MultiStage_network);

    //------------------------------------------------------------
    // Υη׻
    //------------------------------------------------------------
    double nowDistFromRayOrg = pVolDist->mDistance;
    const double deltaDist   = ((pRay->getVolDDAInfo()).getDDAVec()).getNorm();
    

    //------------------------------------------------------------
    // η׻
    //------------------------------------------------------------
    Vector3 accumColor(0.0, 0.0, 0.0); // ٤ζ֤ͤФʤ󤾡
    double  accumOpacity = pRay->getAccumOpacity();

    // IMPROVE  pointer ˤľ ray  DDAinfo,
    // spcpos ѹƤɤ 
    IntVec3 hitGridIdx	  = pRay->getVoxelIdx();
    DDAInfoClass volDinfo = pRay->getVolDDAInfo();

    int hitVGridIdxInt[3];

    // grid  RGB ̾ [0, 255]  short intdepth 
    // 롥 
    double parDepth = 1.0 / (double)(this->mGlobalGrid.getRGBDepth());
    while(nssDDAunit.isValidASpcpos(hitGridIdx) == YES)
    {
	// äɤ֤ˤ

	//------------------------------------------------------------
	// voxel η׻ for uchi
	//------------------------------
	hitGridIdx.getAsArray(hitVGridIdxInt);
	Vector3 col
	    = (this->getGlobalEffiGridIdxColor(hitVGridIdxInt)).getRGBColor();
	assert((col.get(Vector3::X) >= 0.0) && (col.get(Vector3::Y) >= 0.0) &&
	       (col.get(Vector3::Z) >= 0.0));

	// color vector ʿĹ / depth  opacity Ȥ : ʿ
	// ȤȤΤϤ (3.0 = RGB 3 Ĥʿ)
	// ǤϤʿѤƩ٤Ȥ롥
	double colorave = (col.get(Vector3::X) + col.get(Vector3::Y) +
			   col.get(Vector3::Z)) / 3.0;
	double hereOpacity = ((colorave * parDepth) * (1.0 - accumOpacity));
	accumColor 	   += (hereOpacity * col);
	accumOpacity       += hereOpacity;
	if(accumOpacity >= 1.0)
	{
	    pRay->setAccumOpacity(accumOpacity);
	    cout << "P";	// Ʃ٤ overflow ɽ
	    break;
	}
	//------------------------------
	// voxel η׻Ϥޤ
	//------------------------------------------------------------	

	nssDDAunit.forceMoveRayToNextSpace(&volDinfo,
					   pRay->getPos(), pRay->getDir(),
					   &hitGridIdx);
	// polygon ʤɤȤθ򺹤ˤ Max ʾǷ׻Ǥڤ롥
	nowDistFromRayOrg += deltaDist;
	if(nowDistFromRayOrg > pVolDist->mMaxDistance)
	{
	    break;
	}
    }

    // assert(accumColor >= 0.0);
    assert(accumOpacity >= 0.0);

    (*pRetColor) = accumColor;
    pRay->setVolDDAInfo(volDinfo);
    pRay->setVoxelIdx(hitGridIdx);
    pRay->setAccumOpacity(accumOpacity);
    
    return(YES);
}

//------------------------------------------------------------
// ray  volume θ
//	return : 򺹤εΥ (ξˤϸ򺹤ʤ)
//	arg:	 pIntersectPos : volume  ray   ray λ 
//		  		 volume ʤлΰ
//		 pHitVoxelIdx  : ray 򺹤 voxel  index
//------------------------------
double getDistanceWithVolume(Ray *pRay, VolumeClass *pVol,
			     Vector3  *pIntersectPos,
			     IntVec3  *pHitVoxelIdx)
{
    // ray λ volume γ椫
    if(isPosInBox(pRay->getPos(), 
		  (pVol->getEffiGrid()).getEfMinPos(), 
		  (pVol->getEffiGrid()).getEfMaxPos()) == YES) 
    {	
	// ray λ volume ¸ߤ

	//  ray ΰ֤Ȥ֤ ( voxel ζˤ
	// ⤷ʤΤ꤬뤫?)
	(*pIntersectPos) = pRay->getPos(); 
	(*pHitVoxelIdx)  = pVol->getGlobalVolumeIndexFromPos(pRay->getPos());
	return(0.0);		
    }
    else
    {
	// ray λ volume γˤ: ray 
	// volume ͭϰϤȸ򺹤뤫?
	if(isHitBoundingBox((pVol->getEffiGrid()).getEfMinPos(), 
			    (pVol->getEffiGrid()).getEfMaxPos(),
			    pRay->getPos(), pRay->getDir(),
			    pIntersectPos) == DIR_ILLEGAL)
	{
	    // 򺹤ʤäʤˤ⤷ʤ֤롣
	    return(-1.0);
	}else
	{
	    // 򺹤
	    //   (pIntersectPos  isHitBoundingBox ǥåȤ)
	    (*pHitVoxelIdx) = 
		pVol->getGlobalVolumeIndexFromPos((*pIntersectPos));
	    Vector3 d = (*pIntersectPos) - (pRay->getPos());
	    double  dist = d.getNorm();
	    // return((*pIntersectPos) -(pRay->getPos()).getNorm()) 
	    // Ƚ񤯤ȥ󥹥ȥ饯ǥȥ饯ĸƤФΤ
	    // ˤʤäƤޤΤ褦 tmp ͤȤäƤޤ
	    // Ӹᤰߤ󤬷뺧ȤΤȡǤȤޤ
	    // 1998ǯ44() 
	    assert(dist >= 0.0);
	    return(dist);
	}
    }
}

//------------------------------------------------------------
// qsort Ӵؿ
// gcc ϼνǤ̤Ƥ줿 qsort ӴؿȤ
// ʤΤŪ˥㥹ȤΤ褦񤷤ͤ 
//
// static int comparDist(const VolDistClass *i, const VolDistClass *j)
//------------------------------------------------------------
static int comparDist(const void *i, const void *j)
{
    if(((VolDistClass *)i)->mDistance > ((VolDistClass *)j)->mDistance)
    {
	return(1);
    }
    else if(((VolDistClass *)i)->mDistance < ((VolDistClass *)j)->mDistance)
    {
	return(-1);
    }else
    {
	return(0);
    }
}

//------------------------------
// voxel index ϰϤ鳰 YES
//------------------------------
static YorN isOutGlovalVolRange(VolumeClass *pVol, Ray *pRay)
{
    IntVec3 vidx = pRay->getVoxelIdx();
    for(int xyz = 0; xyz < 3; xyz++)
    {
	if((vidx.get(xyz) < 0) ||
	   (pVol->getGlobalGridResSize(xyz) <= vidx.get(xyz)))
	{
	    // volume index is out of range
	    return(YES);
	}
    }
    return(NO);
}

//------------------------------
// Ĥ volume  1 Ĥ ray θ
// 	ray οιޤ
//
// hitVIdxCand  ray ȸ򺹤 voxel index θǤ롣
// ray äƤΤray Ͻ֤Ǥ 
// volume ȤޤߺѤƤʤΤ ray Ϸ׻Ѥߤ voxel
// index ʤξˤϸǤ hitVIdxCand Ȥ
//
// @param pRay
// @param volDist ޤǤεΥΥ volume ξ
//------------------------------
static void castRayToVolume(Ray      *pRay, 	 // 쥤
			    const VolDistClass& volDist) //  volume ξ
{
    //----- 
    // DDAinfo Ƥˤ⤫餺volume ID ۤʤ
    // ˤ volume ŤʤäƤ뤫 DDAinfo ̤ѤˤƤʤ
    // Τ褦ʾ check Ƥ
    // DDAinfo ˽ƤΤʤФ ray ٤ζ֤Ǥʬ䤷
    // Ʊ volume  trace ƤϤ
    // ٤³? 
    if(pRay->isVolDDAinited() == NO)
    {
	// Ƥʤ롣
	//   volume ID ȸ voxel 򥻥å(DDA  cast н)
	assert(volDist.mpVol->checkGridIdxRangeForDEBUG(volDist.mHitVIdxCand)
	       == YES);
	pRay->setVolID(volDist.mpVol->getID()); 
	pRay->setVoxelIdx(volDist.mHitVIdxCand);
    }else
    {
	// Ƥ롣volume ID Ʊ?
	if(pRay->getVolID() != volDist.mpVol->getID())
	{
	    // volume id ۤʤ롣 => ³ǤϤʤä
	    // Ʊ֤2Ĥ volume ƤΤȤ
	    cerr << "Error! : volume [" << pRay->getVolID() << "] and ["
		<< volDist.mpVol->getID() << "] share a same space. " 
		" no volume can share same space." << endl;
	    abort();
	}else
	{
	    // Ƥ,ޤǤƱvolume ID.Τޤ³Ǥ
	    // 	    cout << pRay->getID() << " " << pRay->getVoxelIdx() << 
	    // 		" has same volume ID" << endl;
	}
    }
    
    Vector3 vcolor(0.0, 0.0, 0.0);
    if(volDist.mpVol->castRay(pRay, &volDist, &vcolor) == YES)
    {
	// FIXHERE ϸ mono
	Vector3 col = pRay->getColor();
	vcolor   += col;
	// overflow Ƚ frame buffer ˽񤯻
	pRay->setColor(vcolor);
    }
    if(isOutGlovalVolRange(volDist.mpVol, pRay) == YES)
    {
	pRay->invalidateVolData();
    }
}

//------------------------------
// Ĥζ֤ 1 Ĥ ray ̲᤹
// 	ray trace 
//	Ĥζ֤ˤʣ volume ¸ߤǽ
//
// 1. ray ȸ򺹤 volume 
// 2. ᤤ volume  ray cast Ƥ
//
// @param pRay
// @param pVolList ray ȸ򺹤ǽΤ volume Υꥹ
// @param distanceBtwRayNObj ʪΤȸ򺹤ƤˤϤεΥ
// 	  Ƥʤˤ PlusInfty 
//------------------------------
void processingOneRayTraceWithVolume(Ray *pRay, 
				     SLList < VolumeClass * > *pVolList,
				     const double distanceBtwRayNObj)
{
    assert(!(pVolList->empty())); // volume ¸ߤϤ
    
    // ray ̲᤹ volume 
    int volumeListLen = pVolList->length();
    assert(volumeListLen > 0);
    VolDistClass *pVidDist = new VolDistClass[volumeListLen];
    
    Pix pix;
    int hitVidIdx = 0;
    Vector3 intersectPos;
    IntVec3 hitVoxelIdx;
    for(pix = pVolList->first(); pix != NULL; pVolList->next(pix))
    {
	VolumeClass *pVol = (*pVolList)(pix);
	double dist = getDistanceWithVolume(pRay, pVol, 
					    &intersectPos, &hitVoxelIdx);
	//  distanceBtwRayNObj 꾮 volume 
	//  hit : ãǽΥ⵭Ͽ
	if((0.0 <= dist) && (dist < distanceBtwRayNObj))
	{  
	    // ray  volume 򺹤( volume ˴ޤޤ뤳Ȥ)
	    pVidDist[hitVidIdx].mpVol 	  	= pVol;
	    pVidDist[hitVidIdx].mDistance 	= dist;
	    pVidDist[hitVidIdx].mHitVIdxCand 	= hitVoxelIdx;	    
	    pVidDist[hitVidIdx].mMaxDistance 	= distanceBtwRayNObj;	    
	    hitVidIdx++;
	    assert(pVol->checkGridIdxRangeForDEBUG(hitVoxelIdx) == YES);
	}
    }	    

    // ray ȸ򺹤 volume ¸ߤʤʤ齪λ¿еΥ sort
    assert(hitVidIdx >= 0);
    if(hitVidIdx == 0)
    {
	return;			//  volume ¸ߤʤ
    }
    else if(hitVidIdx > 1)	
    {
	// ʣθ volume 롣Υ sort 롣
	qsort((VolDistClass *)pVidDist, hitVidIdx,
	      sizeof(VolDistClass), 	 comparDist);
	// for DEBUG
	cout << "DEBUG: sort res : processingOneRayTraceWithVolume "
		"(volID, dist) " << endl;
	for(int d = 0; d < hitVidIdx; d++)
	{
	    cout << "(" << pVidDist[d].mpVol << ", " << 
		pVidDist[d].mDistance << ") ";
	}
	cout << endl;
	// for DEBUG
    }// else => hitVidIdx == 1 ĤޤĤʤΤǥȤɬפϤʤ

    // ᤤ volume  ray  cast ray ο򹹿
    for(int i = 0; i < hitVidIdx; i++)
    {
	castRayToVolume(pRay, pVidDist[i]);
    }
}

