//------------------------------------------------------------
// $Id: NetIntUnit.cc,v 1.2.2.1 2000/06/14 15:00:10 yamauchi Exp $
//------------------------------------------------------------
//	netInt.cc Network Interface Unit implementation
//	These functions are member of PEClass.
//	1997/1/21
//	Copyright (C) 1997--1998 YAMAUCHI Hitoshi
//
//  volume rendering б Network Interface Unit  PE ľ
//  񤤤ƤʬʬΥ 
//  ĤΥФ PEClass ΥФǤ
//------------------------------------------------------------
#include "mpi2communication.h"
#include "mpi2exit.h"
#include "NetIntUnit.h"

//------------------------------------------------------------
// public functions
//------------------------------------------------------------
// constructors
//------------------------------
NetIntClass::
NetIntClass(void)
{
    this->mTotalPENum 			= -1;
    this->mSendRayThreshold 		= -1;
    this->mSendIntensityThreshold  	= -1;
    this->mThisRank			= -1;
    this->mHostRank			= -1;
    this->mpSpc2PE 			= NULL;
    this->mpSendRaysBuf			= NULL;
    this->mpSendVerIntBuf		= NULL;
}

NetIntClass::
~NetIntClass(void)
{
    if(this->mpSpc2PE != NULL)
    {
	delete this->mpSpc2PE;
	delete [] this->mpSendRaysBuf;

    }
    if(this->mpSendVerIntBuf != NULL)
    {
	delete [] this->mpSendVerIntBuf;
    }
}


//------------------------------------------------------------
// initialize network interface 
//------------------------------
// initialize :  call Ƥ
//    @param totalPENum	      	      PE 
//    @param raySendThreshold 	     ray  
//    @param intensitySendThreshold  پ FB ؤΤޤȤ
//    @param thisRank		      PE Υ(ºݤ:å)
//    @param hostRank		     host Υ : Pixel 
//    @param pProcRaysQueue	     PEǽrayݻ륭塼λ
//------------------------------
void NetIntClass::
initRayBuff(int totalPENum, int raySendThreshold, int intensitySendThreshold,
	    int thisRank,   int hostRank,
	    mySLQueue < Ray * > *pProcRaysQueue)
{
    // check args
    if((totalPENum <= 0) 	     ||	(raySendThreshold <= 0) || 
       (intensitySendThreshold <= 0) ||	(thisRank < 0)  	||
       (thisRank >= totalPENum)	     || (hostRank < 0))
    {
	cout << "Error! : illegal args at NetIntClass::initRayBuff" << endl;
	cout << "(totalPENum, raySendThreshold, intensitySendThreshold, "
	     << "thisRank, hostRank) = (" 
	     << totalPENum << ", " << raySendThreshold << ", "
	     << intensitySendThreshold << ", " << thisRank << ", " << 
	     hostRank << ")" << endl;
	commMpi2_exit(1);
    }
    if((pProcRaysQueue == NULL) || (pProcRaysQueue->length() != 0))
    {
	cout << "Error! : illegal queue at NetIntClass::initRayBuff" << endl;
	commMpi2_exit(1);
    }
    
    // init PE num and threshold
    this->mTotalPENum 	      		= totalPENum;	
    this->mSendRayThreshold  		= raySendThreshold;
    this->mSendIntensityThreshold	= intensitySendThreshold;
    this->mThisRank 			= thisRank;
    this->mHostRank 			= hostRank;
    
    // ray Хåեβ 
    if(this->mpSendRaysBuf != NULL)
    {
	delete [] this->mpSendRaysBuf; // ¸ߤƤˤϲ
	this->mpSendRaysBuf = NULL;
    }
    
    this->mpSendRaysBuf = new SendBuffer < Ray > [this->mTotalPENum];
    for(int peRank = 0; peRank < this->mTotalPENum; peRank++)
    {
	this->mpSendRaysBuf[peRank].initBuff(this->mSendRayThreshold);
    }

    this->mSendIntensitiesBuf.initBuff(this->mSendIntensityThreshold);
    
    // Хåեν
    for(int i = 0; i < MAX_RECV_BUFFNUM; i++)
    {
	// ΥХåեȼΥХåեΰ礭Ʊ
	// (礭(this->mSendRayThreshold)̤ ray ΤȤ꤬롥)
	this->mRecvRaysBuf[i].init(this->mSendRayThreshold, pProcRaysQueue);
    }
}

//------------------------------------------------------------
// ط
//------------------------------------------------------------
// ޥȯ
// ˤĤơΥ塼 object 졤塼ꤷ
// Ĺۤ׵ȯԤߤ롥
//   @return ׵ȯԤ YES
//------------------------------
YorN NetIntClass::
sendRays(Ray *pRay, int destRank)
{
    assert((0 <= destRank) && (destRank < this->mTotalPENum));
    assert(pRay != NULL);
    
    return(this->mpSendRaysBuf[destRank].
	   	enqNewedObjwithMaySend(pRay, destRank));
}


//------------------------------
// 塼 1 ܤǤ ray ¸ߤȯԤߡbuffer  lock 
// Ƥʤ¤׵ȯԤ롥 
//   @return ׵ȯԤ YES
//------------------------------
YorN NetIntClass::
flushSendRaysBuff(void)
{
    // 
    YorN isIssueSend = NO;
    for(int i = 0; i < this->mTotalPENum; i++)
    {
	if(this->mpSendRaysBuf[i].flushdone(i) == YES)
	{
	    assert(this->mThisRank != i);
	    isIssueSend = YES; // ׵ȯԤ
	}
    }
    return(isIssueSend);
}

//------------------------------------------------------------
// ط
//------------------------------------------------------------
// ޥȯ
//------------------------------
YorN NetIntClass::
recvRays(void)
{
    // λ
    // ΥХåե򸡺λʤ餽νԤ
    for(int recvBuffNum = 0; recvBuffNum < MAX_RECV_BUFFNUM; recvBuffNum++)
    {
	// λƥȤλʤ⤷λʤ 
	// unlock 뤬Ͼˤ롥
	this->mRecvRaysBuf[recvBuffNum].completeRecv(); 
    }

    // ׵ȯ
    YorN issueRecvReq = NO;
    for(int i = 0; i < MAX_RECV_BUFFNUM; i++)
    {
	if(this->mRecvRaysBuf[i].issueRecvIfItCan() == YES)
	{
	    issueRecvReq = YES;
	}else
	{	
	    // ⤦ ray ¸ߤʤΤ loop ȴ
	    break;		
	}
    }
    return(issueRecvReq);
}

//------------------------------
// ƤХåեξ YES ֤
//------------------------------
YorN NetIntClass::
isSendRayBufAllEmpty(void) const
{
    for(int i = 0; i < this->mTotalPENum; i++)
    {
 	if(this->mpSendRaysBuf[i].getSizeOfQueue() > 0)
	{
 	    return(NO);
 	}
    }
    return(YES);
}

//------------------------------
// ΤΤ̵ YES ֤
// Ray ˤĤ
//------------------------------
YorN NetIntClass::
isNoSendRay(void) 
{
    for(int i = 0; i < this->mTotalPENum; i++)
    {
 	// ΤΤ¸ߤȤˤ¨ NO
 	if(this->mpSendRaysBuf[i].isLockRealBuffer() == YES)
 	{
	    this->mpSendRaysBuf[i].flushdone(i);
 	    return(NO);
 	}
    }
    return(YES);
}
    
//------------------------------------------------------------
// receive rays
//------------------------------------------------------------
// ΥХåե 1 ĤǤ⤢ NO ֤
// Ray ˤĤ
//------------------------------
YorN NetIntClass::
isNoRecvRay(void) 
{
    for(int i = 0; i < MAX_RECV_BUFFNUM; i++)
    {
 	// ΤΤ¸ߤȤˤ¨ NO
 	if(this->mRecvRaysBuf[i].isLocking() == YES)
	{
	    this->mRecvRaysBuf[i].recv();
 	    return(NO);		// 
 	}
    }
    return(YES);
}

//------------------------------
// ΥХåե
//------------------------------
int NetIntClass::
getNRecvs(void)
{
    int n = 0;
    for(int i = 0; i < MAX_RECV_BUFFNUM; i++)
    {
	if(this->mRecvRaysBuf[i].isLocking() == YES)
	{
	    n++;
	}
    }
    return(n);
}

//------------------------------------------------------------
//  (Pixel 
//------------------------------------------------------------
// 塼ꤷͤۤݻ줿硤׵᤬ȯԲǽʤȯ
void NetIntClass::
sendIntensity(IntensityClass *pIntensity)
{
    this->mSendIntensitiesBuf.
    	enqNewedObjwithMaySend(pIntensity, this->mHostRank);
}

//------------------------------
// flush queue of intensity
//------------------------------
void NetIntClass::
flushIntensitySendBuf(void)
{
    this->mSendIntensitiesBuf.flushdone(this->mHostRank);
}

//------------------------------
// ޤ٤椫
//------------------------------
YorN NetIntClass::
isSendingIntensity(void)
{
    if(this->mSendIntensitiesBuf.isNoMoreWork() == YES)
    {
	return(NO);		
    }else
    {
	return(YES);
    }
}


//------------------------------------------------------------
// ĺХåե
//------------------------------------------------------------
// initialize intensity vertex sendrecv buffer :   call Ƥ
//    @param VIntensitySendThreshold ĺ 
//    @param pVIntensityRecvQueue    Shading ˼ĺΥ塼
//------------------------------
void NetIntClass::
initVerIntBuff(int verIntSendThreshold, 
	       mySLQueue < VertexIntensity * > *pVerIntRecvQueue)
{
    // check internal status ( ray η׻ǽƤʤƤϤʤ)
    if((this->mTotalPENum < 0) || (this->mThisRank < 0))
    {
	cout << "Error! : not init yet. NetIntClass::initVerInttexBuff"<< endl;
	commMpi2_exit(1);
    }

    // check args 
    if(verIntSendThreshold <= 0)
    {
	cout << "Error! : illegal args at NetIntClass::initVerInttexBuff " 
	"initVerInttexBuff = " << verIntSendThreshold << endl;
	commMpi2_exit(1);
    }
    if((pVerIntRecvQueue == NULL) || (pVerIntRecvQueue->length() != 0))
    {
	cout << "Error! : illegal queue at NetIntClass::initVerInttexBuff"
		<< endl;
	commMpi2_exit(1);
    }

    // ĺХåեν
    if(this->mpSendVerIntBuf != NULL)
    {
	// shading Υơϸ߰󤷤ʤΤǡߤϲϤʤϤ
	cout << "Warning: never reach : NetIntClass::initVerIntBuff" << endl;
	delete [] this->mpSendVerIntBuf; // ¸ߤƤˤϲ
	this->mpSendVerIntBuf = NULL;
    }
    this->mpSendVerIntBuf = 
    	new SendBuffer < VertexIntensity > [this->mTotalPENum];
    for(int i = 0; i < this->mTotalPENum; i++)
    {
	this->mpSendVerIntBuf[i].initBuff(verIntSendThreshold);
    }
    this->mRecvVerIntBuf.init(verIntSendThreshold, pVerIntRecvQueue);
}

//------------------------------
//  : ޤȤ
//------------------------------
void NetIntClass::
sendVerInt(VertexIntensity *pVerInt, int destrank)
{
    assert(destrank != this->mThisRank);
    assert((0 <= destrank) && (destrank < this->mTotalPENum));

    this->mpSendVerIntBuf[destrank].enqNewedObjwithMaySend(pVerInt, destrank);
}


//------------------------------
// ĤǤĤäƤȯԤߤ
//------------------------------
void NetIntClass::
flushVerIntSendBuff(void)
{
    for(int i = 0; i < this->mTotalPENum; i++)
    {
	this->mpSendVerIntBuf[i].flushdone(i);
    }
}

// ΤΤ¸ߤ
YorN NetIntClass::
isSendingVerInt(void)
{
    for(int i = 0; i < this->mTotalPENum; i++)
    {
	if(this->mpSendVerIntBuf[i].isNoMoreWork() == NO)
	{
	    // ޤλŻ¸ߤ
	    return(YES);
	}
    }
    return(NO);
}


// 
void NetIntClass::
recvVerInt(void)
{
    this->mRecvVerIntBuf.recv(); 
}

YorN NetIntClass::
isRecvingVerInt(void)
{
    return(this->mRecvVerIntBuf.isLocking());
}


//------------------------------------------------------------
// PEStat ΥۥȤؤ
//------------------------------------------------------------
// initialize PEStat sendrecv buffer :   call Ƥ
//    @param pPEStatRecvQueue  ¾ PE (=  Host)ޥɤΥ塼
//------------------------------
void NetIntClass::
initPEStatHostCommBuff(mySLQueue < HostCommClass * > *pHostCommRecvQueue)
{
    // check args 
    if((pHostCommRecvQueue == NULL) || (pHostCommRecvQueue->length() != 0))
    {
	cout << "Error! : illegal queue at NetIntClass::initPEStatRecvBuff"
		<< endl;
	commMpi2_exit(1);
    }

    // PEStat, HostComm Хåեν
    // ߤϰĤǤ⤢롥
    // ǴԤΤ Nonblocking ȤȤΤߤǤޤȤ
    // θ̤Ϥʤ
    const int PEStatSendThreshold = 1; 
    this->mSendPEStatBuf.initBuff(PEStatSendThreshold);

    const int HostCommRecvBuffLen = 1; 
    this->mRecvHostCommBuf.init(HostCommRecvBuffLen, pHostCommRecvQueue);
}

//------------------------------
// 
//------------------------------
void NetIntClass::
sendPEStat(PEStatClass *pPEStat)
{
    const int destRank = this->mHostRank; // ϸ Host ˸

    assert(destRank != this->mThisRank);
    assert((0 <= destRank) && (destRank < this->mTotalPENum));

    this->mSendPEStatBuf.enqNewedObjwithMaySend(pPEStat, destRank);
}


//------------------------------
// ĤǤĤäƤȯԤߤ
//------------------------------
void NetIntClass::
flushPEStatSendBuff(void)
{
    const int destRank = this->mHostRank; // ϸ Host ˸
     
    this->mSendPEStatBuf.flushdone(destRank);
}

// ΤΤ¸ߤ
YorN NetIntClass::
isSendingPEStat(void)
{
    if(this->mSendPEStatBuf.isNoMoreWork() == YES)
    {
	// ŻϤʤΤǤϤʤ
	return(NO);
    }
    else
    {
	return(YES);
    }
}


// HostComm 
void NetIntClass::
recvHostComm(void)
{
    this->mRecvHostCommBuf.recv(); 
}

//------------------------------------------------------------
// ׾ν
//------------------------------------------------------------
void NetIntClass::
printStat(const char *pMes, const char *pStageName)
{
    //     // radiosity ʿѥ쥤 FIXME ɬפ?
    //     cout << pMes << "stat procradios: " << this->mNumOfProcRadiosities;

    //------------------------------
    //  buffer 
    //------------------------------
    // 祭塼Ĺν :  rank ˤĤƤȺǸ˺ĹʿĹ
    cout << pMes << pStageName << " SendBuff::MaxQueueLength ";
    int maxqueuetotal = 0; 
    int i;
    for(i = 0; i < this->mTotalPENum; i++)
    {
	cout << this->mpSendRaysBuf[i].mCBStat.getMaxQueueLen() << " ";
	maxqueuetotal += this->mpSendRaysBuf[i].mCBStat.getMaxQueueLen();
    }
    cout << " : ave " << 
	(((double)maxqueuetotal) / ((double)this->mTotalPENum)) << endl;

    // 󤢤ǤĤ obj 줿
    cout << pMes << pStageName << " SendBuff::NumberOfObjParSend ";
    double opstotal = 0.0;
    for(i = 0; i < this->mTotalPENum; i++)
    {
	double n_objsend = this->mpSendRaysBuf[i].mCBStat.getNObjParSend();
	opstotal += n_objsend;

	if(n_objsend == 0.0)
	{
	    cout << "(NotUsed) "; // Ѥʤä
	}else
	{
	    cout << n_objsend << " ";
	}
    }
    cout << " : ave " << 
    	(((double)opstotal) / ((double)this->mTotalPENum)) << endl;

    //------------------------------
    //  buffer 
    //------------------------------
    // 󤢤μǤĤ obj 줿
    // MPI ݾڤΤ餳פ FIXME
    cout << pMes << pStageName << " RecvBuff::NumberOfObjParSend ";
    opstotal = 0.0;
    for(i = 0; i < MAX_RECV_BUFFNUM; i++)
    {
	double n_objrecv = this->mRecvRaysBuf[i].mCBStat.getNObjParSend();
	opstotal += n_objrecv;

	if(n_objrecv == 0.0)
	{
	    cout << "(NotUsed) "; // Ѥʤä
	}else
	{
	    cout << n_objrecv << " ";
	}
    }
    cout << " : ave " << 
    	(((double)opstotal) / ((double)MAX_RECV_BUFFNUM)) << endl;
}
