//------------------------------------------------------------
// $Id: PE.h,v 1.3 2000/05/19 20:35:50 yamauchi Exp $
//------------------------------------------------------------
//	PE.h Processing Element
//	1996/7/10 
//	Copyright (C) 1996-1998 YAMAUCHI Hitoshi :  
//------------------------------------------------------------
#ifndef _PEClass_h
#define _PEClass_h 1

#include "SLList.h"
#include "DLList.h"

#include "mySLQueue.h"
#include "materialArray.h"
#include "objectSet.h"
#include "sysinfo.h"
#include "viewdata.h"
#include "spc2PE.h"
#include "NetIntUnit.h"
#include "rayTrFRGUnit.h"
#include "searchNextSpcDDA.h"
#include "volumeSet.h"
#include "PRC_searchNextSpc.h"
#include "frustumSRG.h"
#include "PEpmarray.h"

class Ray;

//------------------------------------------------------------
// ray ζ֥å׿
//------------------------------------------------------------
// class skipSpaceCounter {
// private:
//     skipSpaceCounter(const skipSpaceCounter& rhs);
//     const skipSpaceCounter& operator=(const skipSpaceCounter& rhs);
// protected:
//     double 		mSkipSpcOfRays;	      // ray ֤ skip 
//     double 		mTravSpcOfRays;	      // ray ֤̲ᤷ
//     void   incTravSpcNum(void)	{ this->mTravSpcOfRays += 1.0;	}
//     void   incSkipSpcNum(void)	{ this->mSkipSpcOfRays += 1.0;	}
// public:
//     skipSpaceCounter(void);
//     ~skipSpaceCounter(void);
//     double getSkipSpcOfRays(void) const { return(this->mSkipSpcOfRays);}
//     double getTravSpcOfRays(void) const { return(this->mTravSpcOfRays);}
// };

//------------------------------------------------------------
// related functions with PE
//------------------------------------------------------------
extern double getTotalUnshotRadiosityEnergy(SLList < objBaseClass * >*pObList);

//------------------------------------------------------------
// PEClass PE μ
//------------------------------------------------------------
class hemiCubeClass;		

class PEClass {
private:
    PEClass(const PEClass& rhs);
    const PEClass &operator=(const PEClass& rhs);
    
protected:
    //------------------------------
    // system, hardware, rendering parameters
    //------------------------------
    IntVec3 	     	mID;	// this processing element number
    int			mMyRank;
    SysInfoClass	mSysInfo; 	// system information
    ViewDataClass	mViewData;	// view point 
    SpaceParam		mSpcParam;	// space parameter

    //------------------------------
    // ray generator for radiosity delta-radiosity 
    //------------------------------
    hemiCubeClass      *mpHemiCube;
    Ray	     **mppRadiosityRays; 	// radiosity  hemicube Ȥ̿

    //------------------------------
    // ray generator for ray-tracing 
    //------------------------------
    rayTrFRGUnitClass	mRayTrFRGUnit;

    //------------------------------
    // Local memory contents of this PE 
    //------------------------------
    SLList < objBaseClass * > 	mObjList;   // PEô֤object
    objectSetClass  		mObjectSet; // ®˥뤿ε
    //  PE ô֤ؤ volume ѡ϶ˤʤ롥
    SLList < VolumeClass * > 	mVolList;  
    // volume ع®˥뤿ε˼Τ롥
    VolumeSetClass		mVolSet;    

    SLList < materialClass >  	mMatList;       // material μΤݻ
    materialArrayClass 		mMaterialSet;   // materials إ
    mySLQueue < Ray * > 	mRayPtrQueue;   //  ray Υ塼
    DLList < objBaseClass * > 	mRadPatchQueue; // Ͳǽѥåꥹ
    mySLQueue < VertexIntensity * > mVerIntPtrQueue;// ĺ٤Υ塼
    mySLQueue < HostCommClass * > mHostCommRecvQueue; // HostComm ѥ塼
    //------------------------------
    // DDA ʬʬΥ (FIXHERE)
    //------------------------------
    searchNextSpcDDA 		mDDAUnit;
    NetIntClass			mNIUnit;

    //------------------------------
    // processor status 
    //------------------------------
    YorN mIsMultiPassStageOver;    // ƥơλɤ
    YorN mIsPrepareShadingOver;	   // shadingνơϽλ
    YorN mIsIdle;		   // Ż̵
    YorN mIsStopGenNewRay;	   // new rays ǽ(host ξ)
    YorN mSendingRayStopComToHost; // hostrayߤ׵(ʤξ) 


    //------------------------------
    // Statistics : ׾
    //------------------------------
    GeneralStatistics 		mGStat;
public:    
    void printStatInfo(const char *pStageName);


protected:		
    //------------------------------
    // for PRC method
    //------------------------------
    searchNextSpcPRC		mPRCUnit;
    
protected:
    //----- initialize 
    // initialize after receiving sysinfo, viewdata, spaceparam
    void initRankNID(void); 		// child of initInternalUnits()
    void initInternalUnits(void);	// 
    
    // initialize after receiving object data
    void getHavingSubSpcNum(int *spcnum); //  PE λļºݤʬ֤ο
    void initObjectSet(void);		  // child of initObjectMaterialSet
    void initObjectMaterialSet(void);	  // init after material received
    void initRecvVolumeSet(void);	  // init after volume received
    void getSetOfSameSpcIDVolume(SLList < VolumeClass * > *pSameSpIdVlist,
				 SLList < VolumeClass * > *pVolPtrListCopy);

    void outRecvObjectInfo(void);	  // output object   information
    //    void outRecvMaterialInfo(void);	// output material information
    
    void exitprog(void)
    {
	cout << "<< PE:: " << this->mID << " : exit" << endl;
	commMpi2_exit(1);
    }

    // load balancing
    void globalLoadBalance(void);

protected:
    // radiosity and raytracing common method
    void initBusyIdleStat(void) 
    { 
	this->mIsIdle = YES; 	// idle -> busy Ѳ
	this->toBusy();		// ֤Ư
    }
    void toBusy(void);
    void toIdle(void);    
    YorN isIdle(void) const 	{ return(this->mIsIdle);	}
    // HitPoint ʬˤ뤫λʤɤϳ hit 뤳Ȥꡥ
    YorN checkHitPosIsInThisSpcPos(const Vector3& hitpoint, 
				   const IntVec3& spcidx) const;
   
    YorN isMultiPassStageEnd(void) 
#ifndef FOR_STANDALONE	
	const {	return(this->mIsMultiPassStageOver); 	}
#endif
    ;
    // hostˤ PE  stat ãѤʤȡtoBusy,
    // toIdle Ѥ;פ̿򸺾Ǥ  
    void sendHostToThisPEIsBusy(YorN isBusy); 
    int  probNgetHostCommand(void); 	// host̿probe,,
    //  2 쥤
    YorN newSecondaryRadiosityRays(Ray *pRay, const materialClass *pMat,
				   objBaseClass * pHitObj);

    
    YorN newSecondaryRayTracingRays(Ray *pRay, const materialClass *pMat,
				    objBaseClass * pHitObj);

    // õȥ쥤ΰư PEGetNextSpace.cc
    YorN moveRayToNextSubspace(Ray *pRay); 

protected:
    // radiosity method
    void initForRadiosity(void); 	  // 
    void initializePatches(void);	  // calcRadioFRG.cc 
    void finalizeForRadiosity(double totalinitNonRadEnergy); // PERadiosity.cc
    void radiateRadFromMaxRadPatch(void); // calcRadioFRG.cc
    void processingRadiosities(void); 
    void processingOneRadiosity(Ray *pRay); 
    void calcNsetIntensityRadio(Ray *pRay, // calclRadioIC.cc
				const materialClass *const pMat,
				objBaseClass *pHitPat);

protected:
    // shading
    void initPrepareShadingStage(void);
    YorN isPrepareShadingStageOver(void) 
#ifndef FOR_STANDALONE
	const { return(this->mIsPrepareShadingOver);	}
#endif
    ;
    // below functions in prepareShading.cc
    YorN checkInSpc(const IntVec3 &sppos);// ⤫
    void prepareGouroudShading(void);	  // gouroud shading 

    // Ĥζ֤ε٤ν
    void sendVerIntOfThisSpc(objectsArrayClass *pOAry, const IntVec3& srcspc, 
 			     const Vector3& spmin,     const Vector3& spmax,
 			     const double tolerance);
    // εپ٤ΤɤȽꤷ
    void testOneIntNsendVerInt(const VertexIntensity& vint,
			       const IntVec3& srcspc, const Vector3& spmin,
			       const Vector3& spmax,  const double tolerance);
    // ֤صپԤԤ(PEϼʬ⤷ʤ)
    void sendVerIntToOtherSpcNrecv(const VertexIntensity& verint); 
    void addIntensity(const VertexIntensity& verint); // ٲû
    void finalizePrepareShadingStage(void);	 // ̿ʿ

protected:
    // ray tracing method
    void initForRayTracing(void);

    // calcRayTrIC.cc
    void calcNsetIntensityRayTr(Ray *pRay, objBaseClass *pHitObj);

    void processingRayTraces(void);		     
    void processingOneRayTrace(Ray *pRay);		     
    objBaseClass *calcInterSectionWithRayAndObj(Ray *pRay);
    void calcRayTrNextSpace(Ray *pRay);
    YorN isObjectInSpc(const IntVec3 &spcpos);    

    void shotSomeFRGRays(void);

protected:
    //---------- network interface unit ----------
    YorN isRayInValidSpace(const Ray *pChkRay, int callpos); // debug

    //----- send ray
    YorN flushSendRaysBuffer(void);
    // void sendRays(void);

    //----- recv rays
    void recvRaysIfItCan(int buffnum);
    void recvRays(void);
    void setRayToNI(Ray* pRay);

    //----- 
    void insertIntensityIntoBuffer(Ray *pRay);
    void sendRayIntensityToFB(void);
    
public:
    // PE stat
    PEClass(void);    
    ~PEClass(void);    
    IntVec3 getID(void) const	{ return(this->mID);	}
    int  getMyRank(void) const;
    int  getRank(const IntVec3& id) const; // id -> rank 
    void setViewData(const ViewDataClass& aViewdata);

    // object and material
    void addOneMaterial(materialClass *);
    // void delOneMaterial(material *);
    void addOneObject(objBaseClass *);
    // void delOneObject(object *);

    // communication
#ifdef FOR_STANDALONE
    // for stand alone
    void receiveSysinfoNinitSpaceInfo(const SysInfoClass&    sys,
				      const ViewDataClass&   vw,
				      const SpaceParam&	     sp); 
    void receiveMaterials(const materialClass &mat);
    void receiveObjects(objBaseClass* pObj);    
    void receiveVolume(VolumeClass* pVolume);    
    void receiveObjInSpcMap(int bytesize,  OBJINSPC_IMPLEMENT_TYPE *pDatArray);
    int  sendOutRadioResultToHost(void){ return(0); } // no mean at standalone

    // output intermediate result
    void outRadiosityMpi2Data(const char* pBasename, YorN isShotRad);
    void writeGeometryFile(FILE *pFp, YorN isShotRad);
    void outRadiosityIntermediateResult(void); // ַ̤ν

    // for PRC method (see PRC_RecvStandAlone.cc)
    void receivePRCSkipMap(int bytesize, OBJINSPC_IMPLEMENT_TYPE *pDatAry);
#else
    void receiveSysInfoNinitSpaceInfo(void);
    void receiveMaterials(void);
    void receiveObjects(void);
    void receiveVolume(void);    
    void receiveObjInSpcMap(void);        
    void sendOutRadioResultToHostAccordHostComm(void); 
    void sendOutRadioResultToHost(void); // δؿƤФФΤޤ

    // output intermediate result
    void outRadiosityMpi2Data(const char* pBasename, YorN isShotRad);
    void writeGeometryFile(FILE *pFp, YorN isShotRad);
    void outRadiosityIntermediateResult(void); // ַ̤ν

    // for PRC method
    void receivePRCSkipMap(void); 
#endif

    void doRadiosity(void);
    void doPrepareShading(void);
    void doRayTracing(void);
    void barrierSync(void);	// ΤΥХꥢƱ time out 꼺Ԥexit

    void setRayPtrToThisPE(Ray* pRay)
    {				// set ray to PE(from otherPE or generate rays)
	assert(pRay->isOK());
	this->mRayPtrQueue.enq(pRay);	
    }
    // void setRayPtrToOtherPE(int recvRank, Ray* pRay);

    IntVec3 getLocalIdxFromSpcPos(const IntVec3& spcpos)
    {
	return(this->mNIUnit.mpSpc2PE->getLocalIdxFromSpcPos(spcpos));
    }
    
    void printSysInfo(const char *pHead); // pHead ϳƹƬ˽Ϥʸ 
    void printObjects(const char *pHead);
    void printMaterials(const char *pHead);
    void printObjInSpcMap(const char *pHead);

    // network interface unit
    YorN isDealSpace(const IntVec3& spcpos) const; //  PE ô֤
    void initSpc2PE(void); 		// child of initInternalUnits()

protected:
    //tokunaga distribute Rays
    frustumSRGClass mFrustum;	
    YorN secondRayGenWithDistrbute(Ray *pRay, const materialClass *pMat,
				   objBaseClass * pHitObj);
    YorN secondRayGenWithoutDistribute(Ray *pRay, const materialClass *pMat,
				       objBaseClass * pHitObj);
public:    
    void makeDDMap(void);
};

#endif				// _PEClass_h
