//------------------------------------------------------------
// $Id: PERayTracing.cc,v 1.1.1.1 1999/11/25 06:26:09 yamauchi Exp $ 
//------------------------------------------------------------
//	PERayTracing.cc : Processing Element Ray-Tracing stage
//	1999ǯ125() PE.cc ʬΥ
//	Copyright (C) 19961999 YAMAUCHI Hitoshi  
//------------------------------------------------------------
#include "PE.h"
#include "mpi2communication.h"
#include "Option.h"
#include "volumeFuncs.h"
#include "mpi2_funcs.h"
#include "ProfileMon.h"		// for program marker array 

//------------------------------------------------------------    
// ray tracing sub functions
//------------------------------------------------------------    
// ray tracing Τν
//------------------------------
void PEClass::
initForRayTracing(void)
{
    // ֤Ǥ ray Υ塼ˤϲäƤƤϤʤΤĴ٤
    // (ȤСradiosity η׻λĤʤɤϤäƤϤʤ)
    assert(this->mRayPtrQueue.empty() != Q_NOT_EMPTY);

    this->mIsMultiPassStageOver = NO;
    if(this->mSysInfo.isRayTracingCalc() == NO)
    {
	// ray-tracing η׻򤷤ʤȤνϤɤ
	// ɤΤƤʤǤŬʥѥ᡼
	// Ƥ뤬ʾ֤ˤʤʤ褦ˤȤ̤ΰ̣
	// 롥̤ñ exit Ƥɤ⤷ʤYam
	this->mViewData.setMaxRayRefParam(DEFAULT_MAX_REFLECTION_RAY_NUM,
					  DEFAULT_MIN_REFLECTION_COEFF_NORM2);
	this->mViewData.setDistributedRayNum(1);
	this->mViewData.setFrustumParam(DEFAULT_FRUSTUM_RES);
	if(Option::getInstance()->getVerboseMode() == YES)
	{
	    cout << genHd(VERBOSEHEAD, this->getID()) << 
	    "ray-tracing calculation is off. change view parameter." << endl;
	    this->mViewData.print(genHd(VERBOSEHEAD, this->getID()));
	}
    }

    // initialize Network Interface unit
    this->mNIUnit.initRayBuff(this->mSysInfo.getTotalPENum(),
			      this->mSysInfo.getProcRayTrAtOnce(), 
			      this->mSysInfo.getSendInt2FBAtOnce(),
			      this->getMyRank(),
			      SysStat::getInstance()->getHostRank(),
			      &(this->mRayPtrQueue));
    
    // initialize First Ray Generation Unit
    this->mRayTrFRGUnit.init(this->mSysInfo.getFirstRaysGenerationMethod(),
			     this->mViewData, this->mSpcParam, this->mID,
			     this->mSysInfo.getTotalPENum(),
			     this->getMyRank());

    if(this->mRayTrFRGUnit.canGenerateNextRay() == YES)
    {
	// 쥤Υå: ξʤɤͤʤ
	// Ȥ餫ʾ礬ꡢξˤϥ쥤κϤǤʤ
	this->shotSomeFRGRays();
    }
}

//------------------------------
//   ray ν ʣܤΥ쥤
//------------------------------
void PEClass::
processingRayTraces(void)
{
    // δؿƤФˤϰʲξƤ
    //     塼˰ܤϥ쥤¸ߤƤ
    assert(this->mRayPtrQueue.empty()  == Q_NOT_EMPTY);
    
    // ʲξλ˥쥤ܼФƤνԤʤ
    // ʲξϤδؿ˥ȥꤷˤƤΤ 
    // do{}while;Υ롼פȤʤ롣 
    // 	   쥤塼¸ߤ
    // 	   쥤 getProcRayTrAtOnce ʲ
    int i = 0; // 쥤
    static Ray *pRay;
    do {
	assert(this->mRayPtrQueue.length() > 0);
	pRay = this->mRayPtrQueue.deq();
	this->processingOneRayTrace(pRay);  
	i++;
    }while((this->mRayPtrQueue.empty()  == Q_NOT_EMPTY) && 
	   (i < this->mSysInfo.getProcRayTrAtOnce()));
    
    // Statistics.addRayTraceProcsAtOnce(i); FIXME פϤȤ뤫?
}


//------------------------------
// RP Block νԤٷ׻νɬפʤԤ̤ Host 
// 롣ɬפʤ¾ PE ؤԤ
//    processingOneRadiosity Ȥʤ궦ʬ뤬 Hit  
//   ray νۤ  
//------------------------------
void PEClass::
processingOneRayTrace(Ray *pRay)
{
    // ʬξˤ ray ¸ߤ֤ôƤʤ PE  
    // DDA  η׻Ԥǽ롥 FRG  ray 
    // ľˤ PE  DDA η׻ԤäơʪΤ¸ߤ֤ 
    // ray 뤳Ȥǲ򤹤٤äơ ray Τ
    // Ϥ PE ô֤ǤʤƤϤʤʤ 
    assert(this->isRayInValidSpace(pRay, 2) == YES);
    objBaseClass *pHitObj = this->calcInterSectionWithRayAndObj(pRay);
    
    // ٷ׻󼡥쥤ȯõ
    if(pHitObj != NULL)
    {
	// ʪΤ ray 򺹤
	materialClass *pMat = this->mMaterialSet.get(pHitObj->getMaterialID());

	// ʪΤε٤׻ray ˥å
	this->calcNsetIntensityRayTr(pRay, pHitObj);

	// ȿͶޤԤ : ߤ ray  IC->SRG ήˤʤä
	// 롥
	if(this->newSecondaryRayTracingRays(pRay, pMat, pHitObj) == NO)
	{
	    // ȿͤޤ ray [ɤ]ʤäޤ
	    // ׻٤ Frame Buffer 롣Ǥ
	    // 줿ˤϤޤǤε٤Ϥɤ餫 ray Ƥ
	    // 롣줬ǽŪ˾ä Frame Buffer 롣

	    this->insertIntensityIntoBuffer(pRay);
	}
	delete pRay;		// deleteRay!
	pRay = NULL;
	// 򺹤ΤǼ֤ؤϹԤʤ(2 쥤Ѥ)  
    }else
    {
	// õ : ֤򳰤 ray 
	this->calcRayTrNextSpace(pRay);
    }	
}

//------------------------------
//  PE ôʬ֤ǤΥ쥤ʪΤθȽ
//------------------------------
objBaseClass *PEClass::
calcInterSectionWithRayAndObj(Ray *pRay)
{
    // ray Τ֤Ϥ PE ô֤?
    if(this->isDealSpace(pRay->getSpcPos()) != YES)
    {
	pRay->print("DEBUG:");	// DEBUG
    }
    assert(this->isDealSpace(pRay->getSpcPos()) == YES);

    //------------------------------
    // ʪΤȤθȽ 
    //------------------------------
    objectsArrayClass *pObjArray = this->mObjectSet.get(pRay->getSpcPos());
    int objNumInThisSpc = pObjArray->getLength();
    int i;
    double distanceBetweenRayNObject = PlusInfty;
    static Vector3 hitPoint;
    objBaseClass *pHitObj = NULL;
    for(i = 0; i < objNumInThisSpc; i++)
    {
	objBaseClass *pObj = pObjArray->get(i);
	if(pObj != pRay->getRadObj()) // ray ȯ̤ȤϸȽꤷʤ
	{
	    double tmpDist = pObj->getDistance(pRay, &hitPoint);
	    assert(tmpDist >= 0.0);

	    // Τ褦ʪΤξˤϤʬ֤¸ߤʤʬ 
	    // hit ǽ롥ξ 
	    switch(pObj->who())
	    {
	    case TRIANGLE_OBJ_E:	
		break;
	    case SPHERE_OBJ_E:
		if(this->checkHitPosIsInThisSpcPos(hitPoint, pRay->getSpcPos())
		   == NO)
		{
		    tmpDist = PlusInfty;
		}
		break;
	    case NULL_OBJ_E:		
	    default:
		cout << "Internal Error! : " << endl;
		commMpi2_exit(1);
		break;
	    }
	    
	    if(distanceBetweenRayNObject > tmpDist)
	    {
		// tmpDist ᤤΤǸ򺹤obj򹹿򥻥åȤ
		distanceBetweenRayNObject = tmpDist;
		pHitObj = pObj;
		pRay->setHitPos(hitPoint);
	    }
	}
    }

    //------------------------------
    // Volume Ȥߺ 
    //------------------------------
    SLList < VolumeClass * > *pVolList = 
    	this->mVolSet.getList(pRay->getSpcPos());
    if(!(pVolList->empty()))
    {
	// volume ¸ߤˤߺѤ׻
	processingOneRayTraceWithVolume(pRay, pVolList, 
					distanceBetweenRayNObject);
    }

    // ⤷ volume κѤǤ NULL ˤ٤
    return(pHitObj);
}

//------------------------------
// DDA ׻ray ֤ۤˤ ray 
//------------------------------
void PEClass::
calcRayTrNextSpace(Ray *pRay)
{
    // 򺹤ʪΤ¸ߤʤ õ (PEGetNExtSpace.cc)
    YorN isSendRay = this->moveRayToNextSubspace(pRay);
 
    if(isSendRay == YES)	// ray 
    {
	this->setRayToNI(pRay); // NIΥХåեؤᤳ & 
    }else
    {
	// ٤ե졼Хåեƥ
	this->insertIntensityIntoBuffer(pRay);	    
	delete pRay;	// deleteRay!
	pRay = NULL;
    }
}

//------------------------------
// 쥤ȯ for ray-tracing
//------------------------------
void PEClass::
shotSomeFRGRays(void)
{
    // ƤӽФȤȤϺǾ 1 ܤȯǽǤϤ
    assert(this->mRayTrFRGUnit.canGenerateNextRay() == YES);

    int i;
    const int maxGenRaysAtOnce = 
	this->mSysInfo.getProcRayTrAtOnce() * GEN_FRAYS_FACTOR_BY_COMMUN;
    for(i = 0; i < maxGenRaysAtOnce; i++)
    {
	//  ray ȯ
	Ray *pRay = this->mRayTrFRGUnit.newNextRay();

	// ʣ PE ǶƱ ray ȯˤ ray ȯ
	// Ƥ줬ô֤ȸ򺹤뤫ʤƤϤʤΤǡ
	// Ȥ꡼ǤäƤºݤˤ PE  ray ȯ
	// ʤ礬롣Τ褦ʾˤ NULL äƤ롣
	if(pRay != NULL)	
	{
	    if(this->isDealSpace(pRay->getSpcPos()) == YES)
	    {
		// ray ȯˤ PE ô֤ˤ
		// Ƿ׻
		this->setRayPtrToThisPE(pRay);
	    }else 
	    {
		// ray ȯˤ PE ô֤ˤʤ
		if(this->isObjectInSpc(pRay->getSpcPos()) == YES)
		{
		    // ֥Ȥ ray θ¸ߤ֤¸
		    //  
		    // ô PE  : NI 塼
		    this->setRayToNI(pRay); 
		}
		else
		{
		    // ֥ȤϤ ray θ¸ߤ֤ˤ¸
		    // ߤʤõԤ 
		    this->calcRayTrNextSpace(pRay);
		}
	    }
	}else
	{
	    // ray ȯʤä뤳ȤϤʤ
	}
	if(this->mRayTrFRGUnit.canGenerateNextRay() == NO)
	{
	    // ray ȯλ
	    if(Option::getInstance()->getVerboseMode() == YES)
	    {
		cout << genHd(VERBOSEHEAD, this->getID()) << "generate rays( " 
		<< this->mRayTrFRGUnit.getTotalGenRays() << " ) done."<< endl;
	    }
	    break;
	}
    }
}

//------------------------------
// ray tracing
//------------------------------
void PEClass::
doRayTracing(void)
{
    if(Option::getInstance()->getVerboseMode() == YES)
    {
	cout << genHd(VERBOSEHEAD, this->getMyRank()) << 
	    "PE in do raytracing OK" << endl;
    }

    ProfileMon *pPfm = ProfileMon::getInstance();
    pPfm->startStage(ProfileMon::ProfSt_raytracing);
    
    this->initForRayTracing();
    if(Option::getInstance()->getVerboseMode() == YES)
    {
	cout << genHd(VERBOSEHEAD, this->getMyRank()) << "PE in inited OK"
	<< endl;
    }

    // init status 
    this->initBusyIdleStat();
    this->mIsStopGenNewRay = NO; // host  ray ȯ׵Ϥʤ
    this->mSendingRayStopComToHost = NO; // host ؤ ray ȯ׵̤

    while(this->isMultiPassStageEnd() == NO)
    {
	this->recvRays(); // ޤ¾
	// ¾
	//  ray  enqueue 줿٤ʤ
	// ᡤǤ send Ϥʤ
	// this->sendRays(); 
	// this->mNIUnit.flushIntensitySendBuf();// ٤ۥȤ
	
	// processing
	if(this->mRayPtrQueue.empty() == Q_NOT_EMPTY)
	{
	    // 쥤ʤ && ٤ ray 
	    // (RP Block) ray 롣ٷ׻ޤࡣ
	    assert(this->isIdle() == NO);
	    pPfm->occurEv(ProfileMon::ProfEv_ProcRayRays);

	    this->processingRayTraces(); 
	    this->globalLoadBalance(); // ray ̤ˤ ray ȯ¤Ԥ
	}
	else if(this->mRayTrFRGUnit.canGenerateNextRay() == YES)
	{
	    // 塼˥쥤ʤ̤ȯͥ쥤
	    this->toBusy();

	    // 쥤ȯ host  ray ȯ׵᤬
	    // Ĵ٤ 
	    this->probNgetHostCommand();
	    if(this->mIsStopGenNewRay == NO) // host Ϥʤ
	    {
		pPfm->occurEv(ProfileMon::ProfEv_genRays); 
		this->shotSomeFRGRays();
	    }
	}
	else if(this->mNIUnit.isSendRayBufAllEmpty() == NO)
	{
	    // 쥤ʤ && buff ˥쥤ĤäƤ
	    assert(this->isIdle() == NO);
	    this->flushSendRaysBuffer(); //  buf ˶ǽʤ
	    pPfm->occurEv(ProfileMon::ProfEv_SendingWait);

 	}else if((this->mNIUnit.isNoSendRay() == YES) && // ǤϤʤ && 
 		 (this->mNIUnit.isNoRecvRay() == YES))   // ǤϤʤ
 	{
	    // ٤ PEStat  flush
	    this->mNIUnit.flushIntensitySendBuf(); //   flush
	    this->mNIUnit.flushPEStatSendBuff();   // PEStat  flush

	    if((this->mNIUnit.isSendingIntensity() == NO) && 
	       (this->mNIUnit.isSendingPEStat()    == NO))
	    {
		// 쥤ʤ && ٤ ray ̵ && λ
		// && λ (ray &&  && PEStat)
		// == ƤλŻλ
		if(this->isIdle() == NO) // busy -> idle
		{
		    this->toIdle();		
		    assert(this->isIdle() == YES);
		    pPfm->occurEv(ProfileMon::ProfEv_Idle);		    
		}else
		{
		    // ƽʤ idle ΤޤޤǤä
		    // ⤷λΥޥɤ endOfRayTracing  YES 
		    // ѹ 
		    this->probNgetHostCommand(); 
		    if(this->isMultiPassStageEnd() == YES)
		    {
			// λޥɤˤϥ塼϶
			assert(this->mRayPtrQueue.empty() != Q_NOT_EMPTY);
		    }
		}
	    }
	    pollingSleep();
	}else
	{
	    this->toBusy();	    
	}
    }
#ifdef FOR_STANDALONE
    this->toIdle();
#endif
    this->mNIUnit.flushIntensitySendBuf();  // ٤ۥȤ
    if(Option::getInstance()->getVerboseMode() == YES)
    {
	cout << genHd(VERBOSEHEAD, this->getMyRank())<<"out of PE loop"<< endl;
    }
    pPfm->endStage(ProfileMon::ProfSt_raytracing);
}


