//------------------------------------------------------------
// $Id: PErecv.cc,v 1.2.2.1 2000/06/14 15:00:11 yamauchi Exp $ 
//------------------------------------------------------------
//	PE.cc Processing Element
//	1996/3/22
//	(C) Copyright Hitoshi YAMAUCHI
//	1997/1/21
//------------------------------------------------------------
#ifdef FOR_STANDALONE
# error "This file is for multi-PE (parallel) version."
#endif

#include <string.h>
#include "mpi2_funcs.h"
#include "mpi2communication.h"
#include "Option.h"
#include "PE.h"

//------------------------------------------------------------
// This is the part of multi-PE (parallel) version.
//------------------------------------------------------------
// get sysinfo and initialize ID and this Spc2PE
//------------------------------
void PEClass::
receiveSysInfoNinitSpaceInfo(void)
{
    if(this->mObjList.empty() == Q_NOT_EMPTY) // ʤˤ
    {
	cerr << "Error! : there are objects in list PE(" << this->mID << 
	    ") at PEClass::receiveSysinfo" << endl;
	this->exitprog();
    }
    commMpi2_BroadSysinfo(&(this->mSysInfo), 
 			  &(this->mViewData), 
 			  &(this->mSpcParam));

    // initialize after receiving sysinfo, viewdata, spaceparam
    this->initInternalUnits();
}

//------------------------------------------------------------
// μ
//------------------------------------------------------------
void PEClass::
receiveMaterials(void)
{
    materialClass mat;
    YorN isEnd = NO;
    while(isEnd == NO)
    {
	commMpi2_BroadMaterial(&mat);
	if(strncmp(mat.getName(), NULLMaterialName, 
		   MAX_MATERIAL_NAME_LEN - 1) != 0)
	{
	    this->mMatList.append(mat);	// null material ǤϤʤ
	}else
	{
	    if(Option::getInstance()->getVerboseMode() == YES)
	    {
		cout << "## PE:: r " << this->getMyRank() << 
		    " : get null material " << endl;
	    }
	    isEnd = YES;
	    // break;		// null material ǽλ
	}
    }
    // this->outRecvMaterialInfo(); ȤꤿƤͭפʾǤϤʤΤǽϤʤ
}

//------------------------------
// volume μ
//------------------------------
void PEClass::
receiveVolume(void)
{
    VolumeClass *pVolume = NULL;
    YorN isEnd = NO;
    while(isEnd == NO)
    {
	commMpi2_recvVolume(SysStat::getInstance()->getHostRank(), 
			    this->getMyRank(), &pVolume);
	if(pVolume->isNullVolume() == YES)
	{
	    cout << "// PE : " << this->mID << " : r " << this->getMyRank()
		<< " get null volume" << endl;
	    isEnd = YES;
	}else
	{
	    this->mVolList.append(pVolume);
	}
    }
    // already got null volume, initialize now
    this->initRecvVolumeSet();
}

//------------------------------------------------------------
// ʪ(sizeof = )μ
//------------------------------------------------------------
void PEClass::
receiveObjects(void)
{
    objBaseClass *pObj = NULL;
    YorN isEnd = NO;
    while(isEnd == NO)
    {
	commMpi2_recvObject(SysStat::getInstance()->getHostRank(), 
			    this->getMyRank(), &pObj);
	if(pObj->who() == NULL_OBJ_E)
	{
	    cout << "// PE : " << this->mID << " : r " << this->getMyRank()
		<< " get null object" << endl;
	    isEnd = YES;
	}else
	{
	    this->mObjList.append(pObj);
	}
    }
    this->initObjectMaterialSet();
    this->outRecvObjectInfo();
}

//------------------------------------------------------------
// ʪ¸߾롣饹μΩ뤿˾
// ϰ켡ˤʤäƤΤǤ objInSpcClass 
// ˳Ǽ.
// (ref. host.cc : hostComputerClass::readAndDeliverGeoData())  
//------------------------------------------------------------
void PEClass::
receiveObjInSpcMap(void)
{
    int size[3];
    size[0] = size[1] = size[2] = this->mSpcParam.getSubDivNum();
    const int totalSize = size[0] * size[1] * size[2];
    assert(totalSize > 0);	//  SpcParam ϼѤߤΤϤ

    OBJINSPC_IMPLEMENT_TYPE *pDatArray = 
	new OBJINSPC_IMPLEMENT_TYPE [totalSize];
    commMpi2_BroadObjInSpc(sizeof(OBJINSPC_IMPLEMENT_TYPE) * totalSize, 
			   pDatArray);     

    // Map ΰȽ
    this->mDDAUnit.mObjInSpcMap.initBySizeWithBaseZero(size); 

    // ƤΥԡ
    for(int i = 0; i < this->mDDAUnit.mObjInSpcMap.getTotalSize(); i++)
    {
	objectExistInASpaceClass oeias(pDatArray[i]);
	this->mDDAUnit.mObjInSpcMap.setAsOneDimArray(i, oeias);
    }
}


//------------------------------
// host  radiosity η̤ʤƤϤʤȤ狼äƤ
// ˤδؿԤġhost ̿᤬褿롥
// radiosity η׻̤ host 
//------------------------------
void PEClass::
sendOutRadioResultToHostAccordHostComm(void)
{
    while(1)
    {
	this->mNIUnit.mRecvHostCommBuf.recv(); // 
	if(this->mHostCommRecvQueue.length() == 0)
	{
	    retrySleep();	// ۥȤΥޥɤϤʤ
	}
	else
	{
	    // Host Υޥɤ
	    assert(this->mHostCommRecvQueue.length() > 0);
	    HostCommClass *pRecvHostCom = this->mHostCommRecvQueue.deq();
	    HostCommClass recvHCom = (*pRecvHostCom);
	    delete pRecvHostCom; // Τ new Ƥ
	    
	    if(recvHCom.getCommand() == HOSTCOM_Send_Patch_Descriptions_e)
	    {
		PEoutDebugMes(this->getMyRank(), recvHCom.toString());
		this->sendOutRadioResultToHost();
		return;		// λΤǤΥޥɤ
	    }
	    else
	    {
		cout << "Error! : received illegal host command. " <<
	    	"PEClass::sendOutRadioResultToHostAccordHostComm"  << endl;
		cout << "this rank = " << this->getMyRank() << endl;
	    }
	}
    }
}


//------------------------------
// radiosity η׻̤ host 
//------------------------------
void PEClass::
sendOutRadioResultToHost(void)
{
    int totalnum = 0;
    Pix pidx;
    for(pidx = this->mObjList.first(); pidx != NULL; this->mObjList.next(pidx))
    {
	commMpi2_sendObject(SysStat::getInstance()->getHostRank(), 
			    this->mObjList(pidx));
	totalnum++;
    }
    objBaseClass *pNullObject = new nullObjectClass;
    commMpi2_sendObject(SysStat::getInstance()->getHostRank(), pNullObject);
    delete pNullObject;

    // output debug message
    char buf[256];
    sprintf(buf, "send objects to host, total objes %d", totalnum);
    PEoutVerboseMes(this->getMyRank(), buf);
}


//------------------------------
// Radiosity ַ̤Ϥ
//------------------------------
void PEClass::
outRadiosityMpi2Data(const char *pOutBaseFname, YorN isShotRad)
{
    cout << "NIN: ǤξˤƱƤη̤Ϥɬ"
    "뤬Ƥʤ " << endl;
    // Ʊƾ嵭 sendOutRadioResultToHost Ƥַײ
}




