//------------------------------------------------------------
// $Id: PEvolume.cc,v 1.2.2.1 2000/06/14 15:00:10 yamauchi Exp $ 
//------------------------------------------------------------
//	PEvolume.cc Processing Element functions for volume
//	1998/11/17
//	Copyright (C) 1998 YAMAUCHI Hitoshi  
//------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream.h>
#include "SLList.h"
#include "PE.h"
#include "mpi2exit.h"
#include "mpi2_funcs.h"
#include "volume.h"


// ѤĤ
static YorN isVolume(VolumeClass *pVol)
{
    IntVec3 resolSize = (pVol->getSubGrid()).getResolSize();

    for(int xyz = 0; xyz < 3; xyz++)
    {
	if(resolSize.get(xyz) <= 1)
	{
	    // 1 ξˤ grid ʤΤѤäƤʤ
	    cout << "Error! : isVolume :: This volume has no volume. "
	    "resolSize[" << xyz << "] = " << resolSize << endl;
	    return(NO);
	}
    }
    return(YES);
}

/**------------------------------------------------------------
 * volume  volume Ĵ٤
 * 	 ѤĤ
 *  	 fill Ƥ뤫 (= Ƥ 0 ʾǤ뤫)
 *
 *    @param Ĵ٤ volume Υꥹ
 *-----------------------------------------------------------*/
static void checkAllVolumes(VolumeSetClass * pVolSet)
{
    const int totalVolNum = pVolSet->getTotalLocalSpaceNum();
    for(int i = 0; i < totalVolNum; i++)
    {
	SLList < VolumeClass * > *pVolList = pVolSet->getListAs1DimArray(i);
	for(Pix p = pVolList->first(); p != NULL; pVolList->next(p))
	{
	    if(isVolume((*pVolList)(p)) == NO)
	    {
		abort();
	    }
	    if(((*pVolList)(p)->isAllFill()) == NO)
	    {
		abort();
	    }
	}
    }
}

/**-----------------------------------------------------------
 * volume ΥꥹȤ餽ΥꥹȤ° volume  subgridinfo 
 * 롥 
 * ΤȤSubVolume  MinPos, MaxPos  grid index 
 *------------------------------------------------------------*/
static SubGridInfoClass 
getSubGridInfoFromVolList(SLList < VolumeClass * > *pSameSpIdVlist)
{
    //  spcpos ˤ grid ϰϤ
    Pix firstPix   = pSameSpIdVlist->first();	
    const IntVec3 spcpos = (*pSameSpIdVlist)(firstPix)->getSubVolumeSpcPos();
    const int volumeID   = (*pSameSpIdVlist)(firstPix)->getID();

    int gridMin[3], gridMax[3];
    gridMin[0] = gridMin[1] = gridMin[2] = INT_MAX;
    gridMax[0] = gridMax[1] = gridMax[2] = -1;
    Pix pix;
    for(pix = pSameSpIdVlist->first(); pix != NULL; pSameSpIdVlist->next(pix))
    {
	// check : spcpos, volumeID ƱǤɬפ
	if(spcpos != (*pSameSpIdVlist)(pix)->getSubVolumeSpcPos())
	{			
	    cout << "Error! : inconsistent spcpos." << endl;
	    commMpi2_exit(1);
	}
	if(volumeID != (*pSameSpIdVlist)(pix)->getID())
	{
	    cout << "Error! : inconsistent volume ID." << endl;
	    commMpi2_exit(1);
	}

	// get grid max, grid min
	for(int xyz = 0; xyz < 3; xyz++)
	{
	    if((*pSameSpIdVlist)(pix)->getGridMinIdx(xyz) < gridMin[xyz])
	    {
		gridMin[xyz] = (*pSameSpIdVlist)(pix)->getGridMinIdx(xyz);
	    }
	    if(gridMax[xyz] < (*pSameSpIdVlist)(pix)->getGridMaxIdx(xyz))
	    {
		gridMax[xyz] = (*pSameSpIdVlist)(pix)->getGridMaxIdx(xyz);
	    }
	}
    }
    GlobalGridInfoClass gginfo = (*pSameSpIdVlist)(firstPix)->getGlobalGrid();

    Vector3 minpos = 
	getPosition(gginfo.getMinPos(), 
		    gginfo.getVoxelEdgeLen(),
		    Vector3(double(gridMin[0]), double(gridMin[1]), 
			    double(gridMin[2])));
    Vector3 maxpos = 
	getPosition(gginfo.getMinPos(), 
		    gginfo.getVoxelEdgeLen(),
		    Vector3(double(gridMax[0]), double(gridMax[1]), 
			    double(gridMax[2])));

    SubGridInfoClass sginfo;
    sginfo.initAll(minpos, maxpos, spcpos, IntVec3(gridMin), IntVec3(gridMax));

    return(sginfo);
}

//------------------------------------------------------------
// initialize volume set
//	ʬۤ줿 volume Ϻڤʤᡢʬ֤Ȥ˺ƹ
//------------------------------------------------------------
// ʬֻɸƱ volume ID  volume  1 Ф
//      ʬֻɸ volume ID ˤäƤդ櫓
//	arg: Ʊ spcpos, volume ID  volume list
//------------------------------
void PEClass::
getSetOfSameSpcIDVolume(SLList < VolumeClass * > *pSameSpIdVlist,
			SLList < VolumeClass * > *pVolPtrListCopy)
{
    Pix firstPix   = pVolPtrListCopy->first();
    IntVec3 spcpos = (*pVolPtrListCopy)(firstPix)->getSubVolumeSpcPos();
    const int id   = (*pVolPtrListCopy)(firstPix)->getID();
    SLList < VolumeClass * > restVlist;

    for(Pix pix = pVolPtrListCopy->first(); pix != NULL;
	pVolPtrListCopy->next(pix))
    {
	VolumeClass *pVol = (*pVolPtrListCopy)(pix);
	if((pVol->getID() == id) && 
	   (pVol->getSubVolumeSpcPos() == spcpos))
	{
	    pSameSpIdVlist->append(pVol);
	}else
	{
	    restVlist.append(pVol);
	}
    }
    (*pVolPtrListCopy) = restVlist; // Ĥ᤹
}

//------------------------------------------------------------
// initialize volume set main
//	ʬۤ줿 volume Ϻڤʤᡢʬ֤Ȥ˺ƹ
//	ǤνϼȼΤ
//	׻Ѥ˽ˤϡ initializeVolumes() 
//------------------------------------------------------------
void PEClass::
initRecvVolumeSet(void)
{
    int subspcInPE[3];
    this->getHavingSubSpcNum(subspcInPE);
    IntVec3 localTotalSpcNum(subspcInPE[0], subspcInPE[1], subspcInPE[2]);
    this->mVolSet.init(localTotalSpcNum);
    
    //  volume  effective range ׻ϰϤ
    Pix pix;
    for(pix = this->mVolList.first(); pix != NULL; this->mVolList.next(pix))
    {
	IntVec3 spcpos = (this->mVolList(pix)->getSubGrid()).getSpcPos();
	Vector3 cellMinPos = this->mSpcParam.getMinPosOfSpcIdx(spcpos);
	Vector3 cellMaxPos = this->mSpcParam.getMaxPosOfSpcIdx(spcpos);

	if((this->mVolList(pix))->calcEffectiveGridRange(cellMinPos,cellMaxPos)
	   == NO)
	{
	    cout << "Error! : PEClass::initVolumeSet" 		  << endl;
	    this->mVolList(pix)->print();
	    cout << "Error! : Can not calc effective grid range." << endl;
	    this->exitprog();
	}
 	// DEBUG : (this->mVolList(pix))->print();
    }

    // ǡݥ󥿥ꥹȤΥԡΥԡФƿ
    // ʬν򤷤褦Ȼפäݥ󥿤礷Τʤɤ
    // Τǡԡˤϰ̣ʤ

    while(!(this->mVolList.empty()))
    {
	// ʬֻɸ volume ID ˤä volume դ櫓
	SLList < VolumeClass * > sameSpIdVlist; // Ʊ spcpos  volume 
	this->getSetOfSameSpcIDVolume(&sameSpIdVlist, &this->mVolList);
	
	// sameSpIdVlist Ʊ spcpos && Ʊ ID  volume ޤä
	// 礵
	
	// volume ʬ volume ξõ
	GlobalGridInfoClass gginfo = (sameSpIdVlist.front())->getGlobalGrid();
	SubGridInfoClass    sginfo = getSubGridInfoFromVolList(&sameSpIdVlist);
	
	VolumeClass *pUnifiedVol = new VolumeClass;
	pUnifiedVol->initGlobalGridInfo(gginfo);
	pUnifiedVol->initSubGridInfo(sginfo);

	// volume (pUnifiedVol)  pSubGrid 礷Ƥ
	while(!(sameSpIdVlist.empty()))
	{
	    VolumeClass *pSubGrid = sameSpIdVlist.remove_front();
	    assert(pSubGrid->isAllFill() == YES);
	    pUnifiedVol->addSubVolume(pSubGrid); 
	    delete pSubGrid;	// delete !
	}
	assert(pUnifiedVol->isAllFill() == YES);

	IntVec3 nowSpcpos = sginfo.getSpcPos();
	Vector3 cellMinPos = this->mSpcParam.getMinPosOfSpcIdx(nowSpcpos);
	Vector3 cellMaxPos = this->mSpcParam.getMaxPosOfSpcIdx(nowSpcpos);

	// grid ͭϰϤη׻
	if(pUnifiedVol->calcEffectiveGridRange(cellMinPos, cellMaxPos) == NO) 
	{
	    cout << "Error! : unified volume has something wrong." << endl;
	    this->exitprog();
	}
	
	// volume Ͽ (Ͻλ pUnifiedVol  delete ٤
	// ƤʤmVolSet ˤϥݥ󥿤ϤƤΤǤ 
	// delete Ǥʤ
	this->mVolSet.add(pUnifiedVol->getSubVolumeSpcPos(), pUnifiedVol);
    }
    // Ƥ volume ʲξ
    // 	 Ѥ
    // 	 fill Ƥ
    // 	 Ƥ 0 ʾǤ롥
    checkAllVolumes(&(this->mVolSet));
}
