//------------------------------------------------------------
// $Id: calcRadioFRG.cc,v 1.1.1.1 1999/11/25 06:26:09 yamauchi Exp $
//------------------------------------------------------------
//	calcRadioFRG.cc
//	calculation of First Ray Generation unit for radiosity 
//	1996/8/9
//	(C) Copyright YAMAUCHI Hitoshi  
//
//	쥤ȯ
//------------------------------------------------------------
#include "Option.h"
#include "FileInfo.h"
#include "hemicube.h"
#include "mpi2_funcs.h"
#include "ProfileMon.h"

// stand alone 
#ifdef FOR_STANDALONE
#include "StandAloneModeObjects.h"
#else
#include "PE.h"
#endif

//------------------------------------------------------------
// get unshot radiosity energy from a objList
//	Ϥ줿 object list ϢʤäƤ patch ̤ͥͥ륮֤
//------------------------------------------------------------
double getTotalUnshotRadiosityEnergy(SLList < objBaseClass * > *pObjList)
{
    double unshotEnergy = 0.0;
    for(Pix pix = pObjList->first(); pix != NULL; pObjList->next(pix))
    {
	objBaseClass *pObj = (*pObjList)(pix);	
	switch(pObj->who())
	{
	case TRIANGLE_OBJ_E:
	    unshotEnergy += pObj->getNonRadEnergy();	    
	    break;	    
	case SPHERE_OBJ_E:
	case NULL_OBJ_E:
	default:
	    break;
	}
    }
    return(unshotEnergy);
}

//------------------------------------------------------------
// initForRadiosity() 
//	- initialize for patches
//   ѥåνԤ
//   ̤ͥ饸ƥͥ륮֤
//------------------------------------------------------------
void PEClass::
initializePatches(void)
{
    if(this->mSysInfo.isRadiosityCalc() == NO)
    {
	if(Option::getInstance()->getVerboseMode() == YES)
	{
	    cout << genHd(VERBOSEHEAD, this->getMyRank()) << 
	    "radiosity calculation is off. skip radiosity stage." << endl;
	}
	this->mIsMultiPassStageOver = YES;
    }

    // ͤѥå򸡺塼Ͽ롥
    // κ̤ͥͥ륮̤׻
    Pix pix;
    for(pix = this->mObjList.first(); pix != NULL; this->mObjList.next(pix))
    {
	objBaseClass *pObj = this->mObjList(pix);
	if(pObj->who() == TRIANGLE_OBJ_E) // ѷѥåʳͤϤʤ
	{
	    if(pObj->getNonRadEnergy() >= this->mViewData.getMinRadEnergy())
	    {
		// ͥ륮ͤ礭äͥ塼
		assert(pObj->isInRadiationQueue() == NO);
		pObj->setInRadiationQueue();
		this->mRadPatchQueue.append(pObj);
	    }
	}
    }

    // make HemiCube for this PE
    this->mpHemiCube = 
	new hemiCubeClass(this->mViewData.getHemres(), this->mID);
    // ray pointer array for communicate HC <-> PE
    // ray μϤκ    
    this->mppRadiosityRays = new Ray *[this->mpHemiCube->getGenRayNum()];
    
    if(Option::getInstance()->getVerboseMode() == YES)
    {
	cout << genHd(VERBOSEHEAD, this->getMyRank()) << "init radpatches " <<
	" ( " << this->mRadPatchQueue.length() << " )" << ", HCGenRays/once = "
	<< this->mpHemiCube->getGenRayNum() << endl;
    }
}

//------------------------------------------------------------
// ͥѥåθ椫Υͥ륮ΤΤ򤷡ͤ
//------------------------------------------------------------
void PEClass::
radiateRadFromMaxRadPatch(void)
{
    double maxEnergy = 0.0;
    objBaseClass *pMaxEnergyPatch = NULL;
    
    // list.length() ŤʤΤ search ĤǤĹפ
    // DLList ˤ length Ͽ褦ˤƤʤΤɬס
    // ϺΤΤ search ƤФᡢqueue Τ褦
    // ƬΤǤϤʤᡣɤϲǽʤϤ
    int queueLen = 0;	

    Pix pidx;
    Pix delpidx = NULL;
    for(pidx = this->mRadPatchQueue.first(); 
	pidx != NULL;
	this->mRadPatchQueue.next(pidx))
    {
	if((this->mRadPatchQueue(pidx))->getNonRadEnergy() > maxEnergy)
	{
	    maxEnergy = (this->mRadPatchQueue(pidx))->getNonRadEnergy();
	    pMaxEnergyPatch = this->mRadPatchQueue(pidx);
	    delpidx = pidx;
	}
	queueLen++;
    }
    // δؿϥѥå塼Ĺ > 0 ǸƤ֤ɬͥѥå¸ߤ
    assert((pMaxEnergyPatch != NULL) && (delpidx != NULL)); 
    this->mRadPatchQueue.del(delpidx, 1); // 1 (meanless here) see libg++ info

    double nonRadEnergyOfThePatch = pMaxEnergyPatch->getNonRadEnergy();

    this->mpHemiCube->genRays(pMaxEnergyPatch, 
			      this->mSysInfo.isDistributedRadiation(),
			      this->mppRadiosityRays);

    int i;
    const int raynum = this->mpHemiCube->getGenRayNum();
    for(i = 0; i < raynum; i++)
    {
	this->mRayPtrQueue.enq(this->mppRadiosityRays[i]);
	//  version ǤϼΤϤΤɬפä 
	// pointer ԡƤɤ 
	// delete this->mpRadiosityRays[i];
	// ݥ󥿤ϴ˥ꥹȤ˥ԡ졢Τؤݥ
	//  2 ĤˤʤäΤڤäƤ 
	this->mppRadiosityRays[i] = NULL;
    }

    //------------------------------
    // Hemi Cube 򤷤뤹
    //------------------------------
    this->mGStat.incHCGenNum();
    if(FileInfo::getInstance()->isRadioOutAtStep() == YES)
    {
	this->outRadiosityIntermediateResult();
    }

    // Profiling Monitor
    double unshotRadEnergy = -1.0;
    if(FileInfo::getInstance()->isTraceOutWithUnshotRadiosity() == YES)
    {
	unshotRadEnergy = 
	    getTotalUnshotRadiosityEnergy(&(this->mObjList));
    }
    ProfileMon::getInstance()->outHemiCubeInfo(queueLen, unshotRadEnergy);

    // verbose, debug mode
    if(Option::getInstance()->getVerboseMode() == YES)
    {
	cout << genHd(VERBOSEHEAD, this->getMyRank()) << "make HC : "
		"radpat nonrad enegy " << nonRadEnergyOfThePatch <<
		", rest nonradpatches " << queueLen << endl; 
	if(Option::getInstance()->getDebugMode() == YES)
	{
	    if(unshotRadEnergy < 0.0)
	    {
		unshotRadEnergy = 
		    getTotalUnshotRadiosityEnergy(&(this->mObjList));
	    }
	    cout << genHd(DEBUGHEAD, this->getMyRank()) << "rest of total "
	    	"energy of this PE " << unshotRadEnergy << endl;
	}
    }
}

void PEClass::
outRadiosityIntermediateResult(void)
{
    if(FileInfo::getInstance()->isInLoop() == YES)
    {
	if(FileInfo::getInstance()->getLoopVarValue() == 
	   this->mGStat.getHCGenNum())
	{
	    char fnamebuf[256];
	    static int output_counter = 0;
	    // output radiosity file 
	    switch(FileInfo::getInstance()->getRadioIntermFileType())
	    {
	    case 0:		// shot radiosity
		sprintf(fnamebuf, "%s.shot.%03d",   
			(const char*)(FileInfo::getInstance()->
				      getRadioIntermediateBaseName()),
			output_counter);
		this->outRadiosityMpi2Data(fnamebuf, YES);
		break;
	    case 1:		// unshot radiosity
		sprintf(fnamebuf, "%s.unshot.%03d", 
			(const char*)(FileInfo::getInstance()->
				      getRadioIntermediateBaseName()),
			output_counter);
		this->outRadiosityMpi2Data(fnamebuf, NO);
		break;
	    case 2:		// both
		sprintf(fnamebuf, "%s.shot.%03d",   
			(const char*)(FileInfo::getInstance()->
				      getRadioIntermediateBaseName()),
			output_counter);
		this->outRadiosityMpi2Data(fnamebuf, YES);
		sprintf(fnamebuf, "%s.unshot.%03d", 
			(const char*)(FileInfo::getInstance()->
				      getRadioIntermediateBaseName()),
			output_counter);
		this->outRadiosityMpi2Data(fnamebuf, NO);
		break;
	    default:
		cout << "Internal Error! : illegal intermediate radio "
		"out type" << endl;
		this->exitprog();
		break;
	    }
	    output_counter++;
	    // next loop
	    FileInfo::getInstance()->nextLoopVar();
	}
    }
}    
