//------------------------------------------------------------
// $Id: calcRayTrFRG.cc,v 1.6.2.1 2000/06/14 15:00:10 yamauchi Exp $
//------------------------------------------------------------
//	calcRayTrFRG.cc
//	calculation of First Ray Generation unit for ray tracing 
//	1996/8/2 
//	(C) 1995-1998 Copyright YAMAUCHI Hitoshi  
//
//	꡼ȻƤ
//
//	쥤ȯ
//	
//	PEClass ΥФˤǤ뤬ƥȤǽˤ뤿ʬ
//	Υ 
//------------------------------------------------------------
#include <assert.h>
#include <stdlib.h>
#include <time.h>

#include "Option.h"
#include "rayTrFRGUnit.h"
#include "screenPointer.h"
#include "mpi2_funcs.h"

// stand alone 
#ifdef FOR_STANDALONE
#  include "StandAloneModeObjects.h"
#else
#  include "ProcObj.h"
#endif			// #ifdef FOR_STANDALONE

//------------------------------
// random
//------------------------------
static void initRandom(void)
{
    srand48(time(NULL));
}

// return random value [0.0. 1.0)
static inline double getDRandom(void)
{
    return(drand48());
}

//------------------------------------------------------------
// rayTrFRGUnitClass 
//------------------------------------------------------------
// screen pointer (ɤΤ褦ʽǤɤ pixel ؤ strategy)
// νȤȼ member ν
// ȤСmVpSpcPos ʤɤϻ֤򼨤
// ϽʬԤɤ˿äƤΤǤǽ롥
void rayTrFRGUnitClass::
initForScreenPointer(int scrSizeX, int scrSizeY, int myRank, int totalPENum)
{
    // ΰ֤֤ȽǤ
    if(isPosInBox(this->mViewPoint, this->mDefSpcMin, this->mDefSpcMax) == YES)
    {
	// ¸ߤ: ¸ߤְ֤Τ
	this->mIsVpInDefSpc	= YES;
	this->mVpSpcPos 	= this->getSpcposOfaPoint(this->mViewPoint);
    }else
    {
	assert(this->mIsVpInDefSpc == NO);
    }

    // ɤνȯ method Ȥˤ screen pointer 
    switch(this->mScanMethodOnScreen)
    {
    case AllScan_frgm:		// ΤΥ (ʬϤʤ)
	this->initForAllScanMethod(scrSizeX, scrSizeY);
	break;

    case ScanlineScan_frgm:	// 饤󤴤Ȥʬ
	this->initForScanlineScanMethod(scrSizeX, scrSizeY, 
					myRank, totalPENum);
	break;
    case DistributedScan_frgm:	// distributed ʬ 
	this->initForDistributedScanMethod(scrSizeX, scrSizeY,
					   myRank, totalPENum);
	break;
    default:
	cout << "Internal Error! : illegal ray generation method in "
	    "ray-tracing." << endl;
	commMpi2_exit(1);
    }
}


//------------------------------
// init for AllScan method
//------------------------------
void rayTrFRGUnitClass::
initForAllScanMethod(int scrSizeX, int scrSizeY)
{
    assert(this->mScanMethodOnScreen == AllScan_frgm);
    this->mpScreenPtr = new scrPtrAllScan(scrSizeX, scrSizeY);

    // AllScan ξˤϻ֤ˤȯ ray ô PE 
    // 롥 
    if(this->mIsVpInDefSpc == YES)
    {
	assert(isPosInBox(this->mViewPoint, this->mDefSpcMin,
			  this->mDefSpcMax) == YES);
	//  PE ϻ¸ߤ֤ôƤ뤫? 
	if(GpPE->isDealSpace(this->mVpSpcPos) == NO)
	{
	    // ˤ뤬¸ߤʬ֤򤳤 
	    // PE ôƤʤΤ ray ȯɬפʤ 
	    this->mCanGenerateNextRay = NO; // ray ȯλȤ
	    if(Option::getInstance()->getDebugMode() == YES)
	    {
		cout << genHd(DEBUGHEAD, this->mPEid) << "generate 0 rays\n" 
		"Because view point is in a define space and this PE "
		"does not deal subspace within a view point" << endl;
	    }	    
	}else
	{
	    if(Option::getInstance()->getDebugMode() == YES)
	    {
		cout << genHd(DEBUGHEAD, this->mPEid) << "AllScan : "
		"view point is in defined space and this PE generate rays."
		<< endl;
	    }
	}
    }else
    {
	assert(isPosInBox(this->mViewPoint, this->mDefSpcMin,
			  this->mDefSpcMax) == NO);
    }
}


void rayTrFRGUnitClass::
initForScanlineScanMethod(int scrSizeX, int scrSizeY, 
			  int myRank, int totalPEnum)
{
    // myRank  PE 0  PE n-1 ȤʤäƤϤ
    assert(myRank < totalPEnum);
    assert(myRank >= 0);
    this->mpScreenPtr = new scrPtrScanline(scrSizeX, scrSizeY, 
					   totalPEnum, myRank); 
    cout << "DEBUG: GenScreen ScanLine Pointer : myRank is " << 
	myRank << ", skip = " << totalPEnum << endl;
    
}

void rayTrFRGUnitClass::
initForDistributedScanMethod(int scrSizeX, int scrSizeY,
			     int myRank, int totalPEnum)
{
    //  block size ѲʤѲ : ̤
    this->mpScreenPtr = new scrPtrDistribute(scrSizeX, scrSizeY,
					     totalPEnum, myRank); 
}

//------------------------------
// ꡼ pixel ñ̤ΰư
// pixel Ǥλ ray ȯ鼡 pixel ذư
// ޤ ray ȯǤΤʤФΤޤ
//------------------------------
inline void rayTrFRGUnitClass::
moveOnScreen(void)
{
    this->mGenRayParPixel++;
    // 1 pixel ˤȯۤ? loop δϤ
    if(this->mGenRayParPixel >= this->mMaxRaysInAPixel)
    {
	this->mGenRayParPixel = 0;
	// this->mpScreenPtr->next ˤäưư롥
	// ʾưǤʤˤϤ򵭤
	this->mCanGenerateNextRay = this->mpScreenPtr->next();
	this->mpScreenPtr->getpos(&(this->mScrX), &(this->mScrY));
    }else
    {
	// ޤ pixel  ray ȯǽ
    }
}

//------------------------------
// , point ¸ߤʬ֤λɸΰ֤
//
// cp = cell periad, p = point, min, max = partition volume min, max
// k  sub space pos Ȥ
// min + k \frac{max - min}{cp} <= p < min + (k + 1)\frac{max - min}{cp}
//   
// \frac{cp (p - min)}{max - min} - 1 < k <= \frac{cp (p - min)}{max - min}
//  Ȥʤ
// \frac{cp}{max - min} = partVolOneSize
//  (partVolOneSize * (p - min)) - 1 
//------------------------------
IntVec3 rayTrFRGUnitClass::
getSpcposOfaPoint(const Vector3& point) const
{
    // partition volume Фа
    Vector3 relativePos = point - this->mPartVolMinPos;
    int i, spos[3];		// space position
    for(i = 0; i < 3; i++)
    {
	// ֤֤礭ǳꡢشݤ
	spos[i] = (int)(floor(relativePos.get(i) / this->mPartVolOneSize));
	assert(spos[i] >= 0);

	if(spos[i] == this->mSubDivNum)
	{
	    if((double)(int(relativePos.get(i) / this->mPartVolOneSize)) ==
	       (relativePos.get(i) / this->mPartVolOneSize))
	    {
		// ֤礭ü point  
		spos[i]--;
	    }else
	    {
	 	cout << "Internal Error! : Is this pos in partition volume?"
		     << endl;
	    }
	}
    }
    return(IntVec3(spos[0], spos[1], spos[2]));
}


//------------------------------
// view vector generator
//------------------------------
inline Vector3 rayTrFRGUnitClass::
genViewVec(void) const
{
    //------------------------------
    // AllScan ξˤϳʻҾΥꥢͤȯ뤬
    // ưΤߤǤꡤ1PE ˤư롥
    // ͤ AllScan mode ư˼ˤ pixel ô
    //  PE ֤Ƕͭ뤢 pixel ˤƱΥץ󥰤Ԥ
    // ʤƤϤʤʤͭˤ 1 Ĥ PE ץ󥰤ơ
    // ô PE ؤ ray Ȥʤɤ뤬 AllScan 
    // mode ͭ̿Υɤ줿ʤΤ AllScan ư
    // ԹϤΤޤ޻ĤƤ 
    //------------------------------
    Vector3 viewVec = 
    	this->mScreenLeftUpFromVp + 
 	((((double)(this->mScrX)) + getDRandom()) * this->mDeltaXscr) +
 	((((double)(this->mScrY)) + getDRandom()) * this->mDeltaYscr);
    viewVec.VNormalize();
    return(viewVec);
}

//------------------------------
// ˤ ray ȯ 
// 1  ray ȯ index ʤƤ
// ƽλ餽Υե饰ΩƤ롣
//      ݻƤȯ쥤ΰ֤ϰưʤ
//------------------------------
Ray *rayTrFRGUnitClass::
newOneRayVpInDefineSpc(void) const
{
    assert(this->mCanGenerateNextRay == YES);
    assert(this->mIsVpInDefSpc == YES);

    // ľ genViewVec() 񤤤ƤɤȻפä԰¤롥
    Vector3 distPixelVec = this->genViewVec();

    assert(fabs((distPixelVec.getNorm()) - 1.0) < D_ERROR_RES);    

    // ˤΤɬ ray ˤ롥
    Ray *pRay = Ray::newForRayTrace(this->mScrX, this->mScrY,
				    this->mViewPoint, 
				    distPixelVec, // ľ genViewVec()?
				    this->mVpSpcPos,
				    this->mDistRaysFactor); //, PEid;
    return(pRay);
}

//------------------------------
// ֤γˤ ray ȯ
//------------------------------
Ray *rayTrFRGUnitClass::
newOneRayVpOutOfDefSpc(void) const
{
    assert(this->mCanGenerateNextRay == YES);
    assert(this->mIsVpInDefSpc == NO);

    Vector3 viewVec = this->genViewVec();

    assert(fabs((viewVec.getNorm()) - 1.0) < D_ERROR_RES);    

    // ֤ȤθȽ 
    // äˤϤɤʬ֤ˤä׻
    static Vector3 hitpos;
    DIR_3D hitSide = isHitBoundingBox(this->mDefSpcMin, this->mDefSpcMax,
				      this->mViewPoint, viewVec,
				      &hitpos);
    if(hitSide != DIR_ILLEGAL)
    {
	// ray ֤ hit 
	static IntVec3 hitSubSpcPos; 
	hitSubSpcPos = this->getSpcposOfaPoint(hitpos);
	
	switch(this->getScanMethodOnScreen())
	{
	case AllScan_frgm:
	    // AllScan ʬԤʤ
	    // ray  PE ô֤ˤʤˤϵѲ
	    if(GpPE->isDealSpace(hitSubSpcPos) == YES) // ƻ:ädirty
	    {
		Ray *pRay = Ray::newForRayTrace(this->mScrX, this->mScrY, 
						this->mViewPoint, viewVec,
						hitSubSpcPos, 
						this->mDistRaysFactor); 
		return(pRay);
	    }else 
	    { 
		return(NULL);	// ֤ˤhitô֤ǤʤѲ
	    }
	    break;
	    
	case ScanlineScan_frgm:
	case DistributedScan_frgm:
	{
	    // ʬ
	    // ꡼ʬ䤵졤ƥ꡼󤬥Ρɤʬۤ
	    // ΤǡɬƥΡɤ ray ȯ
	    //  ray Ϥζ֤ˤȤϸ¤ʤ
	    Ray *pRay = Ray::newForRayTrace(this->mScrX, this->mScrY, 
					    this->mViewPoint, viewVec,
					    hitSubSpcPos, 
					    this->mDistRaysFactor); //,PEid
	    return(pRay);		
	    break;
	}
	default:
	    cout << "Internal Error! : No such screen scan method." << endl;
	    commMpi2_exit(1);
	    return(NULL);
	}
    }else
    {
	// ֤ hit ʤΤȯɬפϤʤ
	return(NULL);
    }
    cout << "never reach. rayTrFRGUnitClass::newOneRayVpOutOfDefSpc()" << endl;
    commMpi2_exit(1);
    return(NULL);
}

// reference ʻҾΥꥢ󥰥Х
// [0.0, 1.0)  deltaX, deltaY ˳ݤԥ
//  random ˥ץФޤ
//	distPixelVec = pixelVecFromVpBase + 
//	    (getDRandom() * deltaX) + (getDRandom() * deltaY);
// ξˤ ray Ȥô뤫ɤȽ
// ꤷƤ뤿ᡢǰʤ֤ζǤϤ1Ĥ
// ԥ2Ĥ PE ƱôƤޤȹͤꡢ
// ɤ PE ôʤȤȤȯƳʻҾ
// Υꥢ󥰤ƤޤΤᡢ쥤§
// Ūȯ뤳Ȥˤ 

//------------------------------------------------------------
// public functions
//------------------------------------------------------------
// constructor
//------------------------------
rayTrFRGUnitClass::
rayTrFRGUnitClass(void)
{
    Vector3 illegalV(-1.0, -2.0, -3.0);
    IntVec3 illegalIntV(-1, -2, -3);

    this->mScanMethodOnScreen   = ELSE_frgm;
    
    this->mPEid 		= illegalIntV;
    this->mViewPoint		= illegalV;
    this->mpScreenPtr		= NULL;

    this->mScreenLeftUpFromVp	= illegalV;
    this->mDeltaXscr		= illegalV;
    this->mDeltaYscr		= illegalV;

    this->mMaxRaysInAPixel 	= -1;
    this->mGenRayParPixel	= -1;

    this->mCanGenerateNextRay	= NO; // ޤǤ next rays Ϥʤ


    this->mDefSpcMin		= illegalV;		
    this->mDefSpcMax		= illegalV;
    this->mPartVolMinPos	= illegalV;	
    
    this->mSubDivNum		= 0.0;	
    this->mPartVolOneSize	= -1.0;

    this->mVpSpcPos   		= illegalIntV;		
    this->mIsVpInDefSpc		= NO;
    this->mNumOfTotalGenRays	= -1;
    this->mDistRaysFactor	= 0.0;

    this->mScrX = -10;
    this->mScrY = -11;
}

//------------------------------
// initializer
//  @param rayGenerationMethod : ʬ ray ˡ(Allscan  ScanlineScan)
//  @param viewDat	: ΰ֡ʤɤλ
//  @param spcDat	: ־
//  @param PEid	: νȯ˥åȤ° PE  ID
//  @param totalPEnum : ƥ PE  (ʬɬ)
//  @param myRank 	: ʬ rank (0 Ϥޤ)
//------------------------------
void rayTrFRGUnitClass::
init(int 			rayGenerationMethod,
     const ViewDataClass& 	viewDat, 
     const SpaceParam& 		spcDat, 
     const IntVec3& 		PEid,
     int   			totalPEnum,
     int 			myRank // myRank  PE  0 Ϥޤ
     )
{
    // ˴ؤѥ᡼
    Vector3 vp  = viewDat.getViewPoint(); 	// ٥ȥ
    Vector3 vv  = viewDat.getViewVector();	// ٥ȥ
    Vector3 vup = viewDat.getViewUpVector();	// ꡼٥ȥ
    double  vFovRad, vHovRad;
    int     screenSizeX, screenSizeY;
    viewDat.getFOV(&vFovRad, &vHovRad);
    viewDat.getScreenSize(&screenSizeX, &screenSizeY);

    // ꡼ľñ̥٥ȥ (ñ̥٥ȥΤϤǰΤ)
    vup.VNormalize();			     
    vv.VNormalize();			     
    
    // ꡼ʿñ̥٥ȥ(vSh) $vSh = vv  \times vup$
    Vector3 vScreenHoriz = vv.VOut_prd(vup); 
    if(vScreenHoriz.getNorm() <= 1e-4)
    {
	cout << NORMALHEAD << " Warning! : maybe view up vector is illegal. "
	     << "Generated image will not be right." << endl;
    }

    vScreenHoriz.VNormalize();		     // $vSh = \frac{vSh}{|vSh|}$
    const double  halfX  = tan(vHovRad);
    const double  halfY  = tan(vFovRad);
    // = vp + .... - vp ʲμ˲äƤ,Ȼ vec
    Vector3 screenLeftUpFromVp
	= vv - (halfX * vScreenHoriz) + (halfY * vup);
    const Vector3 deltaX = 
	((2.0 * halfX) / double(screenSizeX)) * vScreenHoriz;
    const Vector3 deltaY = ((-2.0 * halfY) / double(screenSizeY)) * vup;
    
    initRandom();		// distributed ray-tracing Τ

    // initialize member
    this->mScanMethodOnScreen 	= rayGenerationMethod;
    this->mPEid	 		= PEid;
    this->mViewPoint		= vp;
    this->mpScreenPtr		= NULL;	//  initForScreenPointer ˤ
    this->mScreenLeftUpFromVp	= screenLeftUpFromVp;
    this->mDeltaXscr		= deltaX;
    this->mDeltaYscr		= deltaY;
    this->mMaxRaysInAPixel	= viewDat.getDistributedRayNum();
    assert(this->mMaxRaysInAPixel > 0);
    this->mGenRayParPixel	= 0;

    this->mCanGenerateNextRay	= YES; // ޤ ray ȯƤʤ

    // [־] ְ֤뤿ν 
    this->mDefSpcMin 	  = spcDat.getAlignedDefSpaceMin();//֤Ȥ
    this->mDefSpcMax 	  = spcDat.getAlignedDefSpaceMax();
    this->mPartVolMinPos  = spcDat.getPartVolumeMin();
    Vector3 partVolMaxPos = spcDat.getPartVolumeMax();

    // partition volume ν
    // partition volume ΩΤǤΤǰդ狼ɤ 
    this->mSubDivNum	  = (double)(spcDat.getSubDivNum());

    this->mPartVolOneSize = 
      (partVolMaxPos.get(Vector3::X) - (this->mPartVolMinPos).get(Vector3::X))
         / this->mSubDivNum;

    // initialize parameter for distributed ray-trace 
    this->mNumOfTotalGenRays	= 0;	//  ray ȯ

    // mDistRaysFactor :  ray ε٤ distributed Υ쥤οʬ 1  
    // νŤߤġΤ coefficient value ˤͤݤƤ
    // 4 ܤ distributed ʤг ray  1 Ĥ pixel ؤδͿ 1/4 
    this->mDistRaysFactor = 1.0 / double(this->mMaxRaysInAPixel);
    
    this->mScrX = 0;
    this->mScrY = 0;

    // screen pointer ν
    // 	this->mIsVpInDefSpc, this->mVpSpcPos ν
    this->initForScreenPointer(screenSizeX, screenSizeY,  myRank, totalPEnum);
}


//------------------------------
// ray ȯ
// 	ʬˡŬѻˤϡȯ ray ʬֹ椬ȯ
// 	 PE ôǤʤ礬롥
//------------------------------
Ray *rayTrFRGUnitClass::
newNextRay(void)
{
    // δؿƤ˰ܤȯǽǤ뤳ȤĴ٤ƤϤ
    assert(this->canGenerateNextRay() == YES);
    Ray *pRay = NULL;
    if(this->mIsVpInDefSpc == YES)
    {
	// ֤Τɤˤ뤫طΤϽʬʤ All
	// scan λΤ
	assert(this->getScanMethodOnScreen() == AllScan_frgm);

	// ¸ߤ
	// ξˤϤ pixel ֤ ray ȯǤʤȤ
	// ϤʤΤǡʬˤ餺ɬ1ܤ ray ȯ
	pRay = this->newOneRayVpInDefineSpc();
	this->moveOnScreen();	// ȯ֤ذư
	assert(pRay != NULL);
	this->mNumOfTotalGenRays++; // ȯ ray 򥫥Ȥ
    }else
    {				
	// ˤϤʤ
	// ȯǽ֤ǤäƤ⸽ߤ screen ֤Ǥ PE  ray ȯ
	// ô뤫ɤʤΤǡ1 ȯޤ screen 
	//  pixel ưƤߤ롣
	// Ĥޤޤ꡼ǽ(canGenerateNextRay() == YES) 
	// Ǥĸߤ pixel  PE ôǤʤˤϼȥ꡼
	// Ƥߤ 
	while(this->canGenerateNextRay() == YES)
	{
	    //  pixel ȯǽ? 
	    pRay = this->newOneRayVpOutOfDefSpc();
	    this->moveOnScreen();	// ȯ֤ذư
	    if(pRay != NULL)
	    {   
		this->mNumOfTotalGenRays++; // ȯ ray 򥫥Ȥ
		break;		// ray ȯ
	    }
	}
    }

    // ʬˤ ray  spcpos ºݤôƤ 
    // PE ν ray ¸ߤʤǽ롥
    return(pRay);
}

//#if 0
//------------------------------
// Ƚ bounding box 
//	Graphics GEMS p.p.736
//  bounding box ˻ȡɬ Hit ΤǤȽ
// λϻȤ롣
// ˡǤϸ褦Ǥ롣ޤǸȽ롼
// Ǥ롣 
//------------------------------
// enum quadrant_e {
//     RIGHT,
//     LEFT,
//     MIDDLE			// xlC Τ,򳰤ʤƤϤʤ.Ҥɤ
// 
// };
// 
// YorN isHitBoundingBox(const Vector3& minBox,    const Vector3& maxBox, 
// 		      const Vector3& rayOrigin, const Vector3& rayDir,
// 		      Vector3 *pHitPoint)
// {
//     const int NUMDIM = 3;	// 3-dimension
//     enum quadrant_e quadrant[NUMDIM]; 
//     int i;
//     double candidatePlane[NUMDIM];
//     double coord[NUMDIM];	// hit point
// 	
//     // find candidate planes. This loop can avoided is rays cast all
//     // from the eye (assume perpsective view)  
//     YorN isInside = YES;
//     for(i = 0; i < NUMDIM; i++)
//     {
// 	if(rayOrigin.get(i) < minBox.get(i)) // ȢκǾͤ꾮
// 	{
// 	    quadrant[i] 	= LEFT;
// 	    candidatePlane[i] 	= minBox.get(i);
// 	    isInside		= NO;
// 	}else if(rayOrigin.get(i) > maxBox.get(i)) // Ȣκͤ
// 	{
// 	    quadrant[i] 	= RIGHT;
// 	    candidatePlane[i] 	= minBox.get(i);
// 	    isInside		= NO;
// 	}else
// 	{
// 	    quadrant[i] 	= MIDDLE;	    
// 	    // isInside 	= YES; ͤ YES
// 	}
//     }
//     
//     // ray origin inside bounding box
//     if(isInside == YES)
//     {
// 	(*pHitPoint) = rayOrigin;
// 	return(YES);
//     }
//     
//     // calculate T distance to candidate planes
//     double distT[NUMDIM];
//     for(i = 0; i < NUMDIM; i++)
//     {
// 	// γƺɸ bounding ext.Ǥʤ٥ȥʬ
// 	if((quadrant[i] != MIDDLE) && (rayDir.get(i) != 0.0))
// 	{
// 	    distT[i] = (candidatePlane[i] - rayOrigin.get(i)) / rayDir.get(i);
// 	}else
// 	{
// 	    distT[i] = -1.0;
// 	}
//     }
//     
//     // get largest of the distT's for final choice of intersection
//     int whichPlane = 0;
//     for(i = 1; i < NUMDIM; i++)
//     {
// 	if(distT[whichPlane] < distT[i])
// 	{
// 	    whichPlane = i;
// 	}
//     }
//     
//     // chech final candidate actually inside box
//     if(distT[whichPlane] < 0.0)
//     {
// 	return(NO);
//     }
//     
//     for(i = 0; i < NUMDIM; i++)
//     {
// 	coord[i] = 
// 	    rayOrigin.get(i) + distT[whichPlane] * rayDir.get(i);
// 	if(whichPlane != i)
// 	{
// 	    if(((quadrant[i] == RIGHT) && (coord[i] < minBox.get(i))) || 
// 	       ((quadrant[i] == LEFT)  && (coord[i] > maxBox.get(i))))
// 	    {
// 		return(NO);	// outside box
// 	    }else
// 	    {
// 		coord[i] = candidatePlane[i];
// 	    }
// 	}
//     }
//     pHitPoint->set(coord[0], coord[1], coord[2]);
//     return(YES);
// }
//#endif
//
// #endif				// #ifndef TEST_isHitBoundingBox
// #ifdef TEST_isHitBoundingBox
// main()
// {
//     Vector3 minB( 0,  0,  0);
//     Vector3 maxB(32, 32, 32);
//     Vector3 rayOrg(16,16,-10);    
//     Vector3 rayDir( 0, 0, 1);
//     Vector3 hitpos(-1, -2, -3);
// 
//     DIR_3D dir = isHitBoundingBox(minB, maxB, rayOrg, rayDir, &hitpos);
//     cout << "minB(" << minB   << "), maxB(" << maxB     << ")" << endl;
//     cout << "rayO(" << rayOrg << "), rayDir(" << rayDir << ")" << endl;
//     cout << "hit(" << hitpos << ")" << endl;
// }
// g++ -g -DTEST_isHitBoundingBox calcRayTrFRG.cc Vector3.o IntVec3.o -lm
// #endif				// #ifdef TEST_isHitBoundingBox


#ifdef TEST_FRG
void main()
{
//     rayTrFRGUnitClass 	frgUnit;
//     ViewDataClass   	viewDat; 
//     SpaceParam 	spcDat;
//    IntVec3		PEid(0, 0, 1);

//    frgUnit.init(viewDat, spcDat, PEid);

    while(frgUnit.canGenerateNextRay() == YES)
    {
	Ray *pRay = getTotalGenRays();
    }
}    
// g++ -g -DTEST_FRG calcRayTrFRG.cc Vector3.o IntVec3.o -lm
#endif				// #ifdef TEST_FRG
