/*
 * $Id: Mpi2CommunicatorMpiJavaS1.java,v 1.6.4.1 2000/06/15 20:41:23 yamauchi Exp $
 *     (M\pi)^2 : Interface of Mpi2 Communicator : 
 *		MPI version implementation 
 * 		Using mpiJava
 *	Copyright 1999 By YAMAUCHI Hitoshi   
 *	@author YAMAUCHI Hitoshi  
 *
 * original source mpi2communicationMPI.cc
 *
 * MPI ˤǤμ 
 * Object  Serialize  Java  Serialize Ǥ 
 *	mpiJava + Java Serializer
 *
 * 1999ǯ913()
 */

package jp.ac.tohoku.is.archi.mpi2.net;

/*
 * mpiJava
 */
import  mpi.*;

/*
 * 
 */
import  java.util.*;
import  java.io.*;

import  jp.ac.tohoku.is.archi.mpi2.common.*;
import  jp.ac.tohoku.is.archi.mpi2.obj.*;
import  jp.ac.tohoku.is.archi.mpi2.sysparam.*;
import  jp.ac.tohoku.is.archi.mpi2.util.*;
import  jp.ac.tohoku.is.archi.mpi2.host.*;

/**
 * (M\pi)^2 Ѥ̿ؿ interface
 */
public final class Mpi2CommunicatorMpiJavaS1
implements Mpi2Communicator, Mpi2Enum 
{
    private static final int OFFSET = 0;

    /**
     * get name : ɤʼ̿
     *	Parallel Version (Parallel | OnePE)
     */
    public String getName()
    {
	return("Parallel");
    }

    /**
     * initialize commun (M\pi)^2
     */
    public synchronized String [] init(String [] args)
    {
	try{
	    String [] retargs = MPI.Init(args);
	    int myrank = MPI.COMM_WORLD.Rank();
	    int size   = MPI.COMM_WORLD.Size();
	    int hostrank = size - 1;
	
	    (SysStat.getInstance()).setCommSize(size);
	    (SysStat.getInstance()).setMyRank(myrank);

	    // host Υ󥯤ϰֺǸΤΤˤ 
	    (SysStat.getInstance()).setHostRank(hostrank);

	    return(retargs);
	}catch(MPIException e)
	{
	    System.out.println(e);
	    return(args);
	}
    }
    
    public synchronized void finalizecommun() // finalize method Ϥ褯Τ
    {
	try{
	    MPI.Finalize();
	}
	catch(MPIException e)
	{
	    System.out.println(e);
	}
    }
    public synchronized void barrier()
    {
	try{
	    MPI.COMM_WORLD.Barrier(); 
	}
	catch(MPIException e)
	{
	    System.out.println(e);
	}
    }
    public synchronized void exit(int err_num)
    {
	try{
	    MPI.COMM_WORLD.Abort(err_num); 
	}
	catch(MPIException e)
	{
	    System.out.println(e);
	}
    }

    /**
     * get type
     */
    public int getCommunicatorType()
    {
	return(Mpi2Enum.MPI2COMM_mpiJavaS1);
    }

    /**
     * get identification
     */
    public String getDescription()
    {
	return("Mpi2CommunicatorMpiJavaS1 : MPI Parallel version using " + 
	       "mpiJava + java.io.Serialize");
    }


    /**
     * Sysinfo, ViewData, SpaceParam  serialize 
     */
    private int getSVSSize()
    {
	return(1523);
    }
    /**
     * sysinfo (Broadcast)
     *  PE ѤǤ object, material, sysinfo  receive Ϥʤ
     *  send Ʊ receive 
     */
    public synchronized void 
    broadSysinfo(SysInfo [] sys, ViewData [] view, SpaceParam [] spcparam)
    {
	int root = SysStat.getInstance().getHostRank();
	//     cout << "***** debug :: Broad casting sysinfo .. root rank  = " 
	// 	<< root << endl;
	Assert.out((root >= 0) && 
		   (root < SysStat.getInstance().getCommSize()),
		   "Error! : Mpi2CommunicatorMPI : root rank is illegal.");
	
	if(root == SysStat.getInstance().getMyRank())
	{
	    // ¦
	    try {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ObjectOutputStream oos     = new ObjectOutputStream(baos);
		oos.writeObject(sys[0]);
		oos.writeObject(view[0]);
		oos.writeObject(spcparam[0]);
		oos.flush();
		byte [] serializedobj = baos.toByteArray();
		oos.close();

		// ³Τ
		MPI.COMM_WORLD.Bcast(serializedobj, OFFSET, 
				     serializedobj.length, MPI.BYTE, root);

		Assert.out(serializedobj.length == getSVSSize(),
			   "Error! : SysInfo, ViewData, SpaceParam size " + 
			   "is changed to " + serializedobj.length);
	    }catch(Exception e)
	    {
		System.out.println("Exception : broadSysinfo send. " + e);
	    }
	}else
	{
	    // ¦
	    try {
		byte [] buf = new byte [getSVSSize()];
		MPI.COMM_WORLD.Bcast(buf, OFFSET, getSVSSize(), MPI.BYTE,root);
		ByteArrayInputStream bais = new ByteArrayInputStream(buf);
		ObjectInputStream ois     = new ObjectInputStream(bais);
		sys[0] 		= (SysInfo)ois.readObject();
		view[0] 	= (ViewData)ois.readObject();
		spcparam[0] 	= (SpaceParam)ois.readObject();
		ois.close();
		
	    }catch(Exception e)
	    {
		System.out.println("Exception : broadSysinfo recv. " + e);
	    }
	}
    }

    /**
     * ƶ֤ʪΤ¸ߤ뤫ɽ : ϸĹǤϤʤ
     */
    public synchronized 
    void broadObjectInSpaceMap(Array3D [] objectInSpaceMap)
    {
	int root = SysStat.getInstance().getHostRank();
	Assert.out((root >= 0) && 
		   (root < SysStat.getInstance().getCommSize()),
		   "Error! : Mpi2CommunicatorMPI : root rank is illegal.");
	
	if(root == SysStat.getInstance().getMyRank())
	{
	    // ¦
	    try {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ObjectOutputStream oos     = new ObjectOutputStream(baos);
		oos.writeObject(objectInSpaceMap[0]);
		oos.flush();
		byte [] serializedobj = baos.toByteArray();
		oos.close();

		// ޤ serialize Υ
		int [] size = new int[1];
		size[0] = serializedobj.length;
		MPI.COMM_WORLD.Bcast(size, OFFSET, 1, MPI.INT, root);
		
		// ³Τ
		MPI.COMM_WORLD.Bcast(serializedobj, OFFSET, 
				     serializedobj.length, MPI.BYTE, root);
	    }catch(Exception e)
	    {
		System.out.println("Exception : broadSysinfo send. " + e);
	    }
	}else
	{
	    // ¦
	    try {
		// ޤϼ
		int [] recv_size = new int[1];
		MPI.COMM_WORLD.Bcast(recv_size, OFFSET, 1, MPI.INT, root);

		// Τ
		byte [] serializedobj = new byte [recv_size[0]];
		MPI.COMM_WORLD.Bcast(serializedobj, OFFSET, recv_size[0],
				     MPI.BYTE, root);
		ByteArrayInputStream bais = 
		    new ByteArrayInputStream(serializedobj);
		ObjectInputStream ois = new ObjectInputStream(bais);
		objectInSpaceMap[0]   = (Array3D)ois.readObject();
		ois.close();		
	    }catch(Exception e)
	    {
		System.out.println("Exception : broadSysinfo recv. " + e);
	    }
	}
    }

    /**
     * ʪΤεΥɽ
     */
    // public void broadPRCSkipMap(int bytes, IMPLEMENT_TYPE pDatArray)

    /**
     * send, recv objects
     */
    /**
     * send objects
     * 	@param destRank 	ʪΥץߥƥ rank
     * 	@param objprim 		դʪΥץߥƥ
     */
    public synchronized 
    void sendObject(int destRank, ObjectPrimitive objprim)
    {
	try {
	    ByteArrayOutputStream baos = new ByteArrayOutputStream();
	    ObjectOutputStream oos     = new ObjectOutputStream(baos);
	    oos.writeObject(objprim);
	    oos.flush();
	    byte [] serializedobj = baos.toByteArray();
	    oos.close();

	    int tag = TAG_NULL_OBJ;
	    switch(objprim.who())
	    {
	    case ObjectPrimitive.OBJ_TRIANGLE:
		tag = TriangleObject.getTAG();
		break;
	    case ObjectPrimitive.OBJ_NULL:
		tag = NullObject.getTAG();
		break;
	    case ObjectPrimitive.OBJ_SPHERE:
		// NIN tag = SphereObject.getTAG();
	    default:
		String s = "Error! : send object has illegal tag. dest = "
		    + destRank;
		throw new RuntimeException(s);
	    }	

	    // size  : material ΥѲΤǥ
	    int [] size = new int[1];
	    size[0] = serializedobj.length;

	    Request reqSize = MPI.COMM_WORLD.Isend(size, OFFSET, 1, MPI.INT, 
						   destRank, TAG_INTEGER);
	    reqSize.Wait();

	    // 
	    Request reqObj = MPI.COMM_WORLD.Isend(serializedobj, OFFSET, 
						  serializedobj.length,
						  MPI.BYTE, destRank, tag);
	    reqObj.Wait();
	}catch(Exception e)
	{
	    System.out.println("Exception : sendObject send. " + e);
	}
    }	
    
    /**
     * recv objects
     * 	@param destRank 	ʪΥץߥƥո rank Ĺ 1 
     * 	@param objprim 		ʪΥץߥƥ֤
     */
    // private final static int TRIANGLE_OBJSIZE 	= 100;
    // private final static int NULL_OBJSIZE	= 200;
    // private boolean mIsFirstProbe = true;

    public synchronized void 
    recvObject(int [] srcRank,  ObjectPrimitive [] objprim)
    {
	final int hostrank = (SysStat.getInstance()).getHostRank();

	try{
	    /*
	     * recv size and allocate buffer
	     */
	    Status probeStat = null;
	    while(probeStat == null)
	    {
		probeStat = MPI.COMM_WORLD.Iprobe(MPI.ANY_SOURCE, TAG_INTEGER);
		if(probeStat == null)
		{
		    Mpi2System.pollingSleep();
		}
		/*
		  if(this.mIsFirstProbe == true)
		  {
		  // mpiJava ǽ probe ˼ԤΤǡǽ skip 
		  probeStat = null;
		  this.mIsFirstProbe = false;
		  }
		  */
	    }
	    int [] size = new int[1];
	    Request reqSize = MPI.COMM_WORLD.Irecv(size, OFFSET, 1, MPI.INT,
						   hostrank, TAG_INTEGER);
	    reqSize.Wait();

	    /*
	      switch(probeStat.tag)
	      {
	      case TAG_TRIANGLE_OBJ:
	      size = TRIANGLE_OBJSIZE;
	      break;
	      case TAG_SPHERE_OBJ:
	      // size = SPHERE_OBJSIZE;
	      break;
	      
	      case TAG_NULL_OBJ:
	      size = NULL_OBJSIZE;
	      break;
	      
	      default:
	      String s = "Error! : recv object gets illegal tag. tag = "
	      + probeStat.tag + " " + probeStat.source;
	      throw new RuntimeException(s);
	      }	
	      */
	    
	    byte [] buf = new byte[size[0]];
	    /*
	     * get tag
	     */
	    probeStat = null;
	    while(probeStat == null)
	    {
		probeStat = MPI.COMM_WORLD.Iprobe(hostrank, MPI.ANY_TAG);
		if(probeStat == null)
		{
		    Mpi2System.pollingSleep();
		}
	    }
	    
	    Request reqObj = MPI.COMM_WORLD.Irecv(buf, OFFSET, size[0], 
						  MPI.BYTE, 
						  hostrank, probeStat.tag);
	    reqObj.Wait();
	    
	    ByteArrayInputStream bais = new ByteArrayInputStream(buf);
	    ObjectInputStream ois     = new ObjectInputStream(bais);
	    /*
	     * now I get message
	     */
	    switch(probeStat.tag)
	    {
	    case TAG_TRIANGLE_OBJ:
		// System.out.print("t");
		objprim[0] = (TriangleObject)ois.readObject();
		break;
	    case TAG_NULL_OBJ:
		objprim[0] = new NullObject();
		// System.out.print("n");
		break;
	    case TAG_SPHERE_OBJ:
		// objprim[0] = (SphereObject)ois.readObject();
		// System.out.print("s");
	    default:
		String s = "Error! : recv object has illegal tag. src = "
		    + probeStat.source + ", tag = " + probeStat.tag;
		throw new RuntimeException(s);
	    }
	    ois.close();
	}catch(MPIException e)
	{
	    System.out.println(e);
	}catch(Exception e)
	{
	    System.out.println("Exception : recvObject recv. " + e);
	    e.printStackTrace();
	}
    }


    /**
     * Isend objects
     *	non blocking send
     *	
     *	@param destRank	
     *	@param buff	ʪΤ
     *	@param sendnum  ϸˤȤ뤳Ȥ¿Τᡤ
     *			ΤĤޤǤ٤Τ򼨤Ƥ
     *  @param sendTag	ΤΥ
     */
    public synchronized Request 
    IsendObj(int destRank, EncodedData lockedSendBuf, int sendObjTag)
    {
	try {
	    Request reqID = 
		MPI.COMM_WORLD.Isend(lockedSendBuf.getArray(), OFFSET, 
				     lockedSendBuf.length(), 
				     lockedSendBuf.type(),     destRank,
				     sendObjTag);
	    return(reqID);
	}catch(MPIException e)
	{
	    System.out.println(e);	    
	    return(null);
	}
    }
	
    /**
     * Irecv objects
     *	non blocking recv
     *	
     *	@param srcRank	  
     *	@param buff	  ʪΤ
     *	@param recvNums[0]ʪο
     *	@param recvTags   ʪΤμ̻()
     */
    public synchronized Request 
    IrecvObj(int srcRank, EncodedData encodedBuf, int recvTags)
    {
	try {
	    Request reqID = 
		MPI.COMM_WORLD.Irecv(encodedBuf.getArray(), OFFSET, 
				     encodedBuf.length(),   encodedBuf.type(),
				     srcRank, recvTags);
	    return(reqID);
	}catch(MPIException e)
	{
	    System.out.println(e);	    
	    return(null);
	}
    }
    
    /*
     * send complete? 
     * (waiting completion of non blocking send)
     */
    public synchronized boolean isSendComplete(Request req_handle)
    {
	try {
	    Status stat = req_handle.Test();
	    if(stat != null)
	    {
		return(true);
	    }else
	    {
		return(false);
	    }
	}catch(MPIException e)
	{
	    System.out.println(e);	    
	    return(false);	    
	}
    }

    /*
     * recv complete? 
     * (waiting completion of non blocking recv)
     */
    public synchronized boolean isRecvComplete(Request req_handle)
    {
	try {
	    Status stat = req_handle.Test();
	    if(stat != null)
	    {
		return(true);
	    }else
	    {
		return(false);
	    }
	}catch(MPIException e)
	{
	    System.out.println(e);	    
	    return(false);
	}
    }

    /**
     * send, recv volume (䤬 object Ȥ줹٤?)
     *	ĹΤΤ serialize ϤʤΤǸĹΤ
     *	ϤΤޤޡserialize ٤Τ serialize Ū
     *	褦  
     *------------------------------------------------------------
     * send subvolume
     */
    // public void commMpi2_sendVolume(int dstRank, VolumeClass *pVol);

    /**
     * recv subvolume
     */
    // public void commMpi2_recvVolume(int srcRank, int dstRank, 
    // 				       VolumeClass **ppVol);

    /**
     * material
     */
    public synchronized 
    void broadMaterial(Material [] aMaterial)
    {
	final int root = SysStat.getInstance().getHostRank();
	if(root == SysStat.getInstance().getMyRank())
	{
	    // ¦
	    try {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ObjectOutputStream oos     = new ObjectOutputStream(baos);
		oos.writeObject(aMaterial[0]);
		oos.flush();
		byte [] serializedobj = baos.toByteArray();
		oos.close();

		// ޤ serialize Υ
		int [] size = new int[1];
		size[0] = serializedobj.length;
		MPI.COMM_WORLD.Bcast(size, OFFSET, 1, MPI.INT, root);
		
		// Τ
 		MPI.COMM_WORLD.Bcast(serializedobj, OFFSET, 
 				     serializedobj.length, MPI.BYTE, root);
	    }catch(Exception e)
	    {
		System.out.println("Exception : broadMaterial send. " + e);
	    }
	}else
	{
	    // ¦
	    try {
		// ޤϼ
		int [] recv_size = new int[1];
		MPI.COMM_WORLD.Bcast(recv_size, OFFSET, 1, MPI.INT, root);
		
 		byte [] serializedobj = new byte[(recv_size[0])];
  		MPI.COMM_WORLD.Bcast(serializedobj, OFFSET, recv_size[0],
  				     MPI.BYTE, root);
 		ByteArrayInputStream bais = 
 		    new ByteArrayInputStream(serializedobj);
 		ObjectInputStream ois = new ObjectInputStream(bais);
 		aMaterial[0] = (Material)ois.readObject();
		bais.close();
	    }catch(Exception e)
	    {
		System.out.println("Exception : broadMaterial recv. " + e);
	    }
	}
    }

    /*------------------------------------------------------------
     * probe 
     *-----------------------------------------------------------*/
    /**
     * can recv 
     * 	@param srcRank_and_recvDatNum[] : ֤
     *		srcRank_and_recvDatNum[0] : 
     *		srcRank_and_recvDatNum[1] :  Data 
     *	@param recvStaffTAG 		: Τ TAG
     *	@param dt 			: ݤ Datatype
     *		MPI.BYTE  MPI.DOUBLE 
     *	@return recv ȯԲǽʤ true 
     *
     * ࡤDatatype ĤäΤϤޤ褯ʤ
     */
    public synchronized boolean canRecv(int [] srcRank_and_recvDatNum, 
					int  recvStaffTAG,
					Datatype dt)
    {
	try {
	    Status probeStat = null;
	    probeStat = MPI.COMM_WORLD.Iprobe(MPI.ANY_SOURCE, recvStaffTAG);
	    if(probeStat != null)
	    {
		srcRank_and_recvDatNum[0] = probeStat.source;
		srcRank_and_recvDatNum[1] = probeStat.Get_count(dt);
		// srcRank_and_recvDatNum[1] = probeStat.Get_count(MPI.DOUBLE);
		return(true);
	    }else
	    {
		return(false);	    
	    }
	}catch(MPIException e)
	{
	    System.out.println(e);	    
	    return(false);
	}
    }

    /**
     * ray
     */
    // public reqHandle_st commMpi2_sendRays(int destRank, Ray* pRay, int numOfRays);

    // public reqHandle_st commMpi2_recvRays(int srcRank, Ray *pRay, int numOfRays)


    /**
     * send intensity to Frame Buffer
     */
    // public void commMpi2_sendIntensity2FB(IntensityClass *pIntensity, int numOfInt);

    // public void commMpi2_recvIntensity2FB(IntensityClass *pIntensity, int numOfIntensity);


    /**
     * send, recv PEStatus
     */
    // public void commMpi2_sendPEstat2Host(PEStatClass *pPEstat);
    // public void commMpi2_recvPEstat(PEStatClass *pPEstat);

    /**
     * send, recv Intensity value and vertex position
     */
    // public reqHandle_st commMpi2_sendIntVer(int destRank, intVerClass *pIntver);
    // public void commMpi2_recvIntVer(int srcRank,  intVerClass *pIntver);

    /**
     * send, recv object
     *	1 PE ξ,ºݤˤ̿ʤΤǤδؿľܰ(̾ 
     *	host)˽񤤤Ƥޤ  
     */
    // public void sendIntensityStandAlone(IntensityClass *pObj, int numOfObj);
}



