/*
 * $Id: SphereObject.java,v 1.2 1999/12/08 16:40:56 yamauchi Exp $
 *     (M\pi)^2 : Sphere Object
 *	Copyright 1999 By YAMAUCHI Hitoshi   
 *	@author YAMAUCHI Hitoshi  
 *	
 * original source sphereObject.h, sphereObject.cc
 * 
 * μ
 */

package jp.ac.tohoku.is.archi.mpi2.obj;

import java.lang.String;
import java.lang.Math;
import java.io.Serializable;
import jp.ac.tohoku.is.archi.mpi2.util.Vector3;
import jp.ac.tohoku.is.archi.mpi2.util.IntVec3;
import jp.ac.tohoku.is.archi.mpi2.util.Assert;
import jp.ac.tohoku.is.archi.mpi2.util.Encodable;
import jp.ac.tohoku.is.archi.mpi2.util.ObjectDecoder;
import jp.ac.tohoku.is.archi.mpi2.sysparam.Mpi2Constants;
import jp.ac.tohoku.is.archi.mpi2.sysparam.Mpi2Enum;
import jp.ac.tohoku.is.archi.mpi2.sysparam.Config;


/**
 * μ
 */
public final class SphereObject
extends ObjectPrimitive  
implements Mpi2Constants, Mpi2Enum, Config, Encodable
, Serializable
{
    private Vector3 mCenter;
    private double  mRadius;

    /*    virtual Vector3 getNormal() 
    {
	cerr + "Error! : sphereObjectClass.getNormal() can't get norm" 
	    + endl;
	commMpi2_exit(1);
	return(Vector3(-1.0, -1.0, -1.0));
    }
    */

    /**
     * constructors
     */
    /*
     * 
     *   @param objectNumber 	ʪֹ
     *   @param materialID 	ޥƥꥢ
     *   @param spcpos 		¸ߤֹ
     *   @param center 		濴ɸ
     *   @param radius 		Ⱦ
     *	ʣʬ֤¸ߤˤϡζֿ롥
     *  ɽ̤¸ߤ֤ɤޤǤɤ
     */
    public SphereObject(int objectID,   int materialID, IntVec3 spcpos,
			Vector3 center, double radius)
    {
	super(objectID, materialID, spcpos);
	this.mCenter = center;
	this.mRadius = radius;
    }
    /*
     * copy construct
     */
    public SphereObject(SphereObject rhs)
    {
	super(rhs.getObjectID(), rhs.getMaterialID(), rhs.peekSpcPos());
	this.mCenter = rhs.mCenter;
	this.mRadius = rhs.mRadius;
    }

    /*
     * ʪΤޤǤεΥ(if ret Double.MAX_VALUE =  not hit)
     *   ʬ hit ʤˤ no hit Ȥ
     */
    public double getDistance(Ray aRay, Vector3 retHitPos)
    {
	Vector3 Lvec  = this.mCenter.minusNew(aRay.peekPos());
	double  Ldist = Lvec.getNorm(); // 濴ȻȤεΥ
	// ray direction ƤϤ
	Assert.out(Math.abs((aRay.peekDir()).getNorm() - 1.0) < 1e-5, 
		   "Error!: ray direction is not normalized. SphereObject");
	
	double distance = Double.MAX_VALUE;

	if(Ldist > this.mRadius)
	{				
	    double  LRd = Lvec.innPrd(aRay.peekDir());
	    // ϵγˤ
	    if(LRd <= 0.0)
	    {
		//  ray  90 ٰʾ夺Ƥ뤿򺹤ʤ
		return(Double.MAX_VALUE); // 90 ٤ϴޤ᤿ no hit
	    }
	    else
	    {
		double td2 = 
		   this.mRadius * this.mRadius - (Ldist * Ldist) + (LRd * LRd);
		if(td2 < 0.0)	//  <= Ǥ?
		{
		    return(Double.MAX_VALUE); // no Hit
		}else
		{
		    distance = LRd - Math.sqrt(td2);

		    // retHitPos = aRay.peekPos() + distance * aRay.peekDir()
		    Vector3 r = new Vector3(aRay.peekDir());
		    r.mulScalerHere(distance);
		    r.plusHere(aRay.peekPos());
		    retHitPos.set(r);
		}
	    }
	}
	else if(Ldist < this.mRadius)
	{
	    // ϵˤ
	    double  LRd = Lvec.innPrd(aRay.peekDir());
	    double td2 = 
		this.mRadius * this.mRadius - (Ldist * Ldist) + (LRd * LRd);
	    distance = LRd + Math.sqrt(td2);	

	    // retHitPos = aRay.getPos() + distance * aRay.getDir();
	    Vector3 r = new Vector3(aRay.peekDir());
	    r.mulScalerHere(distance);
	    r.plusHere(aRay.peekPos());
	    retHitPos.set(r);
	}
	else if(aRay.isInSideObject())
	{
	    //Ʃξ (ɽ΢˻)
	    double  LRd = Lvec.innPrd(aRay.peekDir());
	    distance = 2.0 * Math.sqrt(LRd * LRd);

	    // retHitPos = aRay.getPos() + distance * aRay.getDir();
	    Vector3 r = new Vector3(aRay.peekDir());
	    r.mulScalerHere(distance);
	    r.plusHere(aRay.peekPos());
	    retHitPos.set(r);
	}
	else
	{
	    // ϵܤäƤ롣ϳȤ롣
	    return(Double.MAX_VALUE);	// no Hit
	}
	
	// no hit ξˤ Double.MAX_VALUE  return Ƥ롥Ǥ 
	// distance  Double.MAX_VALUE ǤϤʤ
	Assert.out(distance < Double.MAX_VALUE, 
		   "Error!: Distance must not be MAX_DOUBLE. (SphereObject)");
	
	//  hitpos ʬˤΤȽꤷʬ¸
	// ߤʤˤϡ¾ζ֤ hit ƤΤ no hit Ȥ롥
	// ʬ־䡤cellLength ˤĤƤξʪΤä
	// 櫓ǤϤʤΤǡ PE ȽǤ뤳Ȥˤ 
	
	return(distance);
    }
    
    public Vector3 newFlatRadiosityColor(Vector3 hitPos)
    {
	// NIN
	return(new Vector3(Vector3.ZERO));
    }

    public Vector3 newShadedRadiosityColor(Vector3 hitPos)
    {
	// NIN
	return(new Vector3(Vector3.ZERO));
    }

    /*
     * Hitpoint ˡ٥ȥ
     */ 
    public Vector3 peekNormal(Vector3 hitpos) 
    { 
	Vector3 vnormal = hitpos.minusNew(this.mCenter);
	vnormal.normalize();
	
	return(vnormal);	
    }

    /**
     * toString 
     */
    public String toString()
    {
	String s = 
	    " sph " + this.getObjectID() + " " + this.mCenter + " " + 
	    this.mRadius + " " + this.peekSpcPos();
	return(s);
    }

    /**
     * toStringWithHEad
     */
    public String toStringWithHead(String header)
    {
	String s = header + " # materialID = " + this.getMaterialID() + "\n" +
	    header + " sph " + this.getObjectID() + " " + this.mCenter + " " + 
	    this.mRadius + " " + this.peekSpcPos();
	return(s);
    }

    /*
     * mgf ν
     */ 
    public String getMgfForm(int outtype)
    {
	String s;	
	// Now for SphereObject, outtype is meanless.
	switch(outtype)
	{
	case NORMAL_ObjOut:
	case WITH_SHOT_RADOUT:
	case WITH_UNSHOT_RADOUT:
	    s = "sph " + this.getObjectID() + " " + this.mCenter + " " + 
		this.mRadius + " " + this.peekSpcPos();
	    return(s);
	default:
	    throw new RuntimeException("Internal Error! : " +
				       "SphereObject::getMgfForm : " +
				       "Illegal Output Type.");
	    // break;
	}	

    }

    /*
     * ʪΤμ
     */
    public int  who() 
    {
	return(OBJ_SPHERE);
    }

    /*
     * ֤ˤ뤫
     */
    public boolean isAtValidSpcPos(Vector3 partmin, Vector3 subSpcSize,
				   int subdivnum)
    {
	// check spcpos num
	for(int i = 0; i < 3; i++)
	{
	    if((this.peekSpcPos().get(i) < 0) ||
	       (this.peekSpcPos().get(i) >= subdivnum))
	    {
		System.out.println("Error! :sph(sphere): spcpos " + 
				   "is out of define space.");
		return(false);
	    }
	}
	
	// check spcpos range
	int [] centerSpcpos = new int[3];
	for(int i = 0; i < 3; i++)
	{
	    // 濴ְ
	    centerSpcpos[i] = 
		(int)((this.mCenter.get(i) - partmin.get(i)) / 
		      subSpcSize.get(i));
	    // Ⱦ¤ˤְϰ
	    int radSpcposBias = (int)(this.mRadius / subSpcSize.get(i)) + 1;
	    if((this.peekSpcPos().get(i) < centerSpcpos[i] - radSpcposBias) ||
	       (this.peekSpcPos().get(i) > centerSpcpos[i] + radSpcposBias))
	    {
		final String [] s = { "x", "y", "z" };
		String ss = 
		    "Error! : sph(sphere): spcpos " + this.peekSpcPos() + 
		    " illigal spcpos : " + s[i] + "\n" +
		    "center " + this.mCenter    + "\n" + 
		    "radius " + this.mRadius    + "\n" + 
		    "centerpos[" + i + "] = " + centerSpcpos[i] + "\n" + 
		    "rad bias " + radSpcposBias + "\n" + 
		    "spbspace size " + subSpcSize + "\n" + 
		    "partmin " + partmin;
		System.out.println(ss);
		return(false);
	    }
	}
	return(true);
    }


    /*------------------------------------------------------------
     * ̿
     *------------------------------------------------------------*/
    /*
     * ̿ѤΥ֤
     */
    public static int getTAG() 
    { 
	return(Mpi2Enum.TAG_SPHERE_OBJ);	
    }

    /**
     * For Encodable interface (and ObjectDecoder object).
     */
    final static int vectorSize = Vector3.ZERO.sizeInDouble();
    final static int intvecSize = IntVec3.ZERO.sizeInDouble();
    final static int doubleSize =
    1 +  			// Object ID
    1 +				// material ID
    + intvecSize		// spcpos size
    + vectorSize 		// 濴
    + 1;			// Ⱦ
    public int sizeInDouble() 
    {
	return doubleSize;
    }
    public void encodeInDouble(double[] array, int offset) 
    {
	// SphereObject  construct ΤɬפʬΤ
	array[offset] = (double)this.getObjectID();
	offset++;

	array[offset] = (double)(this.getMaterialID());     
	offset++;

	(this.peekSpcPos()).encodeInDouble(array, offset);
	offset += intvecSize;
    
	this.mCenter.encodeInDouble(array, offset);
	offset += vectorSize;

	array[offset] = this.mRadius;     
	offset++;
    }
    static class SphereObjectDecoder implements ObjectDecoder {
    	public int sizeInDouble() {
	    return doubleSize;
	}
    	public Object decodeFromDouble(double[] array, int offset) 
	{
	    // 	for(int i = 0; i < array.length; i++)
	    // 	{
	    // 	    System.out.println("DEBUG Decode " + array[i]);
	    // 	}

	    ObjectDecoder vec3  = Vector3.getObjectDecoder();
	    ObjectDecoder ivec3 = IntVec3.getObjectDecoder();
	
	    int objectID = (int)(array[offset]);
	    offset++;

	    int materialID = (int)(array[offset]);
	    offset++;

	    IntVec3 spcpos = (IntVec3)(ivec3.decodeFromDouble(array, offset));
	    offset += intvecSize;
    
	    Vector3 center = (Vector3)(vec3.decodeFromDouble(array, offset));
	    offset += vectorSize;
    
	    double  radius = array[offset];
	    offset++;
	    
	    // 	System.out.println("SphObj::decode mID = " + materialID + 
	    // 			   ", spcpos " + spcpos + "\nv1 " + v1 + 
	    // 			   ", v2 " + v2 + ", v3 " + v3

	    return(new SphereObject(objectID, materialID, spcpos,
				    center, radius));
	}
    }
    private final static ObjectDecoder odInstance = new SphereObjectDecoder();
    public static ObjectDecoder getObjectDecoder() {
	return odInstance;
    }
}
