/*
 * $Id: CalcRadioSRG.java,v 1.5 2000/06/15 13:55:51 yamauchi Exp $
 *     (M\pi)^2 : CalcRadioSRG 
 *	Copyright 1999 By YAMAUCHI Hitoshi   
 *	@author YAMAUCHI Hitoshi  
 *	
 * original source calcRadioSRG.cc
 *
 * 	Secondary Rays Generation Unit
 *	饸ƥˡˤ󼡸ȯʬ
 *
 * 1999ǯ818()
 *
 *-- Ϥγˤϴï⤤ʤΤˤϻ¿ΤߤǤ롥
 *-- Τ˲Ƶ٤ߤȤϤʿï⤤ʤؤȤΤΤ
 *-- ʡؤ¾ˤ뤳ȤʤäȤοʹ֤
 *-- ̡¿ȤϤäɤȤʤΤȤ
 *-- ˤĤä򤷤
 *--
 *-- Mahler  3 ֤ Bernstein, NewYork Phillharmonic ΤΤꤷ
 */

package jp.ac.tohoku.is.archi.mpi2.radiosity;

import	jp.ac.tohoku.is.archi.mpi2.obj.*;
import	jp.ac.tohoku.is.archi.mpi2.common.*;
import	jp.ac.tohoku.is.archi.mpi2.util.*;
import	jp.ac.tohoku.is.archi.mpi2.sysparam.*;

/*
 * 饸ƥȿͤˤ󼡥쥤ȯ
 *	߿ΤѤ ray tracing ǤϤʤ
 */
public final class CalcRadioSRG implements Mpi2Constants {
    /**
     * ȿͥ饸ƥ Ray 
     */
    private static Ray newReflectionRadiosityRays(Ray orgray,
						  double minRayCoefNorm,
						  ObjectPrimitive hitObj,
						  Material [] matArray)
    {
	Assert.out(hitObj != null, 
		   "Error!: no hit obj, newReflectionRadiosityRays");

	final Material mat = matArray[hitObj.getMaterialID()];

	// ȿ͸그׻
	Vector3 afterReflectRadiosity = 
	    orgray.peekColor().mulVecNew(mat.peekSpecRef());

	/*
	 * commented by tosho : ޤǤƼϤʤ ray Ƥ
	 * ȿ͸그ʲΥͥ륮 ray Ϥʾ¤ʤ
	 * ʬminRayCoefNorm  ray-tracing ƱǤΤϸ
	 * ⤷ʤ 
	 */
	if(afterReflectRadiosity.innPrd(afterReflectRadiosity)
	   < minRayCoefNorm)
	{
	    return(null);
	}

	// ȿ
	final Vector3 normal = hitObj.peekNormal(orgray.peekHitPos());

	// ˡƤϤ
	Assert.out(Math.abs(normal.getNorm() - 1.0) < D_ERROR_RES, 
		   "Error!: normal vector must be normalized. reflection");

	//
	// Vector3 reflectedRayDir = 
	// 	(-2.0 * ((orgray.getDir()).dInn_prd(normal)) * normal) + 
	// 	 orgray.getDir();
	//
	double coeff = (-2.0 * ((orgray.peekDir()).innPrd(normal)));
	Vector3 reflectedRayDir = normal.mulScalerNew(coeff);
	reflectedRayDir.plusHere(orgray.peekDir());

	// ȿͥ쥤 newRay! for radiosity ⥻åȤ
	// Υ󥹥ȥ饯Ǻ쥤ϰäʪΤˤϸ򺹤ʤ 
	Ray newRay = Ray.newForReflectionRadiosity(orgray, reflectedRayDir, 
						   afterReflectRadiosity, 
						   hitObj);
	return(newRay);
    }

    /**
     * ޤˤ󼡥쥤ȯ
     */
    private static Ray newRefractionRadiosityRays(Ray orgray,
						  double minRayCoefNorm,
						  ObjectPrimitive hitObj,
						  Material [] matArray)
    {
	// ray  copy  pMat Ʃ᷸ݤ
	final Material mat = matArray[hitObj.getMaterialID()];
	Vector3 afterRefractRadiosity = 
	    (orgray.peekColor()).mulVecNew(mat.peekSpecTrans());

	// ͥ륮Ʒ׻ɬפʤ
	if(afterRefractRadiosity.innPrd(afterRefractRadiosity) 
	   < minRayCoefNorm)
	{
	    return(null);
	}
    
	// Ʃ׻
	Vector3 refractedRayDir = null;
	if(mat.getIdxRefraction() != 1.0)	
	{
	    // ޤ ޥ쥤
	    Assert.out(hitObj != null, 
		       "Error!: no hit obj, newRefractionRadiosityRays");
	    Vector3 normal = hitObj.peekNormal(orgray.peekHitPos());

	    //ˡƤϤ
	    Assert.out(Math.abs(normal.getNorm() - 1.0) < D_ERROR_RES, 
		       "Error!: normal vector must be normalized. refraction");
	    
	    // Υ٥ȥ T (I ͥ٥ȥ, N ʪˡ٥
	    // , n жΨ >= 1.0)  
	    // T = (n (-N \dot I) - \sqrt(1 - n^2(1 - (N \dot I)^2)) N) + nI
	    double n   = mat.getIdxRefraction(); // Ψ
	    double ndi = (orgray.peekDir()).innPrd(normal); // (N \dot I)
	    double rt  = Math.sqrt(1.0 - (n * n) * (1.0 - (ndi * ndi)));
	    // refractedRayDir = 
	    //	(((n * (-ndi)) - rt) * normal) + (n * orgray.getDir());
	    refractedRayDir = normal.mulScalerNew((n * (-ndi)) - rt);
	    Vector3 dir = (orgray.peekDir()).mulScalerNew(n); 
	    refractedRayDir.plusHere(dir);
	}else
	{
	    // mat.getIdxRefraction() == 1.0  : ƩΨ 1 ̵̣ʡ
	    refractedRayDir = orgray.peekDir();
	}

	//-- Ʃ쥤 newRay! for radiosity ⥻åȤ
	//-- Υ󥹥ȥ饯Ǻ쥤ϰäʪΤˤϸ򺹤ʤ 
	//-- Ū˽ refraction  refraction Ʊ
	Ray newRay = 
	    Ray.newForReflectionRadiosity(orgray, refractedRayDir, 
					  afterRefractRadiosity, hitObj);
	return(newRay);
    }

    /**
     * 2 radiosity ray 
     *     	ٷ׻ IC ǺѤǤ
     *	(ʸ˽񤤤Ƥ߷פȤϰ㤤ΤȤ IC.SRG ³򼨤
     *	 Ƥ롥ʸǤȿ FrameBuffer ؤƤ뤬
     *	 ˤ ray Ǥκݤˤ FrameBuffer ؤιʤ
     *		(see mpi2doc_implementation)
     * 
     * @param orgRay ʪΤȸ򺹤쥤
     * @param hitObj 쥤ȸ򺹤ʪ
     * @param pe     쥤ȯˤϤ PE ˥쥤򥻥åȤ
     */
    public static boolean newSecondaryRadiosityRays(Ray orgRay, 
						    ObjectPrimitive hitObj, 
						    PE  pe)
    {
	// 쥤ȯ? 
	boolean isGenerateRay = false;
	final ViewData viewdat = pe.peekViewData();

	//  ray ȿ͡޲ۤƤ뤫
	if(orgRay.getReflectedNum() >= viewdat.getMaxRayRefNum())
	{				// ۤƤ롣ray ȯʤ
	    return(isGenerateRay);
	}
    
	/*
	 *  ȿ 2  radiosity ray  new PE ν塼Ͽ
	 */
	Ray newHanshaRay = 
	    newReflectionRadiosityRays(orgRay, 
				       viewdat.getMinRayRefCoeffNorm(),
				       hitObj, pe.peekMatArray());
	if(newHanshaRay != null)	// ray ȯ塼
	{
	    pe.setRayToThisPE(newHanshaRay);		   
	    isGenerateRay = true;
	    // ȿ radiosity ray ΤʤФ Hint TOKUNAGA
	}

	/*
	 * Ʃ 2  radiosity ray  new PE ν塼Ͽ
	 */
	Ray newToukaRay = 
	    newRefractionRadiosityRays(orgRay, 
				       viewdat.getMinRayRefCoeffNorm(),
				       hitObj, pe.peekMatArray());
	if(newToukaRay != null)
	{
	    pe.setRayToThisPE(newToukaRay);
	    isGenerateRay = true;
	    // ޥ쥤ΤʤФ Hint TOKUNAGA
	}
	return(isGenerateRay);
    }
}
