/*
 * $Id: HemiCube.java,v 1.5 2000/06/15 13:55:51 yamauchi Exp $
 *     (M\pi)^2 : HemiCube 
 *	Copyright 1999 By YAMAUCHI Hitoshi   
 *	@author YAMAUCHI Hitoshi  
 *	
 * original source hemicube.h hemicube.cc
 *
 *	Hemi-Cube μ
 * 
 * 1999ǯ8 4()
 */

package jp.ac.tohoku.is.archi.mpi2.radiosity;

import  jp.ac.tohoku.is.archi.mpi2.util.IntVec3;
import  jp.ac.tohoku.is.archi.mpi2.util.Vector3;
import  jp.ac.tohoku.is.archi.mpi2.util.Assert;
import  jp.ac.tohoku.is.archi.mpi2.obj.ObjectPrimitive;
import  jp.ac.tohoku.is.archi.mpi2.obj.Ray;
import  jp.ac.tohoku.is.archi.mpi2.obj.TriangleObject;
import  jp.ac.tohoku.is.archi.mpi2.sysparam.Mpi2Constants;

/**
 * Hemi Cube
 * μǤϡ
 * 	1.HemiCube 򤢤 hemres Ǻpatch  Normal Vector 
 * 	  ƽ롣 
 * 	2.θϡpatch  Normal vector ͿƼνԤ
 *	  ȤǺѤ롣
 */

/**
 * Hemi-Cube
 * 	( Hemi-cube ϳƥѥåΤưŪ˺)
 */
public final class HemiCube 
implements Mpi2Constants 
{
    public final static int WALL_CEILING 	= 0; // ŷ
    public final static int WALL_XPLUS		= 1; // 
    public final static int WALL_XMINUS		= 2;
    public final static int WALL_YPLUS		= 3;
    public final static int WALL_YMINUS		= 4;
    public final static int HEMICUBE_PLANE_MAX	= 5;
    
    // set in getMemory()
    private double   [][][]mFFval;	// FormFacter ɤΤΤ1Ĥɤ
    private Vector3  [][][]mFFEOrgPos;	// 줾ΥȤβžΰ
    private Vector3  [][][]mFFEPos;	// žΤ줾ΥȤΰ

    // hemi cube info
    private double   mElemArea;		// 줾ΥȤ
    private int	     mHemres;		// Hemi-cube resolution

    private IntVec3  mPEID;		// ô PE ID
    
    // raiation patch informatio
    private Vector3  mPatchNormal; 	// ͥѥåˡ
    // private Vector3  mGravityCenter;	// ͥѥåνſ
    private double   mPatchArea;	// ͥѥå
    private IntVec3  mSpcpos;		// ͥѥå¸ߤֻɸ
    private TriangleObject mPatch;	// ͥѥå
    private Vector3  mLockingNotYetRad;	// ̤ͤΥå


    /**
     * get hemi-cube resolution
     */
    private int getHemres()	 	
    {
	return(this.mHemres);	
    }

    /**
     * Hemi-Cube ѤΥ : ¿ǽ򰭲뤫 FIXME
     */ 
    private void getMemory()
    {
	Assert.out((this.getHemres() > 1)||(this.getHemres() % 2 == 0),
		   "HemiCube is already initialized"); // inited?
    
	//------------------------------
	// getting form factor values area
	// (this.mpFFcal[ɤμ][⤵][])
	//------------------------------
	this.mFFval = new double[2][][]; // ŷ 1 Ĥ
	//     ŷ
	this.mFFval[WALL_CEILING] = new double [this.getHemres()][];  
	for(int i = 0; i < this.getHemres(); i++)
	{
	    this.mFFval[WALL_CEILING][i] = new double[this.getHemres()];
	}
	//     ɤ1
	// ɤι⤵Ͼ̤Ⱦʬ
	this.mFFval[WALL_XPLUS] = new double [this.getHemres() / 2][];
	for(int i = 0; i < (this.getHemres() / 2); i++) // ɤι⤵Ⱦʬ
	{
	    // ɤƱ
	    this.mFFval[WALL_XPLUS][i] = new double[this.getHemres()];
	}
    
	//------------------------------
	// getting Vector3 values    
	// (this.mFFEOrgPos[ɤμ][⤵][])
	//------------------------------
	this.mFFEOrgPos = new Vector3 [HEMICUBE_PLANE_MAX][][];
	this.mFFEPos    = new Vector3 [HEMICUBE_PLANE_MAX][][];
	
	for(int w = WALL_CEILING; w < HEMICUBE_PLANE_MAX; w++)
	{
	    int maxHeight = 0;		// դι⤵
	    if(WALL_CEILING == w)		// ŷ
	    {
		maxHeight = this.getHemres();
	    }else
	    {
		maxHeight = this.getHemres() / 2; // ŷʳ
	    }
	    this.mFFEOrgPos[w] = new Vector3 [maxHeight][]; 
	    this.mFFEPos[w]    = new Vector3 [maxHeight][]; 
	    
	    // h (height)  mHemres 뤤 mHemres/2
	    for(int h = 0; h < maxHeight; h++)  //  maxHeight ޤ
	    {
		this.mFFEOrgPos[w][h] = new Vector3[this.getHemres()];
		this.mFFEPos[w][h]    = new Vector3[this.getHemres()];
	    }
	}
    }

    /**
     * setFormFactor
     *	Master Hemi-Cube(ž Hemi-Cube)Υեեȳƥԥ
     *	֤ 
     * +--------------+ --
     * |              | ^
     * | Hemi-Cube    | | hight
     * | Plane        | v
     * +--------------+ --
     * |<-- width --.|
     *                  pixel  FF  mFFval[plane][hight][width]
     */
    private void setFormFactor()
    {
	final double dZMAX =  1.0;		  // Hemi-Cube ι⤵
	final double dZMIN =  0.0;		  // -1.0    1.0   
	final double dXMIN = -1.0;		  //   +-------+ ---  1.0 
	final double dXMAX =  1.0;		  //   | Hemi- |
	final double dYMIN = -1.0;		  //   |  Cube |    top view
	final double dYMAX =  1.0;		  //   +-------+ --- -1.0 
	final double dCEIL_EDGE = dXMAX - dXMIN;  //   |<----.| dCEIL_EDGE
	final double dELEM_EDGE		   	  // dELEM_EDGE = ǤΰĹ
	    = dCEIL_EDGE / (double)(this.getHemres());
	final double dELEM_BIAS = dELEM_EDGE/2.0; // Ǥ濴뤿bias
	final double dElemArea = dELEM_EDGE * dELEM_EDGE; // Ǥ:

	// ŷʬ 
	int ix = 0;		// for(ix = 0; ; ix++)
	for(double x =  dXMIN + dELEM_BIAS; x <  dXMAX; x += dELEM_EDGE)
	{
	    int iy = 0;		// for(iy = 0; ; iy++)
	    for(double y = dYMIN + dELEM_BIAS; y < dYMAX; y += dELEM_EDGE)
	    {
		// assert(iy < this.getHemres());
		double dR_sqr = x * x + y * y + 1.0;
		dR_sqr *= dR_sqr;
		this.mFFval[WALL_CEILING][ix][iy] = 
		    dElemArea / (Math.PI * dR_sqr); 
		this.mFFEOrgPos[WALL_CEILING][ix][iy] = 
		    new Vector3(x, y, dZMAX);
		iy++;
	    }
	    ix++;
	}

	//----------------------------------------------------------
	//  Ĥʬ xp(ulus), xm(inus), yp(ulus), ym(inus)Τ
	// Ĥ狼ФȤƱ
	//----------------------------------------------------------
	int iz = 0;
	for(double z =  dZMIN + dELEM_BIAS; z < dZMAX; z += dELEM_EDGE)
	{
	    // assert(iz < (this.getHemres() / 2));
	    int iy = 0;
	    for(double y = dYMIN + dELEM_BIAS; y < dYMAX; y += dELEM_EDGE)
	    {
		// assert(iy < this.getHemres());
		double dR_sqr = 1.0 + y * y + z * z;
		dR_sqr *= dR_sqr;
		this.mFFval[WALL_XPLUS][iz][iy] = 
		    (dElemArea * z) / (Math.PI * dR_sqr);
		this.mFFEOrgPos[WALL_XPLUS][iz][iy] = new Vector3(dXMAX, y, z);
		this.mFFEOrgPos[WALL_XMINUS][iz][iy]= new Vector3(dXMIN, y, z);
		iy++;
	    }
	    iz++;
	}

	iz = 0;
	// z Ϲ⤵(hight)
	for(double z = dZMIN + dELEM_BIAS; z < dZMAX; z += dELEM_EDGE)
	{
	    // assert(iz < (this.getHemres() / 2));
	    ix = 0;		// x  width
	    for(double x = dXMIN + dELEM_BIAS; x < dXMAX; x += dELEM_EDGE)
	    {
		// assert(ix < this.getHemres());
		this.mFFEOrgPos[WALL_YPLUS][iz][ix] = new Vector3(x, dYMAX, z);
		this.mFFEOrgPos[WALL_YMINUS][iz][ix]= new Vector3(x, dYMIN, z);
		ix++;
	    }
	    iz++;
	}
    }
    
    /**
     * rotate hemi cube (by Eular rotation)
     * 	(Maybe, this code includes my oldest code in mpi2)
     * 
     * 	@param aPatchNrm : ѥåˡ  HemiCube 
     */
    private void rotateHC(Vector3 aPatchNrm)
    {
	final double D_NON_DEF = -2.0;  // ꤨʤ
	// double a[3][3];		/* 顼Ѵޥȥꥯ  */
	double a00, a01, a02, a10, a11, a12, a20, a21, a22;
	double dCtheta = D_NON_DEF;	/* (x' axis)     */
	double dStheta = D_NON_DEF;
	double dCkai   = D_NON_DEF;	/* (y' axis)     */
	double dSkai   = D_NON_DEF;
	double dCphi   = 1.0;		/* (z' axis)=0.0 */
	double dSphi   = 0.0;
	
	double dNx = aPatchNrm.get(Vector3.X); // ѥåˡ٥ȥ
	double dNy = aPatchNrm.get(Vector3.Y); /* nrmƤ */
	double dNz = aPatchNrm.get(Vector3.Z);
	
	dCtheta = dNz;
	dStheta = Math.sqrt(1.0 - dNz * dNz ); /* Ȥꤢ˲ */
	if(0.0 == dStheta)
	{
	    dSkai = 0.0;
	    dCkai = 1.0;
	}else
	{
	    dSkai   = dNx / dStheta;
	    dCkai   = dNy / dStheta;
	}
	
	/* -1.0 < sin,cos < 1.0 */
	Assert.out((-1.0 <= dCtheta) && (dCtheta <= 1.0), "illegal Ctheta");
	Assert.out((-1.0 <= dStheta) && (dStheta <= 1.0), "illegal Stheta");
	Assert.out((-1.0 <= dCphi)   && (dCphi <= 1.0)  , "illegal Cphi");
	Assert.out((-1.0 <= dSphi)   && (dSphi <= 1.0)  , "illegal Sphi");
	
	a00 =   dCkai   * dCphi - dSkai * dCtheta * dSphi;
	a10 = - dSkai   * dCphi - dCkai * dCtheta * dSphi;
	a20 =   dStheta * dSphi;
	
	a01 =   dCkai   * dSphi + dSkai * dCtheta * dCphi;
	a11 = - dSkai   * dSphi + dCkai * dCtheta * dCphi;
	a21 = - dStheta * dCphi;
	
	a02 =   dSkai * dStheta;
	a12 =   dCkai * dStheta;
	a22 =   dCtheta;
	
	int maxHeight;
	for(int hcPlane =WALL_CEILING; hcPlane < HEMICUBE_PLANE_MAX; hcPlane++)
	{
	    if(hcPlane == WALL_CEILING)
	    {
		maxHeight = this.getHemres();
	    }else
	    {
		maxHeight = this.getHemres() / 2; // ŷʳϹ⤵Ⱦʬ
	    }
	    for(int height = 0; height < maxHeight; height++)
	    {
		for(int width = 0; width < this.getHemres(); width++)
		{
		    Assert.out(isHemiCubePosXYZRight(hcPlane, height, width),
			       "illegal HC Pos"); 
		    Vector3 orgposRef = 
			this.mFFEOrgPos[hcPlane][height][width];
		    
		    double x = orgposRef.get(Vector3.X);
		    double y = orgposRef.get(Vector3.Y);
		    double z = orgposRef.get(Vector3.Z);
		    
		    double xt = a00 * x + a01 * y + a02 * z; 
		    double yt = a10 * x + a11 * y + a12 * z;
		    double zt = a20 * x + a21 * y + a22 * z;
		    
		    this.mFFEPos[hcPlane][height][width] 
			= new Vector3(xt, yt, zt);
		}
	    }
	}
    }
    
    /**
     * generate rays using by Hemi-Cube : distributed method
     * 	ray ̤򼡡ѲƱ(˶ᤤ)Ϣ³ ray 
     *	ͤʤ褦ˤ
     *	load balancing θ̤ȹͤˡ  
     */
    private void distributeRadiation(Ray [] retRays)
    {
	int ridx = 0;		// index for return ray 
	int halfHemres = (this.getHemres() / 2);
	for(int h = 0; h < halfHemres; h++)
	{
	    for(int w = 0; w < this.getHemres(); w++)	    
	    {
		// ŷ2 : h * w  h + halfHemres * w  ray 
		retRays[ridx] = this.newOneRay(WALL_CEILING, h, w);
		ridx++;
		retRays[ridx] = this.newOneRay(WALL_CEILING, h +halfHemres, w);
		ridx++;

		// ŷǤʤʬ
		for(int plane = WALL_XPLUS;plane < HEMICUBE_PLANE_MAX; plane++)
		{ 
		    retRays[ridx] = this.newOneRay(plane, h, w);
		    ridx++;
		}
	    }
	}
    }

    /**
     * generate rays using by Hemi-Cube : continuous method
     *	ray ̤ѲƱ(˶ᤤ)Ϣ³ ray 
     *	ͤ
     *	load balancing θ̤ȹͤˡ  
     */
    private void continuousRadiation(Ray [] retRays)
    {
	int ridx = 0;		// index for return ray 
	for(int plane = WALL_CEILING; plane < HEMICUBE_PLANE_MAX; plane++)
	{ 
	    //(plane==WALL_CEILING)?(this.getHemres()):(this.getHemres()/2);
	    int maxHeight;
	    if(plane == WALL_CEILING)
	    {
		maxHeight = this.getHemres();
	    }else
	    {
		maxHeight = this.getHemres() / 2;
	    }
	    for(int h = 0; h < maxHeight; h++)
	    {
		for(int w = 0; w < this.getHemres(); w++)	    
		{
		    retRays[ridx] = this.newOneRay(plane, h, w);
		    ridx++;
		}
	    }
	}
    }

    /**
     * generate rays using by rotetaed hemi-cube
     *	Ray α delta radiosity  \delta F_{ij} * PatchArea
     */
    private Ray newOneRay(int plane, int height, int width)
    {
	Assert.out(isHemiCubePosXYZRight(plane, height, width),
		   "illegal position is indicated : newOneRay");

	Vector3 newPos = this.mPatch.getNewGCenter();
	Vector3 newDir = new Vector3(this.mFFEPos[plane][height][width]);
	newDir.normalize();	// 쥤Υ٥ȥ
	Vector3 newRadiosity = new Vector3(this.mLockingNotYetRad);
    
	// 쥤ˤϥեեͥѥåѤݤäƤ
	int ffvalPlane = plane;
	if(plane > WALL_XPLUS)
	{
	    ffvalPlane = WALL_XPLUS; // ¦̤ FF val ƱʤΤ1ĤˤƤ
	}

	// Rad = not yet radiated radiosity value (locked)
	// Rad = Rad * ((FF_{ij}[plane][h][w]) * patchArea)
	double ff_patcharea = 
	    this.mFFval[ffvalPlane][height][width] * this.mPatchArea;
	newRadiosity.mulScalerHere(ff_patcharea);
    
	// 쥤 newray
	Ray newRay = Ray.newForRadiosity(newPos, newDir, newRadiosity,
					 this.mPatch,
					 new IntVec3(this.mSpcpos));
	return(newRay);
    }

    /**
     * 顼å
     */
    private boolean isHemiCubePosXYZRight(int plane, int height,  int width) 
    {
	if((plane < 0) || (plane >= HEMICUBE_PLANE_MAX))
	{
	    throw new RuntimeException("Error! : illegal plane = " + plane
				       + " at MHC position");
	    // return(false);
	}
	
	int MaxHeight = 0;
	if(plane != WALL_CEILING)
	{
	    MaxHeight = this.getHemres() / 2;
	}else
	{
	    MaxHeight = this.getHemres();
	}
	
	if((height < 0) || (height >= MaxHeight)) // ⤵Ѳ
	{
	    throw new RuntimeException("Error!: illegal height = " + height
				       + " at MHC position");
	    // return(false);
	}
	
	if((width < 0) || (width >= this.getHemres())) // Ѳʤ
	{
	    throw new RuntimeException("Error!: illegal hemres = " + 
				       this.getHemres() + " width = "
				       + width + " at MHC position");
	    // return(false);	
	}
	return(true);
    }
    
    /**
     * construct
     */
    public HemiCube(int aHemres, IntVec3 PEID)
    {
	// host  hemi cube ʤ
	Assert.out(PEID.get(IntVec3.X) >= 0, "Host has no hemi-cube."); 
	
	if((aHemres < 2) || (aHemres % 2 != 0))
	{
	    String s = "Error! : illegal hemres value. hemres must be > 1 " 
		+ "and % 2 == 0, hemres = " + aHemres;
	    throw new RuntimeException(s);
	    // commMpi2_exit(1);
	}
	
	this.mFFval		= null;
	this.mFFEOrgPos		= null;
	this.mFFEPos		= null;
	
	this.mElemArea		= -1.0;
	this.mHemres		= aHemres;
	
	this.mPEID		= PEID;	
	
	this.mPatchNormal 	= null;
	// this.mGravityCenter	= null;
	this.mPatchArea		= 0.0;
	this.mSpcpos		= null;
	this.mPatch		= null;
	this.mLockingNotYetRad	= null;
	
	// memory allocation 
	this.getMemory();

	// System.out.println("DEBUG : " + this);
	
	// initialize master hemi-cube
	this.setFormFactor();

	// System.out.println("DEBUG : " + this);
    }

    
    /**
     * ȯ ray ֤ : 1 ray/pixel 
     * 	ȯ ray  =  * 4 + ŷ(ɤ礭ŷȾʬ) = 3 * ŷ 
     */
    public int getGenRayNum()  
    { 
	return(3 * this.mHemres * this.mHemres); 
    }
    
    /**
     * 쵤˰Ĥ HemiCube ʬ ray ȯpRetRays ˥ݥ
     * 򥳥ԡ(ݥ󥿤 new Ƥ)֤ 
     * pRetRays ΰϸƤӽФ¦ݤϤȡ
     *	@param  radpatch 	ͤԤʤäѥå
     *  @param  isDistributedRadiation 	ray ȯ load balance
     *  @param  retRays		 ray 
     */
    public void genRays(TriangleObject radpatch,
			boolean isDistributedRadiation,
			Ray []  retRays)
    {
	// must normalized normal
	Assert.out(Math.abs((radpatch.peekNormal(null)).getNorm() - 1.0) < 
		   D_ERROR_RES,
		   "Error! : genRays : patch normal is not normalized.");
	Assert.out((retRays != null) && (retRays.length > 0), 
		   "return [] ray is not initialized");

	// patch Ϥζ֤˾¸ߤΤ reference  OK 
	
	this.mPatchNormal	= radpatch.peekNormal(null);
	// this.mGravityCenter	= radpatch.getNewGCenter();
	this.mPatchArea		= radpatch.getArea();
	
	// Patch ˤѤϤ
	Assert.out(this.mPatchArea > 0.0, "Patch must have area");
	
	this.mSpcpos		= radpatch.peekSpcPos();
	Assert.out((this.mSpcpos.get(IntVec3.X) >= 0) && 
		   (this.mSpcpos.get(IntVec3.Y) >= 0) && 
		   (this.mSpcpos.get(IntVec3.Z) >= 0),
		   "Illegal spcpos of patch"); 
	this.mPatch = radpatch;
	
	//-- ¾ radiosityͤƤɤ褦ͤ lock 
	//-- ɤϽ񤤤ߡmulti thread ǤʤΤ lock ʤƤ
	//-- Ǥ롣Ĥɬפˤʤ뤫⤷ʤ(塼ʬ 
	//-- atomic ˤƤʤΤǡ MT ˤʤäϤ狼ʤ) 
	//-- 7 Sept. 1996 Hitoshi Yamauchi 
	// critical region begin
	this.mLockingNotYetRad = 
	    radpatch.getNewClearNonRadAndUpdateRadiosity();
	radpatch.unsetInRadiationQueue(); //-- 줬̵ä1998ǯ1211()
	// critical region end
	
	this.rotateHC(this.mPatchNormal);
	
	if(isDistributedRadiation == true)
	{
	    this.distributeRadiation(retRays);
	}else
	{  
	    this.continuousRadiation(retRays);
	}
    }

    public String toString()
    {
	String s = "----- HemiCube -----\n" +
	    "FormFactor value array [" + this.mFFval.length + "][" + 
	    this.mFFval[0].length + "][" + this.mFFval[0][0].length + "]\n" + 
	    "FF Elem orgpos   array [" + this.mFFEOrgPos.length + "][" + 
	    this.mFFEOrgPos[0].length + "][" + this.mFFEOrgPos[0][0].length +
	    "]\n" + 
	    "FF Elem pos      array [" + this.mFFEPos.length + "][" + 
	    this.mFFEPos[0].length + "][" + this.mFFEPos[0][0].length +"]\n" + 
	    "FF Elem Area " + this.mElemArea + "\n" + 
	    "hemres " + this.mHemres + "\n" + 
	    "PE id  " + mPEID + "\n" + 
	    "patch normal " + this.mPatchNormal +"\n" + 
	    // mGravityCenter
	    "patch area " + this.mPatchArea + "\n" + 
	    "space position " + this.mSpcpos + "\n" + 
	    // mPatch;	
	    "lock " + this.mLockingNotYetRad;
	return(s);
    }
    
    /**
     * for test of FormFactor value
     */
    public static void main(String [] args)
    {
	IntVec3 PEID = new IntVec3(0, 0, 0);
	int hemres = 10;
	HemiCube hc = new HemiCube(hemres, PEID);

	for(int x = 0; x < hc.getHemres(); x++)
	{
	    for(int y = 0; y < hc.getHemres(); y++)
	    {
		System.out.println("[" + x + "][" + y + "]= " +
				   hc.mFFval[WALL_CEILING][x][y]);
	    }
	}
    }
}
