/*
 * $Id: RadioRenderer.java,v 1.19.4.2 2000/06/19 17:46:20 yamauchi Exp $
 *     (M\pi)^2 : Radio Renderer (Radiosity Renderer)
 *	Copyright 1999 By YAMAUCHI Hitoshi   
 *	@author YAMAUCHI Hitoshi  
 *	
 * original source PERadiosity.cc
 *
 *	饸ƥˡμ
 * 
 * 1999ǯ817()
 */

package jp.ac.tohoku.is.archi.mpi2.radiosity;

import  java.io.*;

import	jp.ac.tohoku.is.archi.mpi2.common.*;
import	jp.ac.tohoku.is.archi.mpi2.dda.*;
import	jp.ac.tohoku.is.archi.mpi2.net.*;
import	jp.ac.tohoku.is.archi.mpi2.obj.*;
import	jp.ac.tohoku.is.archi.mpi2.host.*;
import	jp.ac.tohoku.is.archi.mpi2.sysparam.*;
import	jp.ac.tohoku.is.archi.mpi2.util.*;

import  jp.ac.tohoku.is.archi.mpi2.jdk12src.Map;
import  jp.ac.tohoku.is.archi.mpi2.jdk12src.Iterator;
import  jp.ac.tohoku.is.archi.mpi2.jdk12src.HashMap;
import  jp.ac.tohoku.is.archi.mpi2.jdk12src.Vector;


/**
 * sub functions for radiosity stage
 *
 * radiosity ν ʣܤ delta-radiosity ray 
 *    Ƚꡢַ׻Ԥ ray  NetworkInterface 
 */
public final class RadioRenderer 
implements Mpi2Enum, Mpi2Constants, HostConstants {
    // PE
    private PE mPE;

    // 饸ƥơνλȽ
    private boolean mIsRadiosityStageOver;

    //  Ray 椫
    private boolean mIsStopGenNewRay;

    // hostؤrayȯ׵ФƤ뤫
    private boolean mSendingRayStopComToHost;

    //  ͥѥå塼 : Ӥ delete ɬפʤΤ VectorQueue 
    private VectorQueue	mRadPatchQueue;

    //  Ԥ ray Υ塼λ μΤ PE.java ֤
    // private Queue mRayQueueRef;

    // ray generator for radiosity delta-radiosity 
    private HemiCube    mHemiCube;
    private Ray	[]      mRadiosityRays;	// radiosity  hemicube Ȥ̿

    // reference of member of PEs
    private HashMap 		mRadSpaceMapRef;
    // private Material [] 	mMatrialArrayRef;
    private SearchNextSpcDDA 	mRadDDAUnitRef;    
    private NetIntUnit		mRadNIUnitRef;
    
    public RadioRenderer(PE pe)
    {
	this.mPE 			= pe;
	this.mIsRadiosityStageOver 	= false;
	this.mIsStopGenNewRay		= false;
	this.mSendingRayStopComToHost	= false;
	this.mRadPatchQueue 		= new VectorQueue();
        this.mHemiCube			= null;
	this.mRadiosityRays		= null;	
	this.mRadSpaceMapRef		= this.mPE.peekSpaceMap();
	// this.mMatrialArrayRef		= this.mPE.peekMatArray();
     	this.mRadDDAUnitRef		= this.mPE.peekDDAUnit();
	this.mRadNIUnitRef		= this.mPE.peekNIUnit();
    }
    
    /**
     * ʥ쥤ȯϲǽ
     */
    private boolean isStopGenNewRay()
    {
	return(this.mIsStopGenNewRay);
    }

    /**
     * Radiosity stage Ͻλ 
     */
    private boolean isRadiosityStageOver()
    {
	return(this.mIsRadiosityStageOver);
    }

    /**
     * initialize for Radiosity 
     */
    void initForRadiosity()
    {
	// initialize Network Interface unit
	this.mPE.initNetIntRayBuff();

	// Υơη׻Ԥ
	if(this.mPE.peekSysInfo().isRadiosityCalc() == false)
	{
	    if(Mpi2Option.getInstance().getVerboseMode())
	    {
		String s = Mpi2Option.toHead(VERBOSEHEAD, 
					     this.mPE.getMyRank()) + 
		    "radiosity calculation is off. skip radiosity stage.";
		System.out.println(s);
	    }
	    this.mIsRadiosityStageOver = true;
	    return;		// Υơη׻Ϥʤ
	}

	// Υơη׻Ԥ
	this.mIsRadiosityStageOver 	= false;
	
	// patch ν
	this.initializePatches();	// calcRadioFRG.cc DELETE COMMENT
	
	// HemiCube ν
	this.initializeHemiCube();

	// NIN this.mRadNIUnitRef.initPEStatHostCommBuff(this.mHostCommRecvQueue);

	// volume ν
	/* volume ط NIN 
	   SLList < VolumeClass * > volumeList;
	   this.mVolSet.getAllElementAsList(volumeList);
	   if(!(volumeList.empty()))
	   {
	   initializeVolumes(volumeList);
	   }
	   */
	
	// PE status init
	this.mPE.initBusyIdleStat();

	// hostrayȯ׵Ϥʤ
	this.mIsStopGenNewRay = false; 	 

	// hostؤrayȯ׵̤
	this.mSendingRayStopComToHost = false; 
    }

    /**
     * initForRadiosity() 
     *	- initialize for patches
     *   ѥåνԤ
     *   ̤ͥ饸ƥͥ륮֤ :: NIN
     */
    private void initializePatches()
    {
	/*
	 * ͤѥå򸡺塼Ͽ롥
	 * κ̤ͥͥ륮̤׻
	 */
	for(Iterator i = this.mRadSpaceMapRef.values().iterator(); 
	    i.hasNext();) 
	{
	    Vector objlist = (Vector)(i.next());
	    for(int j = 0; j < objlist.size(); j++)
	    {
		ObjectPrimitive obj = (ObjectPrimitive)(objlist.elementAt(j));

		// TriangleObject ΤߤͲǽ
		if(obj instanceof TriangleObject)
		{
		    TriangleObject tobj = (TriangleObject)obj;

		    // ̤ͥͥ륮 material 
		    if(tobj.getNonRadEnergy() == 0.0)
		    {
			final Material mat = 
			    (this.mPE.peekMatArray())[tobj.getMaterialID()];
			tobj.addUnshotRadiosity(mat.peekDiffEmit());
		    }else
		    {
			// ˽ƤΤ롥skip ä?
			String s = "Error! : non radiosity energy already " + 
			    "initialized, do you want skip radiosity stage?";
			System.out.println(s);
		    }

		    if(tobj.getNonRadEnergy() >= 
		       (this.mPE.peekViewData()).getMinRadEnergy())
		    {
			// ͥ륮ͤ礭äͥ塼
			//  : ƤٸʤΤǡ˥塼
			// äƤȤȤϤʤ
			Assert.out(tobj.isInRadiationQueue() == false, 
				   "Error!: tobj is already in queue.");
			tobj.setInRadiationQueue();
			this.mRadPatchQueue.enq(tobj);
		    }
		}
	    }
	}
    }

    /**
     * initForRadiosity() 
     *	- initialize HemiCube
     *   Hemi_Cube νԤ
     */
    private void initializeHemiCube()
    {
	// make HemiCube for Radiosity calc of this PE 
	this.mHemiCube = 
	    new HemiCube((this.mPE.peekViewData()).getHemres(), 
			 this.mPE.peekID());
	// ray pointer array for communicate HC <-> PE
	// ray μϤκ    
	this.mRadiosityRays = new Ray[this.mHemiCube.getGenRayNum()];
    
	if(Mpi2Option.getInstance().getVerboseMode() == true)
	{
	    String s = Mpi2Option.toHead(VERBOSEHEAD, this.mPE.getMyRank()) + 
		"init radpatches ( " + this.mRadPatchQueue.length() + 
		" ), HCGenRays/once = " + this.mHemiCube.getGenRayNum();
	    System.out.println(s);
	}
    }

    /**
     * radiosity ν ʣܤ delta-radiosity ray 
     * Ƚꡢַ׻Ԥ ray  NetworkInterface 
     */
    private void processingRadiosities()
    {
	/*
	 * δؿƤФˤϰʲξƤ
	 *     塼˰ܤϥ쥤¸ߤƤ			
	 */
	Assert.out(!(this.mPE.peekRayQueue().isEmpty()), 
		   "Error!: there must be ray : processingRadiosities()");
	/*
	 * ʲξλ˥쥤ܼФƤνԤʤ
	 * ʲξϤδؿ˥ȥꤷˤƤΤ 
	 * do{}while;Υ롼פȤʤ롣 
	 * 	   쥤塼¸ߤ
	 * 	   쥤 getProcRadioAtOnce ʲ
	 *	   ХåեƤ				 
	 */
	int i = 0;
	do {
	    Assert.out(this.mPE.peekRayQueue().length() > 0, 
		       "Error!: There must be ray : processingRadiosities().");
	    Ray ray = (Ray)(this.mPE.peekRayQueue().deq());
	    this.processingOneRadiosity(ray); 
	    i++;
	}while((!(this.mPE.peekRayQueue().isEmpty())) && 
	       (i < this.mPE.peekSysInfo().getProcRadioAtOnce()));
	// FIXME Statistics.addRadioProcsAtOnce(i);
    }

    /**
     * processing one delta-radiosity ray packet
     * 	 processingOneRayTrace() ȤۤƱ Hit  ray 
     * 	ۤʤ  (Diff ιԤۤʤ)
     *      ĤΥ쥤֤ʤ NI ؤϤ
     *	Ray Processing Block κȤԤ
     */
    private void processingOneRadiosity(Ray ray)
    {
	// ray ¸ߤְ֤ǤʤĴ٤롣ʤ exit
// 	Assert.out(this.mPE.isDealSpace(ray.peekSpcPos()) == true, 
// 		   "Error!: ray is in invalid subspace. " + 
// 		   "processingOneRadiosity");
 	if(this.mPE.isDealSpace(ray.peekSpcPos()) == false)
	{
	    String s = "DEBUG Error!: rank " + this.mPE.getMyRank() + 
		", PEID " + this.mPE.peekID() + ", ray is in invalid subspace."
		    + " processingOneRadiosity";
	    System.out.println(s + ray);
	    Mpi2CommFactory.getInstance().exit(1);
	}
	
	Vector objArray;
	final int objNumInThisSpc;
	if(this.mRadSpaceMapRef.containsKey(ray.peekSpcPos()) == true)
	{
	    objArray = (Vector)(this.mRadSpaceMapRef.get(ray.peekSpcPos()));
	    objNumInThisSpc = objArray.size();
	}else
	{
	    // ζ֤ˤʪΤ¸ߤʤ
	    objArray = null;
	    objNumInThisSpc = 0;
	}

	double distanceBetweenRayNObject = Double.MAX_VALUE;

	// Ƚ
	Vector3 hitPoint	= new Vector3();	// 
	ObjectPrimitive hitObj 	= null; 		// ʪ
	for(int i = 0; i < objNumInThisSpc; i++)
	{
	    ObjectPrimitive obj = (ObjectPrimitive)(objArray.get(i));
	    if(obj != ray.peekRadObjRef()) // ̤ͤȤϸȽꤷʤ
	    {
		double tmpDist = obj.getDistance(ray, hitPoint);
		Assert.out(tmpDist >= 0.0, "Error!: distance must be >= 0");

		// Τ褦ʪΤξˤϤʬ֤¸ߤʤʬ 
		// hit ǽ롥ξ
		switch(obj.who())
		{
		case ObjectPrimitive.OBJ_TRIANGLE:	
		    break;
		case ObjectPrimitive.OBJ_SPHERE:
		    if(this.mPE.checkHitPosIsInThisSpcPos(hitPoint, 
							  ray.peekSpcPos())
		       == false)
		    {
			// ζ֤ǤϥҥåȤƤʤ
			tmpDist = Double.MAX_VALUE; 
		    }
		    break;
		case ObjectPrimitive.OBJ_NULL:		
		default:
		    throw new RuntimeException("Internal Error! : this obj " + 
					       "cannot intersect calc" + obj);
		}
		if(distanceBetweenRayNObject > tmpDist)
		{
		    // tmpDist ᤤΤǸ򺹤obj򹹿
		    // 򥻥åȤ
		    distanceBetweenRayNObject = tmpDist;
		    hitObj = obj;
		    ray.setHitPos(new Vector3(hitPoint)); // newɬפϤʤ?
		}
	    }
	}
    
	// ѥåٷ׻󼡥쥤ȯõ
	if(hitObj != null) // ̤
	{
	    // ʪΤ ray 򺹤
	    final Material mat = 
		(this.mPE.peekMatArray())[hitObj.getMaterialID()];
	    if(hitObj.who() == ObjectPrimitive.OBJ_TRIANGLE)
	    {
		/*
		 * triangle Τߤ radiosity 
		 *	Diffѥåεٷ׻
		 *	٤ûray  deltaRadiosity 򤳤 
		 * 	patch ѤǳꡤñѤ delta
		 * 	radiosity Ȥ 
		 */
		this.calcNsetIntensityRadio(ray, mat, hitObj);
	    }

	    /*
	     * ȿͶޤԤ
	     * Ǥ ray tracing ˡΤ褦˵٤ Frame Buffer 
	     * ɬפʤΤ ray ȯɤդ 
	     *
	     * ȯ ray  PE Υ塼뤳ȤȤ
	     */
	    CalcRadioSRG.newSecondaryRadiosityRays(ray, hitObj, this.mPE);
	    // delete ray;   äƤ쥤Ͼ deleteRay!
	}else
	{
	    // 򺹤ʤäΤǼ֤ߤ 
	    boolean isSendRay = this.mPE.moveRayToNextSubspace(ray);
	    if(isSendRay == true)	// ray 
	    {
		Assert.out(ray != null, "Error! : Ray is null");
		this.mPE.setRayToNI(ray); // NIΥХåեؤᤳ  
	    }else
	    {
		// 쥤 deleteRay!
		// delete ray;
	    }
	}	
    }

    /**
     * do radiosity
     */
    public void doRadiosity()
    {
	ProfileMon pfm = ProfileMon.getInstance();
	pfm.startStage(ProfileMon.ProfSt_radiosity);
    
	// 
	this.initForRadiosity();

	// ǽ̤ͤäͥ륮פƤ
	double totalInitNonRadEnergy = 
	    RadioRenderer.getTotalUnshotRadiosityEnergy(this.mRadSpaceMapRef);
	
	// System.out.println("DEBUG : total rad = " + totalInitNonRadEnergy);
    
	while(this.isRadiosityStageOver() == false)
	{
	    // ޤ¾
	    this.mPE.recvRays();		

	    /*
	     *  : ǰŪˤϤǤ뤬¤ Network
	     * Interface  enqueue 줿ǥ塼ĹӤԤʤ
	     * 졤ȯԤ٤ʤϤλȯԤƤ롥
	     * this.sendRays();	
	     */
	
	    // processing
	    if(this.mPE.peekRayQueue().hasElements())
	    { 
		// DELETEME System.out.println("DEBUG : proc ray");	
		// ٤ ray ͭ
		Assert.out(this.mPE.isIdle() == false, 
			   "Error!: here must be busy.");

		pfm.occurEv(ProfileMon.ProfEv_ProcRadRays);

		// // this.toBusy();
		this.processingRadiosities(); // buffer ˶ʬ
		this.mPE.globalLoadBalance(); // ray̤rayȯ¤Ԥ
	    }
	    else if(this.mRadPatchQueue.hasElements())
	    {
		/*
		 * 塼˥쥤ʤ && ͥѥå 
		 *		-> ͥѥåͤ
		 * ͥѥå¸ߤΤ busy ˤ롥
		 * 	 toBusy פ⤷ʤͤν
		 *	ŪʤΤ overhead ˤϤʤʤ   
		 */
		this.mPE.toBusy(); 
	    
		/*
		 * 쥤ȯ host  ray ȯ׵᤬
		 * Ĵ٤ 
		 */
		this.getAndPerformHostCommand();
		if(this.isStopGenNewRay() == false) // host Ϥʤ
		{
		    pfm.occurEv(ProfileMon.ProfEv_genHemiCube);
		    this.radiateRadFromMaxRadPatch(); // ray ȯ
		}
	    }
	    else if(this.mRadNIUnitRef.isSendRayBufAllEmpty() == false)
	    {
		// System.out.println("DEBUG : flush");	
		// 쥤̵ && ̵ && Хåե˥쥤ĤäƤ
		Assert.out(this.mPE.isIdle() == false, 
			   "Error!: here must be busy.");

		// ǽʾˤȯ
		this.mRadNIUnitRef.flushSendRaysBuffer();
		pfm.occurEv(ProfileMon.ProfEv_SendingWait);
		
	    }else if((this.mRadNIUnitRef.isNoSendRay() == true) &&
		     (this.mRadNIUnitRef.isNoRecvRay() == true))  
	    {
		// ǤϤʤ && ǤϤʤ
		// DELETEME System.out.println("DEBUG : may idle");	

		// ٤ PEStat  flush
		this.mRadNIUnitRef.flushPEStatSendBuff();   // PEStat  flush
		if(this.mRadNIUnitRef.isSendingPEStat() == false)
		{
		    /*
		     * 쥤ʤ && ٤ ray ̵ && 
		     * λ     && λ (ray && PEStat)
		     * 		== ƤλŻλ
		     */
		    if(this.mPE.isIdle() == false)
		    {
			this.mPE.toIdle();
			pfm.occurEv(ProfileMon.ProfEv_Idle);		    
		    }else
		    {
			// idle ֤Τޤ
			// host Υޥɤм
			this.getAndPerformHostCommand();
			if(this.isRadiosityStageOver() == true)
			{
			    // λޥɤˤϥ塼϶
			    Assert.out(this.mPE.peekRayQueue().isEmpty() &&
				       this.mRadPatchQueue.isEmpty(),
				       "Error!: queues must be empty");
			}
		    }
		}
		Mpi2System.pollingSleep();
	    }else
	    {
		this.mPE.toBusy();
	    }
	}

	// Ͱʲ̤ͥͥ륮ν
	this.finalizeForRadiosity(totalInitNonRadEnergy);

	pfm.endStage(ProfileMon.ProfSt_radiosity);
    }

    /**
     * finalize for radisity
     */
    private void finalizeForRadiosity(double totalInitNonRadEnergy)
    {
	// ͤͥ륮̤ͥͥ륮ν
	double unshotRadEnergy = 
	    RadioRenderer.getTotalUnshotRadiosityEnergy(this.mRadSpaceMapRef);

	String s = Mpi2Option.toHead(NORMALHEAD, this.mPE.peekID()) + 
	    "unshot radenergy (start end) = " + totalInitNonRadEnergy
	    + " " + unshotRadEnergy;
	System.out.println(s);

	// ̤ͥ饸ƥͤ flush 饸ƥͤ˲ä
	for(Iterator i = this.mRadSpaceMapRef.values().iterator(); 
	    i.hasNext(); ) 
	{
	    Vector objlist = (Vector)(i.next());
	    for(int j = 0; j < objlist.size(); j++)
	    {
		ObjectPrimitive obj = (ObjectPrimitive)(objlist.elementAt(j));

		switch(obj.who())
		{
		case ObjectPrimitive.OBJ_TRIANGLE:
		    ((TriangleObject)obj).flushUnshotRadiosityNInitVerInt();
		    break;	    
		case ObjectPrimitive.OBJ_SPHERE:
		case ObjectPrimitive.OBJ_NULL:
		default:
		    break;
		}
	    }
	}
	String head = Mpi2Option.toHead(VERBOSEHEAD, this.mPE.getMyRank());
	if(this.mPE.peekSysInfo().isRadiosityCalc() == false)
	{
	    // radiosity calculation is not executed
	    if(Mpi2Option.getInstance().getVerboseMode() == true)
	    {
		String verbosemes = head +
		    "flushed each patch emittion (skipping radiosity stage).";
		System.out.println(verbosemes);
	    }
	}else
	{
	    // delete this.mHemiCube;	    
	    this.mHemiCube = null;
	    if(Mpi2Option.getInstance().getVerboseMode() == true)
	    {
		String verbosemes = head + 
		    " destroing hemicube. finalize radiosity stage.";
		System.out.println(verbosemes);
	    }
	}

	// λ
	FileInfo fi = FileInfo.getInstance();
	if(fi.isRadioStageResultOut() == true)
	{
	    this.outRadiosityMpi2Data(fi.getRadioStageOutFileName(), true);
	}
    }

    /**
     * Host Υޥɤ᤹롥
     */
    private void getAndPerformHostCommand()
    {
	if(this.mPE.probeHostCommand() == false)
	{
	    return;		// ޥɤμϤʤä
	}
	
	// Host Υޥɤ
	Assert.out(this.mPE.peekHostCommRecvQueue().length() > 0, 
		   "Error!: queue len must be > 0 " + 
		   "RadioRenderer.getAndPerformHostCommand");
	HostCommand hcomm = 
	    (HostCommand)this.mPE.peekHostCommRecvQueue().deq();

	switch(hcomm.getCommand())
	{
	case End_Radiosity: 	
	    // radiosity stage λ
	    this.mIsRadiosityStageOver = true;
	    break;

	case Suspend_GenNewRays: 
	    // ʬ: stop generate new rays
	    this.mIsStopGenNewRay = true;
	    break;
	    
	case Resume_GenNewRays:
	    // ʬ: resume generate new rays
	    this.mIsStopGenNewRay = false;
	    break;
	    
	case Send_Patch_Descriptions:
	    // ۥȤΥѥå׵
	    {
		// ׻ǥѥå host 
		// this.sendOutRadioResultToHost();
		Mpi2Option.getInstance().
		    verboseOut("NIN sendOutRadioResultToHost at " + 
			       "RadioRenderer.getAndPerformHostCommand");
	    }
	    break;

	    // ʲ radiosity ơǤƤʤϤ
	case End_PrepareShading: 	// prepare shading stage λ
	case End_Raytrace: 		// ray-tracing stage λ
	case Issue_BarrierSync: 	// synchronization
	default:
	    String s = "Internal Error! : PE. r " + this.mPE.getMyRank() + 
		"PE.getAndPerformHostCommand received illegal host command";
	    throw new RuntimeException(s);
	}
	    
	if(Mpi2Option.getInstance().getDebugMode())
	{
	    System.out.println(Mpi2Option.toHead(DEBUGHEAD, 
						 this.mPE.getMyRank()) 
			       + "get " + hcomm);
	}
    }
	

    /*------------------------------------------------------------
     * CalcRadioFRG
     *	calculation of First Rays Generation Unit for radiosity 
     *-----------------------------------------------------------*/
    /**
     * get unshot radiosity energy from an object list
     *	Ϥ줿 object list  Map äƤ patch ̤ͥͥ
     *  ֤ 
     */
    private static double getTotalUnshotRadiosityEnergy(HashMap objlistmap)
    {
	double unshotEnergy = 0.0;
	for(Iterator i = objlistmap.values().iterator(); i.hasNext();)
	{
	    Vector objlist = (Vector)(i.next());
	    for(int j = 0; j < objlist.size(); j++)
	    {
		ObjectPrimitive obj = (ObjectPrimitive)(objlist.elementAt(j));

		// TriangleObject ΤߤͲǽ
		if(obj instanceof TriangleObject)
		{
		    unshotEnergy += ((TriangleObject)obj).getNonRadEnergy();
		}
	    }
	}
	return(unshotEnergy);
    }

    /**
     * ͥѥåθ椫Υͥ륮ΤΤ򤷡ͤ
     */
    private void radiateRadFromMaxRadPatch()
    {
	double maxEnergy 	       	= 0.0;
	TriangleObject maxEnergyPatch	= null;
	int deleteIdx 		       	= -1;
	
	VectorQueue patchqueue = (VectorQueue)this.mRadPatchQueue;
	
	for(int i = 0; i < patchqueue.length(); i++)
	{
	    TriangleObject tripatch = (TriangleObject)(patchqueue.peek(i));
	    if(tripatch.getNonRadEnergy() > maxEnergy)
	    {
		maxEnergy = tripatch.getNonRadEnergy();
		maxEnergyPatch = tripatch;
		deleteIdx = i;
	    }
	}
	
	// δؿϥѥå塼Ĺ > 0 ǸƤ֤ɬͥѥå¸ߤ
	Assert.out((maxEnergyPatch != null), 
		   "Error!: max energy patch must be there."); 
	
	this.mRadPatchQueue.del(deleteIdx);
	
	double nonRadEnergyOfThePatch = maxEnergyPatch.getNonRadEnergy();
	
	this.mHemiCube.genRays(maxEnergyPatch, 
			       this.mPE.peekSysInfo().isDistributedRadiation(),
			       this.mRadiosityRays);
	
	int raynum = this.mHemiCube.getGenRayNum();
	for(int i = 0; i < raynum; i++)
	{
	    (this.mPE.peekRayQueue()).enq(this.mRadiosityRays[i]);
	    /*
	     *  version ǤϼΤϤΤɬפä 
	     * pointer ԡƤɤ 
	     * delete this.mpRadiosityRays[i];
	     * ݥ󥿤ϴ˥ꥹȤ˥ԡ졢Τؤݥ
	     *  2 ĤˤʤäΤڤäƤ 
	     */
	    this.mRadiosityRays[i] = null;
	}

	/*------------------------------
	 * Hemi Cube 򤷤뤹
	 *------------------------------*/
	(this.mPE.peekStatistics()).incHCGenNum();
	if(FileInfo.getInstance().isRadioOutAtStep() == true)
	{
	    this.outRadiosityIntermediateResult();
	}

	// Profiling Monitor
	double unshotRadEnergy = -1.0;
	if(FileInfo.getInstance().isTraceOutWithUnshotRadiosity())
	{
	    unshotRadEnergy = 
		RadioRenderer.
		getTotalUnshotRadiosityEnergy(this.mRadSpaceMapRef);
	}
	ProfileMon.getInstance().outHemiCubeInfo(patchqueue.length(), 
						 unshotRadEnergy);

	if(Mpi2Option.getInstance().getVerboseMode())
	{
	    String s = Mpi2Option.toHead(VERBOSEHEAD, this.mPE.getMyRank()) + 
		"make HC : radpat nonrad enegy " + nonRadEnergyOfThePatch +
		", rest nonradpatches " + this.mRadPatchQueue.length();
	    System.out.println(s);
	    if(Mpi2Option.getInstance().getDebugMode())
	    {
		if(unshotRadEnergy < 0.0)
		{
		    unshotRadEnergy = 
			RadioRenderer.
			getTotalUnshotRadiosityEnergy(this.mRadSpaceMapRef);
		}
		String ds = Mpi2Option.toHead(DEBUGHEAD, this.mPE.getMyRank())
		    + "rest of total energy of this PE " + unshotRadEnergy;
		System.out.println(ds);
	    }
	}
    }

    /** 
     * 饸ƥַ̤ν
     */
    private void outRadiosityIntermediateResult()
    {
	FileInfo fi = FileInfo.getInstance();

	if(fi.isInLoop() == true)
	{
	    int output_counter = 0;
	    if(fi.getLoopVarValue() == 
	       (this.mPE.peekStatistics()).getHCGenNum())
	    {
		String outfbase = "";
		// output radiosity file 
		switch(fi.getRadioIntermFileType())
		{
		case 0:		// shot radiosity
		    outfbase = fi.getRadioIntermediateBaseName() + ".shot." + 
			PrintUsing.getDigitString(3, output_counter);
		    this.outRadiosityMpi2Data(outfbase, true);
		    break;
		case 1:		// unshot radiosity
		    outfbase = fi.getRadioIntermediateBaseName() + ".unshot."+ 
			PrintUsing.getDigitString(3, output_counter);
		    this.outRadiosityMpi2Data(outfbase, false);
		    break;
		case 2:		// both
		    outfbase = fi.getRadioIntermediateBaseName() + ".shot." + 
			PrintUsing.getDigitString(3, output_counter);
		    this.outRadiosityMpi2Data(outfbase, true);
		    outfbase = fi.getRadioIntermediateBaseName() + ".unshot."+ 
			PrintUsing.getDigitString(3, output_counter);
		    this.outRadiosityMpi2Data(outfbase, false);
		    break;
		default:
		    throw new RuntimeException("Internal Error! : illegal "+
					       "intermediate radio out type");
		}
		output_counter++;
		// next loop
		fi.nextLoopVar();
	    }
	}
    }    

    /**
     * output radiosity intermediate result
     * ϰ host ˷׻򽸤ɬפߤ
     * 줾 PE  rank  filename Ƥ줾Ϥ롥
     * ϲɬפFIXHERE
     * @param isShotRad == true  λˤϷ׻Ѥߤ radiosity Ϥ
     *		       	   false λˤ̤ radiosity Ϥ
     */
    public void outRadiosityMpi2Data(String outBaseFname, boolean isShotRad)
    {
	// basename + "r"ranknum + "mpi2.geo"
	String outfilename = outBaseFname + "r" + 
	    PrintUsing.getDigitString(3, this.mPE.getMyRank()) + 
	    ".mpi2.geo";
	PrintWriter printer = null;
        try {
	    printer = new PrintWriter(new FileOutputStream(outfilename));
        }
        catch(IOException e)
        {
            throw new RuntimeException("Can not open printer stream at "+
				       "RadioRenderer.outRadiosityMpi2Data."
				       + e);
        }

	// write Hemi-cube num 
	printer.println("# create HC num is " + 
			this.mPE.peekStatistics().getHCGenNum());
	this.writeGeometryFile(printer, isShotRad);
	printer.close();
	    
	// gzip output file
	Mpi2Option.getInstance().
	    verboseOut("NIN : Can not exec gzip now");
	// system(combuf);
	// System.out.println(NORMALHEAD + 
	// " radiosity intermediate result out. [" + buf + "]");
    }	    
    
    /**
     * write geometry file
     * 	@param isShotRad : true  ξˤ shot ΤߤϤ
     *		           false ξˤ unshot ΤߤϤ
     */
    private void writeGeometryFile(PrintWriter p, boolean isShotRad)
    {
	int outputType = ((isShotRad == true) ? 
			  WITH_SHOT_RADOUT : WITH_UNSHOT_RADOUT);
	Material nowmat = null;
	for(Iterator i = this.mRadSpaceMapRef.values().iterator(); 
	    i.hasNext();) 
	{
	    Vector objlist = (Vector)(i.next());
	    for(int j = 0; j < objlist.size(); j++)
	    {
		ObjectPrimitive obj = (ObjectPrimitive)(objlist.elementAt(j));
		if(!(obj instanceof NullObject))
		{
		    if(nowmat != 
		       (this.mPE.peekMatArray())[obj.getMaterialID()])
		    {
			// change material
			nowmat =(this.mPE.peekMatArray())[obj.getMaterialID()];
			p.println("m " + nowmat.peekName());
		    }
		    p.println(obj.getMgfForm(outputType));
		}
	    }
	}
    }

    /*------------------------------------------------------------
     * CalcRadioIC
     *	calculation of Intensity Calcuration unit for radiosity 
     *-----------------------------------------------------------*/
    /**
     * radiosity intensity η׻
     *
     * PE.mRadPatchQueue ؤΥΤǤΥե礷
     * 뤤ϡRadioRenderer ˥᥽åɤߤƤ ICUnit 饢
     * ȤȤͤٷ׻Ϥ
     * ΤǤ褫
     */
    private void calcNsetIntensityRadio(final Ray ray, 
					final Material mat,
					ObjectPrimitive hitobj)
    {
	Assert.out(hitobj.who() == ObjectPrimitive.OBJ_TRIANGLE, 
		   "Error!: now, only TriangleObject can increase its"
		   + "radiosity.");

	TriangleObject hitPatch = (TriangleObject)hitobj;

	double  dAj   = hitPatch.getArea(); // ѥå
	Vector3 Rho_j = new Vector3(mat.peekDiffRef()); // ȿΨ
	// dBi*Fie*Ai(ref.Ray.h)
	Vector3 dBiFieAi  = new Vector3(ray.peekColor()); 
    
	// 饸ƥ dRad = 
	// ȿΨ * 饸ƥ * եե * ͥѥå / 
	// Vector3 dRad = (Rho_j * dBiFieAi) / dAj; 
	Rho_j.mulVecHere(dBiFieAi);	// Rho_j = Rho_j * dBiFieAi
	Rho_j.divScalerHere(dAj);
	Vector3 dRad = new Vector3(Rho_j); 

	hitPatch.addUnshotRadiosity(dRad); // 饸ƥͤ

	// HemiCube.initialize ˤƴ˥ѥåʬϰƤ
	// Ͳ && FRGΥ塼ˤʤ lock ƥ塼
	if((hitPatch.getNonRadEnergy() > 
	    this.mPE.peekViewData().getMinRadEnergy()) && 
	   (hitPatch.isInRadiationQueue() == false)) 
	{
	    hitPatch.setInRadiationQueue(); 	// lock 
	    this.mRadPatchQueue.enq(hitPatch);	// ͥѥå塼

	    if(Mpi2Option.getInstance().getVerboseMode())
	    {
		String s = Mpi2Option.toHead(VERBOSEHEAD, 
					     this.mPE.getMyRank()) + 
		    "enqueue radiation patch : patch non rad enetgy = " + 
		    hitPatch.getNonRadEnergy();
		System.out.println(s);
	    }
	}
    }
}
