/*
 * $Id: RayTracer.java,v 1.11.4.2 2000/06/19 17:46:20 yamauchi Exp $
 *     (M\pi)^2 : Ray-Tracer
 *	Copyright 1999 By YAMAUCHI Hitoshi   
 *	@author YAMAUCHI Hitoshi  
 *	
 * original source 
 *
 *	쥤ȥ졼ˡμ
 * 
 * 1999ǯ827()
 */

package jp.ac.tohoku.is.archi.mpi2.raytrace;

import	jp.ac.tohoku.is.archi.mpi2.common.*;
import	jp.ac.tohoku.is.archi.mpi2.dda.*;
import	jp.ac.tohoku.is.archi.mpi2.net.*;
import	jp.ac.tohoku.is.archi.mpi2.obj.*;
import	jp.ac.tohoku.is.archi.mpi2.host.*;
import	jp.ac.tohoku.is.archi.mpi2.sysparam.*;
import	jp.ac.tohoku.is.archi.mpi2.util.*;

import  jp.ac.tohoku.is.archi.mpi2.jdk12src.Map;
import  jp.ac.tohoku.is.archi.mpi2.jdk12src.Iterator;
import  jp.ac.tohoku.is.archi.mpi2.jdk12src.HashMap;
import  jp.ac.tohoku.is.archi.mpi2.jdk12src.Vector;

/**
 * sub functions for ray-tracing stage
 *
 * ray-tracing ν ʣܤ ray 
 *    Ƚꡢַ׻Ԥ ray  NetworkInterface 
 */
public final class RayTracer 
implements Mpi2Enum, Mpi2Constants, HostConstants, Config
{
    // PE ؤλ
    private PE mPE;

    // 쥤ȥ졼󥰥ơνλȽ
    private boolean mIsRayTracingStageOver;

    //  Ray 椫
    private boolean mIsStopGenNewRay;

    // hostؤrayȯ׵ФƤ뤫
    private boolean mSendingRayStopComToHost;
    
    // ray generator for ray-tracing
    private RayTrFRGUnit mRayTrFRGUnit;

    // reference of member of PEs
    private HashMap 		mRaySpaceMapRef;
    private SearchNextSpcDDA 	mRayDDAUnitRef;    
    private NetIntUnit		mRayNIUnitRef;
    
    public RayTracer(PE pe)
    {
	this.mPE 			= pe;
	this.mIsRayTracingStageOver 	= false;
	this.mIsStopGenNewRay		= false;
	this.mSendingRayStopComToHost	= false;
	this.mRayTrFRGUnit		= new RayTrFRGUnit();
	this.mRaySpaceMapRef		= this.mPE.peekSpaceMap();
     	this.mRayDDAUnitRef		= this.mPE.peekDDAUnit();
	this.mRayNIUnitRef		= this.mPE.peekNIUnit();
    }
    
    /**
     * ray tracing Τν
     */
    public void initForRayTracing()
    {
	/*
	 * ֤Ǥ ray Υ塼ˤϲäƤƤϤʤΤĴ٤
	 * (ȤСradiosity η׻λĤʤɤϤäƤϤʤ)
	 */
	Assert.out(this.mPE.peekRayQueue().isEmpty(), 
		   "Error!: there must be no ray : initForRayTracing()");
	
	if(this.mPE.peekSysInfo().isRayTracingCalc() == false)
	{
	    /*--
	     *-- ray-tracing η׻򤷤ʤȤνϤ
	     *-- ɤΤƤʤǤŬʥѥ᡼
	     *-- ꤷƤ뤬ʾ֤ˤʤʤ褦ˤȤ
	     *-- ̤ΰ̣Ǥ롥̤ñ exit Ƥɤ
	     *-- ⤷ʤYAMAUCHI */
	    if(Mpi2Option.getInstance().getVerboseMode())
	    {
		String s = Mpi2Option.toHead(VERBOSEHEAD, 
					     this.mPE.getMyRank()) + 
		    "ray-tracing calculation is off. skip raytracing " +
		    "stage. no image is generated.";
		System.out.println(s);
	    }

	    // ˽λȤ return
	    this.mIsRayTracingStageOver = true;	
	    return;
	}
    
	// ray tracing ơη׻Ԥ
	this.mIsRayTracingStageOver = false;

	// initialize Network Interface unit
	this.mPE.initNetIntRayBuff();
    
	// initialize First Ray Generation Unit
	this.mRayTrFRGUnit.init(this.mPE.peekSysInfo().
				getFirstRaysGenerationMethod(),
				this.mPE.peekViewData(), 
				this.mPE.peekSpaceParam(), 
				this.mPE.peekID(),
				this.mPE.peekSysInfo().getTotalPENum(),
				this.mPE.getMyRank());

	if(this.mRayTrFRGUnit.canGenerateNextRay() == true)
	{
	    /*
	     * 쥤Υå: ξʤɤͤ
	     * Ȥ餫ʾ礬ꡢξˤϥ쥤κϤ
	     * ʤ
	     */
	    this.shotSomeFRGRays();
	}
    }

    /**
     *   ray ν ʣܤΥ쥤
     */
    public void processingRayTraces()
    {
	/*
	 * δؿƤФˤϰʲξƤ
	 *     塼˰ܤϥ쥤¸ߤƤ
	 */
	Assert.out(!(this.mPE.peekRayQueue().isEmpty()), 
		   "Error!: there must be rays. processingRayTraces");
    
	/*
	 * ʲξλ˥쥤ܼФƤνԤʤ
	 * ʲξϤδؿ˥ȥꤷˤƤΤ 
	 * do{}while;Υ롼פȤʤ롣 
	 * 	   쥤塼¸ߤ
	 * 	   쥤 getProcRayTrAtOnce ʲ
	 */
	int i = 0; // 쥤
	do {
	    Assert.out(this.mPE.peekRayQueue().length() > 0, 
		       "Error!: no ray : processingRayTraces");
	    Ray ray = (Ray)(this.mPE.peekRayQueue().deq());
	    this.processingOneRayTrace(ray);  
	    i++;
	}while((!(this.mPE.peekRayQueue().isEmpty())) && 
	       (i < this.mPE.peekSysInfo().getProcRayTrAtOnce()));
	// Statistics.addRayTraceProcsAtOnce(i); FIXME פϤȤ뤫?
    }


    /**
     * RP Block νԤٷ׻νɬפʤԤ̤ 
     * Host 롣ɬפʤ¾ PE ؤԤ 
     * processingOneRadiosity Ȥʤ궦ʬ뤬 Hit  ray 
     * ۤ   
     */
    public void processingOneRayTrace(Ray ray)
    {
	/*
	 * ʬξˤ ray ¸ߤ֤ôƤʤ PE 
	 *  DDA  η׻Ԥǽ롥 FRG  ray 
	 * ľˤ PE  DDA η׻ԤäơʪΤ¸ߤ
	 * ֤ ray 뤳Ȥǲ򤹤٤äơ 
	 * ray Τ֤Ϥ PE ô֤ǤʤƤϤʤʤ  
	 */
	Assert.out(this.mPE.isDealSpace(ray.peekSpcPos()) == true, 
		   "Error!: ray is in invalid subspace. " + 
		   "processingOneRayTrace");
	
	// Ƚ
	ObjectPrimitive hitobj = this.calcInterSectionWithRayAndObj(ray);
    
	// ٷ׻󼡥쥤ȯõ
	if(hitobj != null)
	{
	    // ʪΤ ray 򺹤
	    final Material mat = 
		(this.mPE.peekMatArray())[hitobj.getMaterialID()];

	    // ʪΤε٤׻ray ˥å
	    CalcRayTrIC.calcNsetIntensityRayTr(ray, hitobj, 
					       this.mPE.peekSysInfo().
					       getShadingScheme());

	    /*
	     * ȿͶޤԤ : ߤ ray  IC.SRG ήˤʤ
	     * Ƥ롥
	     */
	    if(CalcRayTrSRG.
	       newSecondaryRayTracingRays(ray, hitobj, this.mPE.peekViewData(),
					  this.mPE) == false)
	    {
		/*
		 * ȿͤޤ ray [ɤ]ʤä
		 * ޤǷ׻٤ Frame Buffer 롣
		 * 줿ˤϤޤǤε٤Ϥɤ餫 ray 
		 * Ƥ롣줬ǽŪ˾ä Frame
		 * Buffer 롣 
		 */
		this.mPE.insertIntensityIntoSendBuffer(ray);
		// DELEMETE System.out.println("DEBUG insertintnsity");
	    }
	    // delete ray;		// deleteRay! for C++
	    // ray = null;
	    // 򺹤ΤǼ֤ؤϹԤʤ(2쥤Ѥ)  
	}else
	{
	    // õ : ֤򳰤 ray 
	    this.calcRayTrNextSpace(ray);
	}	
    }

    /**
     *  PE ôʬ֤ǤΥ쥤ʪΤθȽ
     */
    private ObjectPrimitive calcInterSectionWithRayAndObj(Ray ray)
    {
	Assert.out(this.mPE.isDealSpace(ray.peekSpcPos()), 
		   "Error!: this ray is in invalid subspace " +
		   "calcInterSectionWithRayAndObj"); 

	/*
	 * ʪΤȤθȽ 
	 */
	Vector objArray;
	final int objNumInThisSpc;
	
	/* DELETEME
	for(Iterator i = this.mRaySpaceMapRef.values().iterator(); 
	    i.hasNext(); ) 
	{
	    Vector objary = (Vector)(i.next());
	    System.out.println("DEBUG : values is " + objary);
	}
	
	for(Iterator i = this.mRaySpaceMapRef.keySet().iterator(); 
	    i.hasNext(); ) 
	{
	    IntVec3 pos = (IntVec3)(i.next());
	    System.out.println("DEBUG : key is " + pos);
	}
	*/

	if(this.mRaySpaceMapRef.containsKey(ray.peekSpcPos()) == true)
	{
	    objArray = (Vector)(this.mRaySpaceMapRef.get(ray.peekSpcPos()));
	    objNumInThisSpc = objArray.size();
	}else
	{
	    // ζ֤ˤʪΤ¸ߤʤ
	    objArray = null;
	    objNumInThisSpc = 0;
	}

	double distanceBetweenRayNObject = Double.MAX_VALUE;

	Vector3 hitPoint	= new Vector3();	// 
	ObjectPrimitive hitobj 	= null; 		// ʪ
	for(int i = 0; i < objNumInThisSpc; i++)
	{
	    ObjectPrimitive obj = (ObjectPrimitive)objArray.get(i);
	    if(obj != ray.peekRadObjRef()) //rayȯ̤ȤϸȽꤷʤ
	    {
		double tmpDist = obj.getDistance(ray, hitPoint);
		Assert.out(tmpDist >= 0.0, "Error!: distance must be >= 0");

		/* DELETEME 
		if(tmpDist < Double.MAX_VALUE)
		{
		    System.out.println("DEBUG: hit");
		}
		*/

		// Τ褦ʪΤξˤϤʬ֤¸ߤʤʬ 
		// hit ǽ롥ξ 
		switch(obj.who())
		{
		case ObjectPrimitive.OBJ_TRIANGLE:	
		    break;
		case ObjectPrimitive.OBJ_SPHERE:
		    if(this.mPE.checkHitPosIsInThisSpcPos(hitPoint, 
							  ray.peekSpcPos())
		       == false)
		    {
			// ζ֤ǤϥҥåȤƤʤ
			tmpDist = Double.MAX_VALUE;
		    }
		    break;
		case ObjectPrimitive.OBJ_NULL:		
		default:
		    throw new RuntimeException("Internal Error! : this obj " + 
					       "cannot intersect calc" + obj);
		}

		if(tmpDist < distanceBetweenRayNObject)
		{
		    // tmpDist ᤤΤǸ򺹤obj򹹿
		    // 򥻥åȤ  
		    distanceBetweenRayNObject = tmpDist;
		    hitobj = obj;
		    ray.setHitPos(new Vector3(hitPoint));
		}
	    }
	}
	/*
	 * Volume Ȥߺ 
	 * NIN
	 */
	/*
	SLList < VolumeClass * > *pVolList = 
	    this.mVolSet.getList(ray.getSpcPos());
	if(!(pVolList.empty()))
	{
	    // volume ¸ߤˤߺѤ׻
	    processingOneRayTraceWithVolume(ray, pVolList, 
					    distanceBetweenRayNObject);
	}
	*/
	// ⤷ volume κѤǤ null ˤ٤礬뤫
	return(hitobj);
    }

    /**
     * DDA ׻ray ֤ۤˤ ray 
     */
    private void calcRayTrNextSpace(Ray ray)
    {
	// 򺹤ʪΤ¸ߤʤ õ (PEGetNExtSpace.cc)
	boolean isSendRay = this.mPE.moveRayToNextSubspace(ray);
 
	if(isSendRay == true)	// ray 
	{
	    this.mPE.setRayToNI(ray); // NIΥХåեؤᤳ  
	}else
	{
	    // ٤ե졼Хåեƥ
	    this.mPE.insertIntensityIntoSendBuffer(ray);	    
	    // delete ray;	// deleteRay! for C++
	    // ray = null;
 	}
    }

    /**
     * 쥤ȯ for ray-tracing
     */
    private void shotSomeFRGRays()
    {
	// ƤӽФȤȤϺǾ 1 ܤȯǽǤϤ
	Assert.out(this.mRayTrFRGUnit.canGenerateNextRay() == true, 
		   "Error!: can not generate any rays. shotSomeFRGRays");

	int maxGenRaysAtOnce = 
	    this.mPE.peekSysInfo().getProcRayTrAtOnce();
	//  * GEN_FRAYS_FACTOR_BY_COMMUN; ???

	int i;
	for(i = 0; i < maxGenRaysAtOnce; i++)
	{
	    //  ray ȯ
	    Ray ray = this.mRayTrFRGUnit.newNextRay();

	    /*
	     * ʣ PE ǶƱ ray ȯˤ ray ȯ
	     * Ƥ줬ô֤ȸ򺹤뤫ʤƤϤ
	     * ʤΤǡȤ꡼ǤäƤºݤˤ PE  
	     * ray ȯʤ礬롣Τ褦ʾˤ 
	     * null äƤ롣 
	     */
	    if(ray != null)	
	    {
		if(this.mPE.isDealSpace(ray.peekSpcPos()) == true)
		{
		    // ray ȯˤ PE ô֤ˤ
		    // Ƿ׻
		    this.mPE.setRayToThisPE(ray);
		}else 
		{
		    // ray ȯˤ PE ô֤ˤʤ
		    if(this.mPE.isDealSpace(ray.peekSpcPos()) == true)
		    {
			/* 
			 * ΤΥޥåפ򸫤ơ֥Ȥ ray 
			 * θ¸ߤ֤¸ߤˤϤô
			 *  PE  : NI 塼 
			 * (Multistage Network ʤΤľǽ)
			 */
			this.mPE.setRayToNI(ray); 
		    }
		    else
		    {
			/*
			 * ΤΥ֥¸ߥޥåפ򸫤ơ֥ 
			 * Ȥ ray θ¸ߤ֤ˤ¸ߤ
			 * ʤˤϡõԤäʪΤ¸
			 * ֤ؤȰư롥
			 */
			this.calcRayTrNextSpace(ray);
		    }
		}
	    }else
	    {
		// ray ȯʤä뤳ȤϤʤ
	    }
	    if(this.mRayTrFRGUnit.canGenerateNextRay() == false)
	    {
		// ray ȯλ
		if(Mpi2Option.getInstance().getVerboseMode())
		{
		    String s = Mpi2Option.toHead(VERBOSEHEAD, 
						 this.mPE.peekID())
			+ "generate rays( " +
			this.mRayTrFRGUnit.getTotalGenRays() + 
			" ) done.";
		    System.out.println(s);
		}
		break;
	    }
	}
	// System.out.println("DEBUG : shot some rays " + i);
    }

    /*------------------------------------------------------------
     * ray tracing
     *-----------------------------------------------------------*/
    public void doRayTracing()
    {
	if(Mpi2Option.getInstance().getVerboseMode())
	{
	    System.out.println(Mpi2Option.toHead(VERBOSEHEAD,	    
						 this.mPE.getMyRank())
			       + "PE in do raytracing OK");
	}

	ProfileMon pfm = ProfileMon.getInstance();
	pfm.startStage(ProfileMon.ProfSt_raytracing);

	// ray-tracing Τν
	this.initForRayTracing();
	if(Mpi2Option.getInstance().getVerboseMode())
	{
	    System.out.println(Mpi2Option.toHead(VERBOSEHEAD,
						 this.mPE.getMyRank())
			       + "PE in inited OK");
	}

	// init status 
	this.mPE.initBusyIdleStat();

	// host  ray ȯ׵Ϥʤ
	this.mIsStopGenNewRay = false; 

	// host ؤ ray ȯ׵̤
	this.mSendingRayStopComToHost = false; 
	
	while(this.isRayTracingStageOver() == false)
	{
	    if(Use_Profile_Monitor)
	    {
		/* PM_Array
		   static int beforerays = -PM_RAY_CHANGE_TICK;
		   int pm_rnum = this.mPE.peekRayQueue().length();
		   if(abs(pm_rnum - beforerays) > (PM_RAY_CHANGE_TICK / 2))
		   {
		   pm_raynum(pm_rnum);
		   beforerays = pm_rnum;
		   }
		   */
	    }

	    // ޤ¾
	    this.mPE.recvRays(); 

	    /*
	     * ¾
	     *  ray  enqueue 줿٤ʤ
	     * 뤿ᡤǤ send Ϥʤ this.sendRays(); 
	     */
	    // this.mNIUnit.flushIntensitySendBuf();// ٤ۥȤ
	
	    // processing
	    if(this.mPE.peekRayQueue().hasElements())
	    {
		/*
		 * 쥤ʤ && ٤ ray 
		 * (RP Block) ray 롣ٷ׻ޤࡣ
		 */
		Assert.out(this.mPE.isIdle() == false, 
			   "Error!: processing rayQueue().hasElements()");
		pfm.occurEv(ProfileMon.ProfEv_ProcRayRays);

		this.processingRayTraces(); 

		// ray ̤ˤä ray ȯ¤Ԥ
		this.mPE.globalLoadBalance(); 
	    }
	    else if(this.mRayTrFRGUnit.canGenerateNextRay() == true)
	    {
		// 塼˥쥤ʤ̤ȯͥ쥤
		this.mPE.toBusy();

		// 쥤ȯ host  ray ȯ׵᤬
		// Ĵ٤ 
		this.getAndPerformHostCommand();
		if(this.mIsStopGenNewRay == false) // host Ϥʤ
		{
		    pfm.occurEv(ProfileMon.ProfEv_genRays); 
		    this.shotSomeFRGRays();
		}
	    }
	    else if(this.mRayNIUnitRef.isSendRayBufAllEmpty() == false)
	    {
		// 쥤ʤ && buff ˥쥤ĤäƤ
		Assert.out(this.mPE.isIdle() == false,
			   "Error!: here must be busy. RayTracer:481");

		// ǽʾˤȯԤߤ
		this.mRayNIUnitRef.flushSendRaysBuffer();
		pfm.occurEv(ProfileMon.ProfEv_SendingWait);
		// System.out.println("DEBUG sending only");
	    }else if((this.mRayNIUnitRef.isNoSendRay() == true)&&//Ǥʤ
		     (this.mRayNIUnitRef.isNoRecvRay() == true)) //Ǥʤ
	    {
		// ٤ PEStat  flush
		this.mRayNIUnitRef.flushIntensitySendBuf(); //   flush
		this.mRayNIUnitRef.flushPEStatSendBuff();   // PEStat  flush
		
		if((this.mRayNIUnitRef.isSendingIntensity() == false) && 
		   (this.mRayNIUnitRef.isSendingPEStat()    == false))
		{
		    // 쥤ʤ && ٤ ray ̵ && λ
		    // && λ (ray &&  && PEStat)
		    // == ƤλŻλ
		    if(this.mPE.isIdle() == false) // busy . idle
		    {
			this.mPE.toIdle();		
			// System.out.println("DEBUG toIdle");
			pfm.occurEv(ProfileMon.ProfEv_Idle);		    
		    }else
		    {
			/* 
			 * ƽʤ idle ΤޤޤǤä
			 * ⤷λΥޥɤ 
			 * endOfRayTracing  true  
			 * ѹ 
			 */
			this.getAndPerformHostCommand(); 
			if(this.isRayTracingStageOver() == true)
			{
			    // λޥɤˤϥ塼϶
			    Assert.out(this.mPE.peekRayQueue().isEmpty(),
				       "Error!: Host say all over, then "
				       + "queue must be empty.");
			}
		    }
		}
		/*
		System.out.println("DEBUG nosend, norecv ray, intensity " +
				   this.mRayNIUnitRef.isSendingIntensity() +
				   ", PEStat " + 
				   this.mRayNIUnitRef.isSendingPEStat());
		System.out.println("DEBUG nosend");
		DELETEME
		*/
		Mpi2System.pollingSleep();
	    }else
	    {
		this.mPE.toBusy();	    
	    }
	}
	/*
	  #ifdef FOR_STANDALONE
	  this.toIdle();
	  #endif
	  */
	this.mRayNIUnitRef.flushIntensitySendBuf();  // ٤ۥȤ
	if(Mpi2Option.getInstance().getVerboseMode())
	{
	    String s = Mpi2Option.toHead(VERBOSEHEAD, this.mPE.getMyRank()) + 
		"out of PE loop at RayTracer.doRayTracing.";
	    System.out.println(s);
	}
	pfm.endStage(ProfileMon.ProfSt_raytracing);
    }
    
    /**
     * ʥ쥤ȯϲǽ
     */
    private boolean isStopGenNewRay()
    {
	return(this.mIsStopGenNewRay);
    }

    /**
     * RayTracing stage Ͻλ 
     */
    private boolean isRayTracingStageOver()
    {
	return(this.mIsRayTracingStageOver);
    }

    /**
     * Host Υޥɤ᤹롥
     */
    private void getAndPerformHostCommand()
    {
	if(this.mPE.probeHostCommand() == false)
	{
	    return;		// ޥɤμϤʤä
	}
	
	// Host Υޥɤ
	Assert.out(this.mPE.peekHostCommRecvQueue().length() > 0, 
		   "Error!: queue len must be > 0 " + 
		   "RayTracer.getAndPerformHostCommand");
	HostCommand hcomm = 
	    (HostCommand)this.mPE.peekHostCommRecvQueue().deq();

	switch(hcomm.getCommand())
	{
	case End_Raytrace: 	
	    // raytracing stage λ
	    this.mIsRayTracingStageOver = true;
	    break;

	case Suspend_GenNewRays: 
	    // ʬ: stop generate new rays
	    this.mIsStopGenNewRay = true;
	    break;
	    
	case Resume_GenNewRays:
	    // ʬ: resume generate new rays
	    this.mIsStopGenNewRay = false;
	    break;
	    
	case Send_Patch_Descriptions:
	    // ۥȤΥѥå׵
	    {
		// ׻ǥѥå host 
		// this.sendOutRadioResultToHost();
		Mpi2Option.getInstance().
		    verboseOut("NIN sendOutRadioResultToHost at " + 
			       "RayTracer.getAndPerformHostCommand");
	    }
	    break;

	    // ʲ ray tracing ơǤƤʤϤ
	case End_Radiosity: 		// radiosity stage λ
	case End_PrepareShading: 	// prepare shading stage λ
	case Issue_BarrierSync: 	// synchronization
	default:
	    String s = "Internal Error! : PE. r " + this.mPE.getMyRank() + 
		"PE.getAndPerformHostCommand received illegal host command";
	    throw new RuntimeException(s);
	}
	    
	if(Mpi2Option.getInstance().getDebugMode())
	{
	    System.out.println(Mpi2Option.toHead(DEBUGHEAD, 
						 this.mPE.getMyRank()) 
			       + "get " + hcomm);
	}
    }
}
