/*
 * $Id: ShadeCalculator.java,v 1.7 2000/06/15 13:55:51 yamauchi Exp $
 *     (M\pi)^2 : PE (Processing Element)
 *	Copyright 1999 By YAMAUCHI Hitoshi   
 *	@author YAMAUCHI Hitoshi  
 *	
 * original source prepareShading.cc
 *
 *	Gouroud Shading μ
 * 
 * 1999ǯ9 7()
 */

package jp.ac.tohoku.is.archi.mpi2.shading;

import  jp.ac.tohoku.is.archi.mpi2.common.*;
import  jp.ac.tohoku.is.archi.mpi2.dda.*;
import  jp.ac.tohoku.is.archi.mpi2.host.*;
import  jp.ac.tohoku.is.archi.mpi2.net.*;
import  jp.ac.tohoku.is.archi.mpi2.obj.*;
import  jp.ac.tohoku.is.archi.mpi2.sysparam.*;
import  jp.ac.tohoku.is.archi.mpi2.util.*;

import  jp.ac.tohoku.is.archi.mpi2.jdk12src.Map;
import  jp.ac.tohoku.is.archi.mpi2.jdk12src.Iterator;
import  jp.ac.tohoku.is.archi.mpi2.jdk12src.HashMap;
import  jp.ac.tohoku.is.archi.mpi2.jdk12src.Vector;


/**
 * Prepare Shading μ
 */
public final class ShadeCalculator 
implements Mpi2Enum, Mpi2Constants, HostConstants, Config
{ 		
    /*
     * PE ؤλ
     */
    private PE mPE;

    /*
     *  stage Ͻλɤ
     */
    private boolean mIsPrepareShadingOver;

    /*
     * ĺ٤Υ塼
     */
    private Queue mVerIntQueue; 

    /**
     * constructor
     */
    public ShadeCalculator(PE pe)
    {
	Assert.out(pe != null, "init with null PE at ShadeCalculator");
	this.mPE = pe;
	this.mIsPrepareShadingOver = false;
	// PROF this.mVerIntQueue = new VectorQueue();
	this.mVerIntQueue = new RingBufferQueue();
    }

    private boolean isPrepareShadingStageOver()
    {
	return(this.mIsPrepareShadingOver);
    }

    /**
     * Shading ν
     *	shading Τη׻Ⱦθ
     */
    public void doPrepareShading()
    {
	switch(this.mPE.peekSysInfo().getShadingScheme())
	{
	case FlatShading:
	    // flat shading ǤϤ뤳ȤϤʤ
	    break;
	case GouroudShading:
	    this.initPrepareShadingStage(); // stage ν start
	    this.prepareGouroudShading();
	    this.finalizePrepareShadingStage();
	    break;
	default:
	    throw new RuntimeException("Error!: I don't know such shading " +
				       "scheme. " + this.mPE.peekSysInfo().
				       getShadingScheme());
	}
    }

    /**
     * ̿Хåեʤɤν
     */
    private void initPrepareShadingStage()
    {
	// ΥơϤ줫(λƤʤ)
	this.mIsPrepareShadingOver = false;
	this.mPE.initBusyIdleStat();

	// ĺХåեޤ
	NetIntUnit niUnit = this.mPE.peekNIUnit();
	int verIntSendThresh = this.mPE.peekSysInfo().getSendVerIntAtOnce();
	niUnit.initVerIntBuff(verIntSendThresh, this.mVerIntQueue);
    }
    
    /**
     * GouroudShading Τξ
     */
    private void prepareGouroudShading()
    {
	/*
	 * NIN
	 * 	pm_change_area(PA_prepshade_e);
	 * 	pm_change_stat(PC_prec_intensity_e);
	 */

	// Cell Ĺĺΰ֤εƸ
	final double cellLen   = this.mPE.peekSpaceParam().getCellLen();
	final double tolerance = cellLen / 64.0; // Ƚ

	// PE λʪξ֤Ȥ˥󤹤
	final Vector3 defSpcMinPos = 
	    this.mPE.peekSpaceParam().peekDefSpaceMin();
	if(!(defSpcMinPos.equals(Vector3.ZERO)))
	{
	    System.out.println("Warning! : μǤdefine spaceκǾͤ "
			       + "(0.0,0.0,0.0)ǤʤȤޤ⤷ʤ");
	}

	IntVec3 spcpos = new IntVec3();
	final int subdivnum = this.mPE.peekSpaceParam().getSubDivNum();
	for(int spz = 0; spz < subdivnum; spz++)
	{
	    for(int spy = 0; spy < subdivnum; spy++)
	    {
		for(int spx = 0; spx < subdivnum; spx++)
		{
		    // ֹ
		    spcpos.set(spx, spy, spz);
		    this.averageSendIntensity(spcpos, defSpcMinPos,
					      cellLen, tolerance);
		}
	    }
	}

	if(REDUCE_VERBOSE_INFO)
	{
	    if(Mpi2Option.getInstance().getVerboseMode())
	    {
		System.out.println(Mpi2Option.toHead(VERBOSEHEAD, 
						     this.mPE.getMyRank()) + 
				   "send Intensity-Vertex.. done.");
	    }
	}

	/*
	 *  PE ôʪΤĺ٤ϤȤꤢ塼ˤ줿
	 * ȤδλȼδλƥơνλԤ 
	 */
	NetIntUnit niUnit = this.mPE.peekNIUnit();
	while(this.isPrepareShadingStageOver() == false)
	{
	    // Хåե flush
	    niUnit.flushVerIntSendBuff();
		
	    // 
	    niUnit.recvVerInt();
	    
	    if((niUnit.isSendingVerInt() == true) ||
	       (niUnit.isRecvingVerInt() == true))
	    {
		// 椢뤤ϼ
		this.mPE.toBusy(); // host ưã	    
		niUnit.recvVerInt();
	    }
	    else
	    {
		// Ȥˤʤʤä
		this.mPE.toIdle(); // host ؽλã
	    }
	    this.getAndPerformHostCommand(); // host ϢԤ
	    Mpi2System.pollingSleep();
	}
	
	if(REDUCE_VERBOSE_INFO)
	{
	    if(Mpi2Option.getInstance().getVerboseMode())
	    {
		System.out.println(Mpi2Option.toHead(VERBOSEHEAD, 
						     this.mPE.getMyRank()) + 
				   "get all Intensity-Vertex.. done.");
	    }
	}
    }

    /**
     * ƶ֤ȤƱĺε٤
     * 	줾ζΤ
     *------------------------------
     * ҤȤĤζ֤ʬ
     *	IMPROVEME ߺѤη˽񤱤ФΤߤ
     */
    private void aveVerIntOneSpcLocal(Vector objVec)
    {
	int len = objVec.size();
	VertexIntensity [] senderIvc = new VertexIntensity[3];
	for(int i = 0; i < senderIvc.length; i++)
	{
	    senderIvc[i] = new VertexIntensity();
	}

	for(int i = 0; i < len; i++)
	{
	    // 긵
	    ObjectPrimitive senderObj = (ObjectPrimitive)(objVec.elementAt(i));
	    // ѷΤ߽
	    if(senderObj.who() == ObjectPrimitive.OBJ_TRIANGLE)
	    {
		TriangleObject tri = (TriangleObject)senderObj;
		tri.getRadVer(senderIvc);
		for(int j = 0; j < len; j++)
		{
		    if(i != j)	// ʬȤǤʤԤ
		    {
			// ټ¦
			ObjectPrimitive recvObj =
			    (ObjectPrimitive)(objVec.elementAt(j));
			if(recvObj.who() == ObjectPrimitive.OBJ_TRIANGLE)
			{
			    TriangleObject recvTri = (TriangleObject)recvObj;
			    // sv   ...sender vertex idx
			    for(int sv = 0; sv < 3; sv++)
			    {
				// addIntensityToVertex  porigonID ĺ
				// ΰ֤Ʊʤ鵱٤ä  
				recvTri.addIntensityToVertex(senderIvc[sv]);
			    }
			}
		    }
		}
	    }
	}
    }


    /** 
     * ζ֤ε٤ʿѲ¾֤
     */
    private void averageSendIntensity(final IntVec3 spcpos, 
				      final Vector3 defSpcMinPos,
				      final double  cellLen,
				      final double  tolerance)
    {
	IntVec3 peid = this.mPE.peekNIUnit().getNewPEIDFromSpcPos(spcpos);
	if(this.mPE.peekID().equals(peid))
	{
	    // ʪο 0 ʤ餳ζ֤ԤɬפϤʤ
	    Map objmap = this.mPE.peekSpaceMap();
	    if(objmap.containsKey(spcpos) == false)
	    {
		// ζ֤ˤʪΤ¸ߤʤ
		return;
	    }

	    Vector objArray = (Vector)(objmap.get(spcpos));
	    int objNumInThisSpc = objArray.size();
			
	    Assert.out(objArray.size() > 0, 
		       "Error!: There must be object(s):averageSendIntensity");
	    /*
	     * 1.ζεͤʿѤ 
	     *  1 ĤʪΤʤнɬפϤʤ 
	     */
	    if(objArray.size() > 1)
	    {
		this.aveVerIntOneSpcLocal(objArray);
	    }
	    
	    /*
	     * 2.ζε٤¾ζ֤ؤ¤
	     *   ζ֤ min, max 
	     */
	    /*
	     * Vector3 spmin = defSpcMinPos + cellLen * iv2v(spcpos); 
	     */
	    Vector3 spmin = new Vector3(spcpos);
	    spmin.mulScalerHere(cellLen);
	    spmin.plusHere(defSpcMinPos);
	    /*
	     * Vector3 spmax = spmin + cellLen * Vector3(1.0, 1.0, 1.0);
	     */
	    Vector3 spmax = new Vector3(1.0, 1.0, 1.0);
	    spmax.mulScalerHere(cellLen);
	    spmax.plusHere(spmin);

	    //   ʪΤФ
	    this.sendVerIntOfThisSpc(objArray, spcpos, 
				     spmin, spmax, tolerance);
	}
    }

	
    /**
     * Ĥζεپ¾롣1Ĥζ֤ʪξ 
     * objArray ؤƤ롣
     */
    private void sendVerIntOfThisSpc(Vector objVec, 	IntVec3 srcspc, 
				     Vector3 spmin,     Vector3 spmax,
				     double tolerance)
    {
	int objnum = objVec.size();
	for(int i = 0; i < objVec.size(); i++)
	{
	    ObjectPrimitive obj = (ObjectPrimitive)(objVec.elementAt(i));

	    /*
	     * ˤʤΤ radiosity Ѥ뤳Ȥǽ 
	     * primitive (ߤϡѥѥåΤ)
	     */
	    if(obj.who() == ObjectPrimitive.OBJ_TRIANGLE)
	    {
		TriangleObject tri = (TriangleObject)obj;

		// ͤФ
		VertexIntensity [] retVerInts3 = new VertexIntensity[3];
		retVerInts3[0] = new VertexIntensity();
		retVerInts3[1] = new VertexIntensity();
		retVerInts3[2] = new VertexIntensity();
		tri.getRadVer(retVerInts3);
		for(int ver = 0; ver < 3; ver++) // ѷĺ3
		{
		    /*
		     * ʪΤΰĤεͤ٤Τɤ
		     * Ƚꤷ롣nonblocking Ԥ
		     */
		    this.testOneIntNsendVerInt(retVerInts3[ver], srcspc, 
					       spmin, spmax, tolerance);
		}
	    }
	}
    }

    /**
     * εپ٤ΤɤȽꤷ
     *   δؿϺǹ addhoc adhoc ؿ̿̾Ƥɤ̤
     *   ϵʬŪɽΤɤʤΤ
     *   1999ǯ9 6()  
     */
    private void testOneIntNsendVerInt(VertexIntensity v, IntVec3 srcspc, 
				       Vector3 spmin,     Vector3 spmax,
				       double   tolerance)
    {
	// verint ѹβǽ
	VertexIntensity verint = new VertexIntensity(v);

	Assert.out(tolerance > 0.0, "Error!: tolerance must be > 0.0");
	
	for(int i = 0; i < 3; i++) // vector i  (x, y, z)
	{
	    /*
	     * ĺ϶ 
	     */
	    if(InDEBUG)
	    {
		if((spmax.get(i) - (verint.peekVertexPos()).get(i)) <
		   (-tolerance))
		{
		    String s = 
			"Error! : vertex is out of range:Large"	+ "\n" + 
			"min	" + spmin 		     	+ "\n" + 
			"ver	" + (verint.peekVertexPos())	+ "\n" + 
			"max	" + spmax		     	+ "\n" + 
			"out axis is [" + i + "], val = " 	+ 
			(spmax.get(i)-(verint.peekVertexPos()).get(i))+ "\n" + 
			"tolerance " + (-tolerance);
		    throw new RuntimeException(s);
		}
		if(((verint.peekVertexPos()).get(i) - spmin.get(i)) <
		   (-tolerance))
		{
		    String s = 
			"Error! : vertex is out of range:Small"	+ "\n" + 
			"min	" + spmin 		     	+ "\n" + 
			"ver	" + verint.peekVertexPos()	+ "\n" + 
			"max	" + spmax		     	+ "\n" + 
			"out axis is [" + i + "], val = " 	+ 
			((verint.peekVertexPos()).get(i)-spmin.get(i))+ "\n" +
			"tolerance " + (-tolerance);
		    throw new RuntimeException(s);
		}
	    }
	    // ޤϳ X,Y,Z  +- 
	    if((spmax.get(i) - (verint.peekVertexPos()).get(i)) < tolerance)
	    {
		// plus Ȥɲ(ˤʤ)
		verint.setDestSpcPos(srcspc.plusNew(AdjDir.get(i * 2)));
		if(this.checkInSpc(verint.peekDestSpcPos()) == true)
		{
		    this.sendVerIntToOtherSpcNrecv(verint);
		}
	    }else if((verint.peekVertexPos().get(i) - spmin.get(i)) < 
		     tolerance)
	    {
		// minus Ȥɲ(ˤʤ)
		verint.setDestSpcPos(srcspc.plusNew(AdjDir.get(i * 2 + 1)));
		if(this.checkInSpc(verint.peekDestSpcPos()) == true)
		{
		    this.sendVerIntToOtherSpcNrecv(verint);
		}
	    }else
	    {
		// θˤϤʤʤ
	    }
	}
	// ʣ (X+Y+, X+Y-, X-Y+, X-Y- : XZ Ʊ)
	Vector3 vertexpos = verint.peekVertexPos();
	double xp = spmax.get(Vector3.X) - vertexpos.get(Vector3.X); 
	double xm = vertexpos.get(Vector3.X) - spmax.get(Vector3.X); 
	double yp = spmax.get(Vector3.Y) - vertexpos.get(Vector3.Y);
	double ym = vertexpos.get(Vector3.Y) - spmax.get(Vector3.Y);
	double zp = spmax.get(Vector3.Z) - vertexpos.get(Vector3.Z);
	double zm = vertexpos.get(Vector3.Z) - spmax.get(Vector3.Z);
	// XY, XZ
	if(xp < tolerance) 
	{
	    //----- XPlus
	    if(yp < tolerance)
	    {	
		// Y+

		// srcspc + AdjDir.get(DIR_XPLUS) + AdjDir.get(DIR_YPLUS);
		IntVec3 dir = srcspc.plusNew(AdjDir.get(DIR_3D.XPLUS));
		dir.plusHere(AdjDir.get(DIR_3D.YPLUS));

		verint.setDestSpcPos(dir);
		if(this.checkInSpc(verint.peekDestSpcPos()) == true)
		{
		    /*
		     * ޤ X+ θ Y+ (ߤ
		     * Ȥʤ) 
		     * Ū Mesh η׻ǤϤΤ褦Ū2
		     * ɤߤ MPI  rank ä
		     *  (IMPROVEME)
		     */
		    // verint.mRestDir = DIR_3D.YPLUS; // X+ . Y+
		    this.sendVerIntToOtherSpcNrecv(verint);
		}
	    }else if(ym < tolerance)
	    {	
		// Y-

		// srcspc + AdjDir.get(DIR_3D.XPLUS) + AdjDir.get(DIR_3D.YMINUS)
		IntVec3 dir = srcspc.plusNew(AdjDir.get(DIR_3D.XPLUS));
		dir.plusHere(AdjDir.get(DIR_3D.YMINUS));
		
		verint.setDestSpcPos(dir);
		if(this.checkInSpc(verint.peekDestSpcPos()) == true)
		{
		    // verint.mRestDir = DIR_3D.YMINUS; // X+ . Y-
		    this.sendVerIntToOtherSpcNrecv(verint);
		}
	    }
	    if(zp < tolerance)
	    {	
		// Z+

		// srcspc + AdjDir.get(DIR_3D.XPLUS) + AdjDir.get(DIR_3D.ZPLUS)
		IntVec3 dir = srcspc.plusNew(AdjDir.get(DIR_3D.XPLUS));
		dir.plusNew(AdjDir.get(DIR_3D.ZPLUS));
		
		verint.setDestSpcPos(dir);
		if(this.checkInSpc(verint.peekDestSpcPos()) == true)
		{
		    // verint.mRestDir = DIR_3D.ZPLUS; // X+ . Z+
		    this.sendVerIntToOtherSpcNrecv(verint);
		}
	    }else if(zm < tolerance)
	    {	
		// Z-
		//srcspc + AdjDir.get(DIR_3D.XPLUS) + AdjDir.get(DIR_3D.ZMINUS)
		IntVec3 dir = srcspc.plusNew(AdjDir.get(DIR_3D.XPLUS));
		dir.plusHere(AdjDir.get(DIR_3D.ZMINUS));
		
		verint.setDestSpcPos(dir);
		if(this.checkInSpc(verint.peekDestSpcPos()) == true)
		{
		    // verint.mRestDir = DIR_3D.ZMINUS; // X+ . Z-
		    this.sendVerIntToOtherSpcNrecv(verint);
		}
	    }
	}else if(xm < tolerance)
	{
	    //----- X Minus
	    if(yp < tolerance)
	    {			
		// Y+

		//srcspc + AdjDir.get(DIR_3D.XMINUS) + AdjDir.get(DIR_3D.YPLUS)
		IntVec3 dir = srcspc.plusNew(AdjDir.get(DIR_3D.XMINUS));
		dir.plusHere(AdjDir.get(DIR_3D.YPLUS));

		verint.setDestSpcPos(dir);
		if(this.checkInSpc(verint.peekDestSpcPos()) == true)
		{
		    // verint.mRestDir = DIR_3D.YPLUS; // X- . Y+
		    this.sendVerIntToOtherSpcNrecv(verint);
		}
	    }else if(ym < tolerance)
	    {			
		// Y-

		//srcspc + AdjDir.get(DIR_3D.XMINUS)+ AdjDir.get(DIR_3D.YMINUS)
		IntVec3 dir = srcspc.plusNew(AdjDir.get(DIR_3D.XMINUS));
		dir.plusHere(AdjDir.get(DIR_3D.YMINUS));
		
		verint.setDestSpcPos(dir);
		if(this.checkInSpc(verint.peekDestSpcPos()) == true)
		{
		    // verint.mRestDir = DIR_3D.YMINUS; // X- . Y-
		    this.sendVerIntToOtherSpcNrecv(verint);
		}
	    }
	    if(zp < tolerance)
	    {		
		// Z+
		
		//srcspc + AdjDir.get(DIR_3D.XMINUS) + AdjDir.get(DIR_3D.ZPLUS)
		IntVec3 dir = srcspc.plusNew(AdjDir.get(DIR_3D.XMINUS));
		dir.plusHere(AdjDir.get(DIR_3D.ZPLUS));

		verint.setDestSpcPos(dir);
		if(this.checkInSpc(verint.peekDestSpcPos()) == true)
		{
		    // verint.mRestDir = DIR_3D.ZPLUS; // X- . Z+
		    this.sendVerIntToOtherSpcNrecv(verint);
		}
	    }else if(zm < tolerance)
	    {	
		// Z-

		//srcspc + AdjDir.get(DIR_3D.XMINUS)+ AdjDir.get(DIR_3D.ZMINUS)
		IntVec3 dir = srcspc.plusNew(AdjDir.get(DIR_3D.XMINUS)); 
		dir.plusHere(AdjDir.get(DIR_3D.ZMINUS));

		verint.setDestSpcPos(dir);
		if(this.checkInSpc(verint.peekDestSpcPos()) == true)
		{
		    // verint.mRestDir = DIR_3D.ZMINUS; // X- . Z-
		    this.sendVerIntToOtherSpcNrecv(verint);
		}
	    }
	}
	//-------------------- YZ
	if(yp < tolerance) 
	{
	    //----- YPlus
	    if(zp < tolerance)
	    {			
		// Z+
		// srcspc + AdjDir.get(DIR_3D.YPLUS) + AdjDir.get(DIR_3D.ZPLUS)
		IntVec3 dir = srcspc.plusNew(AdjDir.get(DIR_3D.YPLUS));
		dir.plusHere(AdjDir.get(DIR_3D.ZPLUS));

		verint.setDestSpcPos(dir);
		if(this.checkInSpc(verint.peekDestSpcPos()) == true)
		{
		    // verint.mRestDir = DIR_3D.ZPLUS; // Y+ . Z+
		    this.sendVerIntToOtherSpcNrecv(verint);
		}
	    }else if(zm < tolerance)
	    {	
		// Z-

		//srcspc + AdjDir.get(DIR_3D.YPLUS) + AdjDir.get(DIR_3D.ZMINUS)
		IntVec3 dir = srcspc.plusNew(AdjDir.get(DIR_3D.YPLUS));
		dir.plusHere(AdjDir.get(DIR_3D.ZMINUS));
		verint.setDestSpcPos(dir);
		if(this.checkInSpc(verint.peekDestSpcPos()) == true)
		{
		    // verint.mRestDir = DIR_3D.ZMINUS; // Y+ . Z-
		    this.sendVerIntToOtherSpcNrecv(verint);
		}
	    }
	}else if(ym < tolerance)
	{
	    //----- Y Minus
	    if(zp < tolerance)
	    {	
		// Z+

		//srcspc + AdjDir.get(DIR_3D.YMINUS) + AdjDir.get(DIR_3D.ZPLUS)
		IntVec3 dir = srcspc.plusNew(AdjDir.get(DIR_3D.YMINUS));
		dir.plusHere(AdjDir.get(DIR_3D.ZPLUS));

		verint.setDestSpcPos(dir);
		if(this.checkInSpc(verint.peekDestSpcPos()) == true)
		{
		    // verint.mRestDir = DIR_3D.ZPLUS; // Y- . Z+
		    this.sendVerIntToOtherSpcNrecv(verint);
		}
	    }else if(zm < tolerance)
	    {	
		// Z-

		//srcspc + AdjDir.get(DIR_3D.YMINUS)+ AdjDir.get(DIR_3D.ZMINUS)
		IntVec3 dir = srcspc.plusNew(AdjDir.get(DIR_3D.YMINUS));
		dir.plusHere(AdjDir.get(DIR_3D.ZMINUS));

		verint.setDestSpcPos(dir);
		if(this.checkInSpc(verint.peekDestSpcPos()) == true)
		{
		    // verint.mRestDir = DIR_3D.ZMINUS; // Y- . Z-
		    this.sendVerIntToOtherSpcNrecv(verint);
		}
	    }
	}
    }
    
    /**
     * ζֻɸ⤫
     */
    private boolean checkInSpc(final IntVec3 spcpos)
    {
	final IntVec3 maxIdx = 
	    this.mPE.peekSpaceParam().peekMaxDefSpcIdxPlusOne();
	int x = spcpos.get(IntVec3.X);
	if((x < 0) || (x >= (maxIdx.get(IntVec3.X))))
	{
	    return(false);
	}
	int y = spcpos.get(IntVec3.Y);
	if((y < 0) || (y >= (maxIdx.get(IntVec3.Y))))
	{
	    return(false);
	}
	int z = spcpos.get(IntVec3.Z);
	if((z < 0) || (z >= (maxIdx.get(IntVec3.Z))))
	{
	    return(false);
	}
	return(true);
    }


    /**
     * پ
     */
    private void sendVerIntToOtherSpcNrecv(VertexIntensity verint)
    {
	IntVec3 destPos   = verint.peekDestSpcPos();
	NetIntUnit niUnit = this.mPE.peekNIUnit();
	IntVec3 destPEId  = niUnit.getNewPEIDFromSpcPos(destPos);

	if(destPEId.equals(this.mPE.peekID()))
	{
	    // ֤ϼʬôΤä
	    this.addIntensity(verint);
	}else
	{
	    // ֤ϼʬôΤǤϤʤä -> 
	    
	    // PEID -> rank
	    int destRank = this.mPE.getRank(destPEId);

	    // Τϰ new 
	    VertexIntensity sendVerInt = new VertexIntensity(verint);

	    // rank  PE 
	    niUnit.sendVerInt(sendVerInt, destRank);

	    // ƥ
	    niUnit.recvVerInt();
	
	    // λƤ塼ƤФ
	    while(this.mVerIntQueue.length() > 0)
	    {
		VertexIntensity v = (VertexIntensity)(this.mVerIntQueue.deq());
		this.addIntensity(v);
		// delete pV;		// Τ new Ƥ
	    }
	}
    }

    /**
     * addIntensity
     *	٤ä뤫Ĵ١ǽʤä
     *	äĺοФƤ
     */
    private void addIntensity(VertexIntensity verint)
    {
	Map objmap = this.mPE.peekSpaceMap();
	if(objmap.containsKey(verint.peekDestSpcPos()) == true)
	{
	    Vector objVec = (Vector)(objmap.get(verint.peekDestSpcPos()));
	    int destObjNum = objVec.size();

	    for(int i = 0; i < destObjNum; i++)
	    {
		ObjectPrimitive obj = (ObjectPrimitive)objVec.elementAt(i);
		// ٲäƤɤ(ĤΥݥꥴ󤬽ޤäĺ
		// ǽ)  
		// ϸߥѥåΤ
		if(obj.who() == ObjectPrimitive.OBJ_TRIANGLE) 
		{
		    TriangleObject tri = (TriangleObject)obj;
		    // ݥꥴ ID  vertex ֹӤƱʤвä
		    tri.addIntensityToVertex(verint);
		}
	    }
	}
    }

    /**
     * GouroudShading Ԥͽ׻
     *   ׻ɬפʾϽλƤ
     *	̿Ʋû٤ʿѤ
     */
    private void finalizePrepareShadingStage()
    {
	Map objMap = this.mPE.peekSpaceMap();

	for(Iterator i = objMap.values().iterator(); i.hasNext(); ) 	
	{
	    Vector objVec = (Vector)(i.next());
	    for(int j = 0; j < objVec.size(); j++)
	    {
		ObjectPrimitive obj = (ObjectPrimitive)(objVec.elementAt(j));
		if(obj.who() == ObjectPrimitive.OBJ_TRIANGLE)
		{
		    TriangleObject tri = (TriangleObject)(objVec.elementAt(j));
		    tri.calcVerIntAveVal();
		}
	    }
	}
    }

    /**
     * Host Υޥɤ᤹롥
     */
    private void getAndPerformHostCommand()
    {
	if(this.mPE.probeHostCommand() == false)
	{
	    return;		// ޥɤμϤʤä
	}
	
	// Host Υޥɤ
	Assert.out(this.mPE.peekHostCommRecvQueue().length() > 0, 
		   "Error!: queue len must be > 0 " + 
		   "RadioRenderer.getAndPerformHostCommand");
	HostCommand hcomm = 
	    (HostCommand)this.mPE.peekHostCommRecvQueue().deq();

	switch(hcomm.getCommand())
	{
	case End_PrepareShading: 	// prepare shading stage λ
	    this.mIsPrepareShadingOver = true;
	    break;

	    // ʲ shading stage ǤȯʤϤ
	case End_Radiosity: 	
	case End_Raytrace:
	case Suspend_GenNewRays: 
	case Resume_GenNewRays:
	case Send_Patch_Descriptions:
	case Issue_BarrierSync: 	// synchronization
	default:
	    String s = "Internal Error! : PE. r " + this.mPE.getMyRank() + 
		"PE.getAndPerformHostCommand received illegal host command";
	    throw new RuntimeException(s);
	}
	    
	if(Mpi2Option.getInstance().getDebugMode())
	{
	    System.out.println(Mpi2Option.toHead(DEBUGHEAD, 
						 this.mPE.getMyRank()) 
			       + "get " + hcomm);
	}
    }
}
