// -*- mode: Java; file-coding-system: *euc-japan*; -*-
// $Id: Marshal.java,v 1.4 1999/10/05 14:25:29 yamauchi Exp $
// Copyright (C) 1999 MAEDA Atusi

package jp.ac.tohoku.is.archi.mpi2.util;

import mpi.Datatype;
import mpi.MPI;

/**
 * {@link EncodedData}뤿Υ饹.
 */
public abstract class Marshal {
  interface EncodedDataFactory {
    EncodedData create();
  }
  static class ByteArrayEncodedDataFactory implements EncodedDataFactory {
    public EncodedData create() {
      return new ByteArrayEncodedData();
    }
  }
  static class DoubleArrayEncodedDataFactory implements EncodedDataFactory {
    public EncodedData create() {
      return new DoubleArrayEncodedData();
    }
  }
  private static EncodedDataFactory factory =
    new DoubleArrayEncodedDataFactory();

  /**
   * {@link EncodedData}1֤.
   */
  public static EncodedData createEncodedData() {
    return factory.create();
  }

  /**
   * ǡ1Ĥ뤿{@link EncodedData} convenience method.
   * EncodedDataꡢ{@link Encodable}򤽤encodeơ
   * Ǥ֤ˤ֤
   * @param data ٤ǡ
   */
  public static EncodedData createEncodedData(Encodable data) {
    EncodedData e = factory.create();
    Encoder encoder = e.getEncoder();
    encoder.prepare(data);
    encoder.finishPrepare();
    encoder.encode(data);
    encoder.finishEncode();
    return e;
  }

  /**
   * Ʊ{@link Encodable}뤿{@link EncodedData}
   *  convenience method.
   * EncodedDataꡢ{@link Encodable}Ƥ򤽤encodeơ
   * Ǥ֤ˤ֤
   * @param array ٤EncodableEncodableϤ٤Ʊ
   * Υ֥ȤǤʤФʤʤ.
   */
  public static EncodedData createEncodedData(Encodable[] array) {
      EncodedData e = factory.create();
      
      // System.out.println("DEBUG : e " + e);

      Encoder encoder = e.getEncoder();

      // System.out.println("DEBUG : encoder " + encoder);

      encoder.allocate(array[0], array.length);

      // System.out.println("DEBUG : allcate onesize " + array[0].sizeInDouble() 
      // + ", len " + array.length); DELETEME

      for (int i = 0; i < array.length; i++) {
	  // System.out.println("DEBUG : [" + i + "]");	  
	  encoder.encode(array[i]);
      }
      // System.out.println("DEBUG encode");	  
      encoder.finishEncode();
      // System.out.println("DEBUG finish");	  
      return e;
  }

  /**
   * ǡ1Ĥ뤿{@link EncodedData} convenience method.
   * EncodedDataꡢǡ1ļ礭ΥХåեդ
   * Ǥ֤ˤ֤
   * @param od Encodable礭Τ餻뤿ObjectDecoder
   */
  public static EncodedData createEncodedData(ObjectDecoder od) {
    EncodedData e = factory.create();
    Decoder decoder = e.getDecoder();
    decoder.prepare(od);
    decoder.finishPrepare();
    return e;
  }

  /**
   * Ʊ{@link Encodable}뤿{@link EncodedData}
   *  convenience method.
   * EncodedDataꡢǤΥǡդ
   * Ǥ֤ˤ֤
   * @param od ͽEncodable礭򼨤ObjectDecoder
   * @param nelem Encodableǿ
   */
  public static EncodedData createEncodedData(ObjectDecoder od, int nelem) {
    EncodedData e = factory.create();
    Decoder decoder = e.getDecoder();
    decoder.allocate(od, nelem);
    return e;
  }

  /**
   * ޤä礭Υǡ뤿{@link EncodedData} 
   * convenience method.
   * EncodedDataꡢꤷ礭ΥХåեդ
   * Ǥ֤ˤ֤
   * @param size Encodable礭(ǿ)
   */
  public static EncodedData createEncodedData(int size) {
    EncodedData e = factory.create();
    Decoder decoder = e.getDecoder();
    decoder.allocate(size);
    return e;
  }

  private static Datatype mode = MPI.DOUBLE;
  /**
   * {@link EncodedData}μˡbyte[]ѤΤdouble[]Ѥ
   * ΤɤȤڤؤ.
   * @param mode MPI.BYTEޤMPI.DOUBLE
   */
  public static void setMode(Datatype mode) {
    if (mode == MPI.BYTE) {
      factory = new ByteArrayEncodedDataFactory();
      Marshal.mode = mode;
    } else if (mode == MPI.DOUBLE) {
      factory = new DoubleArrayEncodedDataFactory();
      Marshal.mode = mode;
    } else {
      throw new RuntimeException("Illegal mode: " + mode);
    }
  }

  /**
   * ߡ{@link #createEncodedData}ƤӽФȡbyte[]Ѥ
   * double[]ѤΤɤ餬֤뤫Υ⡼ɤ֤.
   * @returns MPI.BYTEޤMPI.DOUBLE
   */
  public static Datatype getMode() {
    return mode;
  }
}
