#! /bin/sh
#
# $Id: proc_r2mark.sh,v 1.2 2000/07/03 10:41:07 yamauchi Exp $
# R^2mark : 
# 	run benchmarks for the Java parallel version
#	Copyright (C) 2000 YAMAUCHI Hitoshi  
#
R2MARKURL="http://www.mpi-sb.mpg.de/~hitoshi/research/r2mark/index.shtml"

isexit="no"
if [ ! $# -eq 2 ]; then
    isexit="yes"
else
    if [ $1 != "office"  -a  $1 != "soda"  -a  $1 != "conf" ] ; then
	isexit="yes"
    else
	SIMNAME=$1;
    fi

    if [ $2 != "test" -a $2 != "inaccurate" -a $2 != "accurate" ] ; then
	isexit="yes"
    else
	ACCURACY=$2
    fi
fi

if [ ${isexit} = "yes" ]; then
    echo "Usage: proc_r2mark.sh [office|soda|conf] [test|accurate|inaccurate]"
    echo "	[office|soda|conf] is a scene name."
    echo "	[test|accurate|inaccurate] is rendering option."	
    echo "	accurate takes much time than inaccurate."	
    echo "	Some results are in the next URL."	
    echo "	${R2MARKURL}"
    echo "R^2mark 1.0 : Copyright (C) 2000 YAMAUCHI Hitoshi"
    exit;
fi

#
# variables
#
#----- for mpiJava
MPIJAVADIR=../../../../mpiJava-1.2
MPIJAVASCRIPTS=${MPIJAVADIR}/src/scripts
MPIJAVALIB=${MPIJAVADIR}/lib

#----- for Mpi2
MPI2CLASSPATH=../../../../Java
MPI2=jp.ac.tohoku.is.archi.mpi2.host.Mpi2MPI

#----- exports environment variables
CLASSPATH=.:${MPI2CLASSPATH}:${MPIJAVALIB}/classes:${CLASSPATH}
export CLASSPATH

PATH=.:${MPIJAVASCRIPTS}:${PATH}
export PATH

LD_LIBRARY_PATH=${MPIJAVALIB}/lib:${LD_LIBRARY_PATH}
export LD_LIBRARY_PATH

#----- for simuration
SIMBASE=${SIMNAME}.${ACCURACY}
MPI2CONFIG="${SIMBASE}.conf"
MPI2VIEW="${SIMBASE}.view"
MACHINEFILE='machinefile'
MACHINEOPT="-machinefile ${MACHINEFILE}"
MPIRUN='prunjava'
LOGBASE=log.r2mparallel.java.${SIMBASE}

#
# create machine files
#
if [ ! -f ${MACHINEFILE} ]; then
    echo "There is no machinefile for mpich, try to generate."
    echo "hostname > ${MACHINEFILE}"
    hostname > ${MACHINEFILE}
    for i in 1 2 3 4 5 6 7
    do
	echo "hostname >> ${MACHINEFILE}"
	hostname >> ${MACHINEFILE}
    done
fi

# 
# start simulation
#
date > log.time.$$
for pe in 001 002 004 008 
do
    #----- expand maximum heap size 
    # this is the case of JDK 1.2.1...
    # Java's problem is not its performance, memory usage is the problem.
    case ${pe} in
    001)
	JAVAOPTION="-Xmx512m"
	;;
    002)
	JAVAOPTION="-Xmx384m"
	;;
    004)
	JAVAOPTION="-Xmx256m"
	;;
    008)
	JAVAOPTION="-Xmx160m"
	;;
    esac
    export JAVAOPTION
    #----- expand maximum heap size 
    LOGNAME=${LOGBASE}.${pe}
    CONFNAME=${MPI2CONFIG}.${pe}
    PENUM=`expr ${pe} + 1`
    PENUMOPT="$PENUM"
    if [ ! -f ${LOGNAME} ]; then
	echo "running ${MPIRUN} ${PENUMOPT} ${MPI2} --conf ${CONFNAME} > ${LOGNAME} 2>&1, exported JAVAOPTION=${JAVAOPTION}"
	${MPIRUN} ${PENUMOPT} ${MPI2} --conf ${CONFNAME} > ${LOGNAME} 2>&1
	date >> log.time.$$
    else
	echo "There is ${LOGNAME}. Skip."
    fi
    # ɽ
done
